# README #

This library has been programmed in C++. Its classes contain algorithms for
instantiating, evaluating, transforming, and creating test suites from symbolic
finite state machines (SFSM). Here, SFSMs practically are finite state machines
with quantifier free first order expressions over sets of input and output
variables as transition input and output symbols. In a journal article
submitted for publishing we describe symbolic and concrete executions of SFSMs,
their semantics and how to create sets of finite sequences of variable
valuations to test whether some implementation conforms to some SFSMs.  Here,
we provide an implementation of these algorithms and some transformation
algorithms.
This library is a sister library to the [`libfsmtest`](https://bitbucket.org/JanPeleska/libfsmtest) library. 

Note that this library contains some dependencies as submodules, so you should run

    git submodule init
    git submodule update

or similar commands after cloning.

## Documentation
Currently, all documentation is in this README or inline

## Dependencies

This project requires the following software:

* [`cmake`](https://cmake.org/)
* [`Z3`](https://github.com/Z3Prover/z3/) (libsfsmtest can be modified to use different SMT-Solvers that accept SMT-LIB2 input)

We tested compilation with the following compilers:

| Compiler | Versions   |
|----------|------------|
| GCC      | 11.2, 12.1 |
| Clang    | 13.0       |
| MSVC     | 14.32      |

### Installation of dependencies

Generally, we assume `cmake` to be installed for the build process and `z3` to
be installed for library execution.  Furthermore, we assume both to be found in
a directory listed in the `PATH` environment variable.
To install these dependencies, please see to your operating system
distribution package repository, if available, or see the linked project
homepages for prebuilt binaries or build and installation instructions. For
some platforms, we provide a guide on how to obtain the dependencies below.

#### MacOS
Use `brew` to install the dependencies.

    brew install cmake
    brew install z3

Make sure that `$(brew --prefix)/bin` is in your PATH environment variable.
For the build instructions we assume that `cmake` can be found in some
directory in the PATH environment variable.
`libsfsmtest` assumes that `z3` can be found in some directory in the PATH
environment variable.

#### Arch linux
Obtain the following packages from the official repos:

    pacman -S extra/cmake
    pacman -S community/z3

#### Debian, Ubuntu and other debian based distros
Obtain the following packages from the official repos:

    apt-get install cmake
    apt-get install z3

#### Fedora
Obtain the following packages from the official repos:

    yum install cmake
    yum install z3

#### Windows
TODO

## Building

To build the library, a compiler supporting the C++17 standard is required.
For building under Linux, MacOS or Windows, we offer CMake build files that require at
least CMake 3.20.
Currently, the libsfsmtest can only generate test suites on linux.

### Building with CMake and make

To build the library using the CMake files, create a build directory and invoke
CMake, pointing it to the `src` directory:

    mkdir build;
    cd build;
    cmake ../src

Alternatively, you can use the `build.sh` script on Linux, which will generate the
directories `build.Debug` and `build.Release`, for the debug and the release
build respectively.

Then, regardless of which option you chose, navigate to the build directory of
your choice, i.e. either `build`, `build.Debug` or `build.Release`, and simply
invoke `make`. This will then build the library and all executables shipped
with it. The resulting binaries will be located in the build directory as well.

### Building with CMake and Xcode

To build the library using the CMake files you can use the `build-for-xcode.sh`
script, which will generate an Xcode project file in a build directory `xcodebuild.Debug`
Then, use the project file to build the library and all binaries.

### Building with CMake and Visual Studio

We assume that the cmake binary is in your `PATH` environment variable. Then,
you can execute the `build-for-vs.bat` script, which will generate a Visual
Studio Solution in a build directory `visualstudio.Debug`
Afterwards, use the solution file to build the library and all binaries.

### Build artifacts

The build scripts are configured to generate a set of files. The core piece is
a library file, `libsfsmtest.a` (or `sfsmtest.lib` on Windows systems), that is
used as the basis for all other executables and that can be included in other
projects as well (see the LICENSE file for details).
Based on this library we provide the executables `usage-demo`, which
demonstrates the usage of some library functions and `sfsmMethod`, which can be
used to generate test suites for SFSMs.

For general usage of the library, most use cases are presented at least once in
the source code of `usage-demo`. It can be executed as a test of the library.

## Container images
We also provide a container image containing the sfsmMethod executable.
You can build it with the provided Dockerfile `Dockerfile-sfsmMethod`:

    docker build -t sfsm_method -f Dockerfile-sfsmMethod

The image uses a volume `/data` for the files `sfsmMethod` needs to work. If
there is an CSV file `example.csv` in the current directory specifying some
SFSM you wish to process, you could execute the built image as follows to
calculate a test suite that is complete for all implementations with at most 1
additional state:

    docker run --volume $PWD:/data sfsm_method 1 example.csv

The command line arguments after `sfsm_method` are passed to the contained
`sfsmMethod` executable.

# Usage
## Expression format
All expressions in the library are parsed and presented as S-expressions, adhering to the following grammar (in EBNF notation):

    OPERAND = IDENTIFIER | OPERATOR;
    OPERANDLIST = {{OPERANDLIST} OPERAND};
    OPERATOR = '(' OPERANDLIST ')';
    EXPRESSION = OPERAND;

Identifiers are strings from the following set of characters: `_a-zA-Z!%&/=?+*<>@.,0-9'\-^`.

## Input format
Currently, this library only supports a Comma Separated Values (CSV) style
input format for SFSMs, where the column delimiter of the CSV file has to be a
semicolon.  The transitions of the SFSM are specified by a table at the start
of the file and further information about the SFSM is stored in sections
subsequent to the table. These additional sections are completely stored in the
first column of the file.
Examples can be found in the `resources` directory, e.g. `resources/brake/brake.csv`.

### Transition table
The table rows correspond to SFSM states whereas the columns correspond to
guard conditions.  The first row of the CSV file shall start with an empty cell
and then list guard conditions, column by column.  Each subsequent row of the
transition table starts with the name of the corresponding state in the first
column of the CSV file and then describes the targets and output expressions
this state reaches when a guard conditions is fulfilled in the following
columns.  A transition is specified as follows: If state `s_0` has a transition
to state `s_1` when guard condition `(< x c_0)` is fulfilled and the output
expression for that transition is `(= y (* x 2))`, then in the row beginning
with `s_0` and in the column for guard condition `(< x c_0)` there is an entry
`s_1/(= y (* x 2))`. Multiple transitions for the same target state and guard
condition can be delimited by commas.
See the following table as an example. Here, there are states `s_0` and `s_1`,
guard conditions `g_0...g_3` and output expressions `o_1...o_3`.
If guard conditions `g_0` or `g_1` are satisfied, state `s_0` will transition
to state `s_1` with output expression `o_3` and `o_2` respectively.
Note that for this example to work as is, we would have to introduce the
identifiers `g_0...g_3` and `o_1...o_3` as boolean identifiers in the
`identifiers` section (see below) and might want to equate them to some more
complex formula in the `invariants` section.

        |   g_0   |   g_1   |   g_2   |   g_3   |
    ---------------------------------------------
    s_0 | s_1/o_3 | s_1/o_2 |         | s_0/o_2 |
    s_1 | s_1/o_3 |         | s_0/o_1 | s_0/o_2 |

### Additional sections
Further sections are all located in the first column of the CSV file, separated
from the transition table and each other by empty rows.
Each section starts with a keyword, specifying the contents of the section
listed in the following rows. Empty rows within a section are not allowed.
The section keywords are `invariants`,`identifiers`,`init`,`inputsymbols` and `outputsymbols`.

#### Invariants
The `invariants` section contains S-expressions whose conjunction has to be
satisfied for every pre- and post-state of a transition.  This can be used to
specify bounds for variables and to assign constant values to some identifiers.

#### Identifiers
The `identifiers` section declares identifiers and their types. Each line declares one identifier by the following grammar (EBNF):

    LOCATION = "in" | "input" | "out" | "output";
    TYPE = "int" | "integer" | "real" | "bool" | "boolean";
    DECLARATION = {LOCATION} TYPE IDENTIFIER;

An identifier declared with location `in` or `input` is considered an input variable,
an identifier with location `out` or `output` is considered an output variable.

#### Init
The `init` section contains S-expressions whose conjunction has to hold before
the first transition is made.
This is important if the initial state is not completely specified, i.e. if
there are input variable valuations admitted by the invariant that do not
satisfy any guard condition of the initial state, as in the case of states that
are not completely specified, there is an implicit transition looping to the
state itself with an output expression that asserts that all output variables
retain the value they had in the transition before or in case of the first
transition a value satisfying the `init` expressions. 

#### Input symbols
The `inputsymbols` section allows to introduce further guard conditions that
are not used in the transition table but will be used for all calculations as
if they were guard conditions in the SFSM. This is important for the equivalence
class calculation. Expected mutations in the implementation can be added here.

#### Output symbols
The `outputsymbols` section allows to introduce further output expressions that
are not used in the transition table but will be used for all calculations as
if they were output expressions in the SFSM. This is important for the equivalence
class calculation. Expected mutations in the implementation can be added here.

### Complete example
The following sketch shows a table that, if given as a CSV input file, would
constitute a valid input file. When interpreting the `|` delimiters as the
delimiters of a CSV file, this is actually valid input.

                           | (< x c) | (>= x c) | (> x c) | (>= x (* c 2)) |
    s_0                    | s_1/o_3 | s_1/o_2  |         | s_0/o_2        |
    s_1                    | s_1/o_3 |          | s_0/o_1 | s_0/o_2        |
                           |         |          |         |                |
    identifiers            |         |          |         |                |
    input real x           |         |          |         |                |
    real c                 |         |          |         |                |
    output real y          |         |          |         |                |
    bool o_1               |         |          |         |                |
    bool o_2               |         |          |         |                |
    bool o_3               |         |          |         |                |
                           |         |          |         |                |
    invariants             |         |          |         |                |
    (= c 500)              |         |          |         |                |
    (= o_3 (= y (* x 2)))  |         |          |         |                |
    (= o_2 (= y x))        |         |          |         |                |
    (= o_1 (= y 0))        |         |          |         |                |
                           |         |          |         |                |
    init                   |         |          |         |                |
    (= y 0)                |         |          |         |                |
                           |         |          |         |                |
    inputsymbols           |         |          |         |                |
    (>= x (^ c 2))         |         |          |         |                |
                           |         |          |         |                |
    outputsymbols          |         |          |         |                |
    (> y (* x 2))          |         |          |         |                |


