/*
 * Copyright 2022 Niklas Krafczyk, Jan Peleska
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#include "creators/input/SfsmFromCsvCreator.hpp"

#include <memory>
#include <string>
#include <fstream>
#include <algorithm>
#include <stdexcept>
#include <vector>
#include <map>
#include <cassert>
#include <iterator>
#include <sstream>
#include <optional>

#include "utils/CsvUtils.hpp"
#include "sfsm/Sfsm.hpp"
#include "sfsm/Transition.hpp"
#include "sfsm/Expressions.hpp"
#include "creators/input/SfsmFromFileCreator.hpp"

using namespace std;

namespace libfsmtest::SFSMs {

unique_ptr<Sfsm> SfsmFromCsvCreator::createSfsm() {
    readFile();

    return make_unique<SFSMs::Sfsm>(std::move(states), 0, inputAlphabet, outputAlphabet, additionalInputs, additionalOutputs, invariant, initialCondition, identifierDefinitions);
}

void SfsmFromCsvCreator::readFile() {
    ifstream inputFile(fileName);
    if (!inputFile.is_open()) throw runtime_error("Failed to open sfsm input file");

    auto trim = [](auto &string) {
        std::string const whitespaces = " \t";
        auto firstChar = string.find_first_not_of(whitespaces);
        auto lastChar = string.find_last_not_of(whitespaces);
        if(firstChar != std::string::npos) {
            string = string.substr(firstChar, lastChar - firstChar + 1);
        } else {
            string = "";
        }
    };
    Csv csv = readCsv(inputFile, ';', nullptr, trim);

    auto columnEmpty = [](auto const &column) {
        return std::all_of(column.begin(), column.end(), [](auto const &cell) {
            return cell.empty();
        });
    };

    // Sanity checks
    if(csv.size() <= 2
        or columnEmpty(csv.at(0)) or columnEmpty(csv.at(1))) {
        // empty
        throw std::runtime_error("No transition table found in file " + fileName);
    }
    auto const firstEmptyColumn = std::find_if(csv.begin(), csv.end(), columnEmpty);
    auto const end = firstEmptyColumn; //End of table iterator

    vector<string> stateNames;
    vector<string> inputAlphabetStrings;
    vector<string> outputAlphabetStrings;
    vector<string> additionalInputStrings;
    vector<string> additionalOutputStrings;
    vector<string> invariantStrings;
    vector<string> initialConditionStrings;
    vector<string> identifierStrings;

    // Find the first empty cell after the transition table, which marks the
    // beginning of the additional sections introducing identifier definitions
    // and invariants
    // TODO: Should also allow whitespace cells.
    auto endOfTransitionTable = std::find(csv.at(0).cbegin() + 1, csv.at(0).cend(), "");
    // Copy state names from the first column of the CSV file, skipping the topmost (empty) cell.
    stateNames.assign(csv.at(0).cbegin() + 1, endOfTransitionTable);

    // Search the 'invariants' keyword in the first column
    auto beginOfInvariants = std::find(endOfTransitionTable, csv.at(0).cend(), std::string("invariants"));
    // Search the 'init' keyword in the first column
    auto beginOfInitialConditions = std::find(endOfTransitionTable, csv.at(0).cend(), std::string("init"));
    // Search the 'identifiers' keyword in the first column
    auto beginOfIdentifiers = std::find(endOfTransitionTable, csv.at(0).cend(), std::string("identifiers"));
    // Search the 'inputsymbols' keyword in the first column
    auto beginOfInputSymbols = std::find(endOfTransitionTable, csv.at(0).cend(), std::string("inputsymbols"));
    // Search the 'outputsymbols' keyword in the first column
    auto beginOfOutputSymbols = std::find(endOfTransitionTable, csv.at(0).cend(), std::string("outputsymbols"));

    std::vector<decltype(beginOfInvariants)> iterators = {beginOfInvariants, beginOfInitialConditions, beginOfIdentifiers, beginOfInputSymbols, beginOfOutputSymbols, csv.at(0).cend()};
    std::sort(iterators.begin(), iterators.end());
    std::map<decltype(beginOfInvariants), decltype(beginOfInvariants)> intervals;
    for (size_t i = 0; i < iterators.size() - 1; ++i) {
        intervals.emplace(iterators.at(i), iterators.at(i+1));
    }

    if (beginOfInvariants != csv.at(0).cend()) {
        // Determine the end of the invariants section, which is either the end
        // of the column or the beginning of one of the other sections, which
        // ever comes first
        auto const endOfInvariants = intervals.at(beginOfInvariants);
        // Copy all invariant strings to 'invariantStrings' except for empty cells
        std::remove_copy(beginOfInvariants + 1, endOfInvariants, std::back_inserter(invariantStrings), "");
    } else {
        invariantStrings.emplace_back("true");
    }
    if (beginOfInitialConditions != csv.at(0).cend()) {
        // Determine the end of the initial conditions section, which is either the end
        // of the column or the beginning of one of the other sections, which
        // ever comes first
        auto const endOfInitialConditions = intervals.at(beginOfInitialConditions);
        // Copy all initial condition strings to 'initialConditionStrings' except for empty cells
        std::remove_copy(beginOfInitialConditions + 1, endOfInitialConditions, std::back_inserter(initialConditionStrings), "");
    } else {
        throw std::runtime_error(std::string("Cannot read SFSM from file '") + fileName + "': No initial conditions found.");
    }
    if (beginOfIdentifiers != csv.at(0).cend()) {
        // Determine the end of the identifiers section, which is either the end
        // of the column or the beginning of one of the other sections, which
        // ever comes first
        auto const endOfIdentifiers = intervals.at(beginOfIdentifiers);
        // Copy all identifier strings to 'identifierStrings' except for empty cells
        std::remove_copy(beginOfIdentifiers + 1, endOfIdentifiers, std::back_inserter(identifierStrings), std::string(""));
    } else {
        throw std::runtime_error(std::string("Cannot read SFSM from file '") + fileName + "': No identifier definitions found.");
    }
    if (beginOfInputSymbols != csv.at(0).cend()) {
        // Determine the end of the identifiers section, which is either the end
        // of the column or the beginning of one of the other sections, which
        // ever comes first
        auto const endOfInputSymbols = intervals.at(beginOfInputSymbols);
        // Copy all identifier strings to 'inputAlphabetStrings' except for empty cells
        std::remove_copy(beginOfInputSymbols + 1, endOfInputSymbols, std::back_inserter(additionalInputStrings), std::string(""));
    }
    if (beginOfOutputSymbols != csv.at(0).cend()) {
        // Determine the end of the identifiers section, which is either the end
        // of the column or the beginning of one of the other sections, which
        // ever comes first
        auto const endOfOutputSymbols = intervals.at(beginOfOutputSymbols);
        // Copy all identifier strings to 'outputAlphabetStrings' except for empty cells
        std::remove_copy(beginOfOutputSymbols + 1, endOfOutputSymbols, std::back_inserter(additionalOutputStrings), std::string(""));
    }
    // Process the invariant strings and generate the invariant expression
    this->invariant = SfsmFromFileCreator::processInvariants(fileName, invariantStrings, static_cast<size_t>(beginOfInvariants - csv.at(0).begin()));
    // Process the initialCondition strings and generate the initialCondition expression
    this->initialCondition = SfsmFromFileCreator::processInvariants(fileName, initialConditionStrings, static_cast<size_t>(beginOfInitialConditions - csv.at(0).begin()));
    // Process the identifier definition strings and generate the identifier definitions
    this->identifierDefinitions = SfsmFromFileCreator::processIdentifierDefinitions(fileName, identifierStrings, static_cast<size_t>(beginOfIdentifiers - csv.at(0).begin()));

    // Create input alphabet from first row
    for (auto it = begin(csv) + 1; it != end; ++it) {
        inputAlphabetStrings.emplace_back(it->at(0));
    }

    states.reserve(stateNames.size());
    Index id = 0;
    for(auto const &name : stateNames) {
        states.emplace_back(name, id++);
    }

    for (auto it = begin(csv) + 1; it != end; ++it) {
        auto const columnIndex = distance(csv.begin() + 1, it);
        assert(columnIndex >= 0);
        Index inputIndex;
        auto const input = it->at(0);
        if (auto inputIt = std::find(inputAlphabetStrings.begin(), inputAlphabetStrings.end(), input);
                inputIt == inputAlphabetStrings.end()) {
            inputIndex = inputAlphabetStrings.size();
            inputAlphabetStrings.emplace_back(input);
        } else {
            auto inputItIndex = std::distance(inputAlphabetStrings.begin(), inputIt);
            assert(inputItIndex >= 0);
            inputIndex = static_cast<Index>(inputItIndex);
        }

        for (Index stateIndex = 0; stateIndex < stateNames.size(); ++stateIndex) {
            Index const lineIndex = stateIndex + 1;
            Transitions transitions = states.at(stateIndex).getTransitions();
            // Cell defining transition from State stateIndex for current input
            istringstream cell(it->at(lineIndex));

            // There might be multiple transitions specified for one input, separated by comma
            string singleTransition;
            while (getline(cell, singleTransition, ',')) {
                istringstream sTstream(singleTransition);
                string postState;
                string output;
                getline(sTstream, postState, '/');
                getline(sTstream, output);

                if (postState.empty()) {
                    continue;
                }

                Index outputIndex;
                if (auto outputIt = std::find(outputAlphabetStrings.begin(), outputAlphabetStrings.end(), output);
                        outputIt == outputAlphabetStrings.end()) {
                    outputIndex = outputAlphabetStrings.size();
                    outputAlphabetStrings.emplace_back(output);
                } else {
                    auto difference2 = std::distance(outputAlphabetStrings.begin(), outputIt);
                    assert(difference2 >= 0);
                    outputIndex = static_cast<Index>(difference2);
                }

                auto stateIt = std::find(stateNames.begin(), stateNames.end(), postState);
                if (stateIt == stateNames.end()) {
                    throw std::runtime_error(std::string("Post state is not defined in line ") + std::to_string(lineIndex) + ", column " + std::to_string(inputIndex) + ": '" + postState + "'");
                }
                auto difference3 = std::distance(stateNames.begin(), stateIt);
                assert(difference3 >= 0);
                auto postStateIndex = static_cast<Index>(difference3);

                transitions.emplace_back(stateIndex, inputIndex, outputIndex, postStateIndex);
            }
            states.at(stateIndex) = states.at(stateIndex).modify(std::nullopt, std::nullopt, transitions);
        }
    }
    for(auto const &inputStr : inputAlphabetStrings) {
        try {
            inputAlphabet.push_back(parseToExpressions(inputStr).front());
        } catch(std::runtime_error &e) {
            std::string error("Error while parsing guard condition in file '");
            error += fileName;
            error += ": ";
            error += e.what();
            throw std::runtime_error(error);
        }
    }
    for(auto const &outputStr : outputAlphabetStrings) {
        try {
            outputAlphabet.push_back(parseToExpressions(outputStr).front());
        } catch(std::runtime_error &e) {
            std::string error("Error while parsing output expression in file '");
            error += fileName;
            error += ": ";
            error += e.what();
            throw std::runtime_error(error);
        }
    }
    for(auto const &inputStr : additionalInputStrings) {
        try {
            additionalInputs.push_back(parseToExpressions(inputStr).front());
        } catch(std::runtime_error &e) {
            std::string error("Error while parsing guard condition in file '");
            error += fileName;
            error += ": ";
            error += e.what();
            throw std::runtime_error(error);
        }
    }
    for(auto const &outputStr : additionalOutputStrings) {
        try {
            additionalOutputs.push_back(parseToExpressions(outputStr).front());
        } catch(std::runtime_error &e) {
            std::string error("Error while parsing output expression in file '");
            error += fileName;
            error += ": ";
            error += e.what();
            throw std::runtime_error(error);
        }
    }
}

} // namespace libfsmtest
