/*
 * Copyright 2022 Niklas Krafczyk, Jan Peleska
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#pragma once

#include <string>
#include <vector>

#include "creators/SfsmCreator.hpp"
#include "sfsm/State.hpp"
#include "sfsm/Sfsm.hpp"
#include "sfsm/Expressions.hpp"
#include "sfsm/ExpressionDefinitions.hpp"

namespace libfsmtest::SFSMs {

/**
 * Factory for reading an SFSM specification from a CSV file.
 */
class SfsmFromCsvCreator : public libfsmtest::SFSMs::SfsmCreator {
private:
    /** The name or path of the CSV file to be used */
    const std::string fileName;

    /** The states of the read FSM */
    libfsmtest::SFSMs::States states;

    /** The guard conditions of the SFSM */
    std::vector<libfsmtest::SFSMs::Sfsm::GuardCondition_t> inputAlphabet;

    /** The output expressions of the SFSM */
    std::vector<libfsmtest::SFSMs::Sfsm::OutputExpression_t> outputAlphabet;

    /** The additional guard conditions of the SFSM */
    std::vector<libfsmtest::SFSMs::Sfsm::GuardCondition_t> additionalInputs;

    /** The additional output expressions of the SFSM */
    std::vector<libfsmtest::SFSMs::Sfsm::OutputExpression_t> additionalOutputs;

    /** The invariant of the SFSM */
    Expression invariant;

    /** The initial condition of the SFSM */
    Expression initialCondition;

    /** The identifier definitions of the SFSM */
    IdentifierDefinitions identifierDefinitions;

    /**
     * Reads and parses the CSV file content of #fileName.
     * Fills #states and #presentationLayer.
     */
    void readFile();

public:
    /**
     * Constructor.
     * Won't start reading or parsing the input.
     *
     * @param fileNameParam The path of the CSV file to create a SFSM from.
     */
    explicit SfsmFromCsvCreator(std::string fileNameParam) : fileName(std::move(fileNameParam)) {}

    /**
     * Creates an SFSM from the given input CSV file.
     * @see readFile()
     * @return A unique pointer to the constructed SFSM.
     */
    std::unique_ptr<Sfsm> createSfsm() override;

};

} // namespace libfsmtest
