/*
 * Copyright 2022 Niklas Krafczyk, Jan Peleska
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#include "creators/input/SfsmFromFileCreator.hpp"

#include <string>
#include <memory>
#include <stdexcept>
#include <filesystem>
#include <fstream>
#include <sstream>
#include <algorithm>
#include <iterator>

#include "SfsmFromRawCreator.hpp"
#include "SfsmFromCsvCreator.hpp"
#include "sfsm/definitions.hpp"
#include "sfsm/ExpressionDefinitions.hpp"
#include "sfsm/Expressions.hpp"


using namespace std;

namespace libfsmtest::SFSMs {

SfsmFromFileCreator::SfsmFromFileCreator(const std::string &fileNameParam) : fileName(fileNameParam), prefix(fileNameParam) {}

std::unique_ptr<Sfsm> SfsmFromFileCreator::createSfsm() {
    // Try to recognise the format of the file name passed to the constructor.
    FileFormat_t recognisedFormat = recogniseFormat();

    std::unique_ptr<Sfsm> sfsm;
    if (recognisedFormat == RAW) {
        // Check whether there are additional files to specify a presentation layer
        if(checkForStateNameFile(prefix)) {
            // File for the state names found, so use them
            sfsm = libfsmtest::SFSMs::createSfsm<SfsmFromRawCreator>(fileName, prefix + ".state", prefix + ".in", prefix + ".out", prefix + ".ain", prefix + ".aout", prefix + ".inv", prefix + ".init", prefix + ".id");
        } else {
            // Use only sfsm specification file and the input and output definitions.
            sfsm = libfsmtest::SFSMs::createSfsm<SfsmFromRawCreator>(fileName, prefix + ".in", prefix + ".out", prefix + ".ain", prefix + ".aout", prefix + ".inv", prefix + ".init", prefix + ".id");
        }
    } else if (recognisedFormat == CSV) {
            sfsm = libfsmtest::SFSMs::createSfsm<SfsmFromCsvCreator>(fileName);
    } else {
        throw std::runtime_error("This file format is not supported.");
    }

    return sfsm;
}

FileFormat_t SfsmFromFileCreator::recogniseFormat() {
    FileFormat_t format;

    if (filesystem::exists(fileName + ".sfsm")) {
        fileName.append(".sfsm");
        format = RAW;
    } else if (filesystem::exists(fileName + ".csv")){
        fileName.append(".csv");
        format = CSV;
    } else if (filesystem::exists(fileName)) {
        // There is a file without an extension for the given filename, so try to identify its format.
        format = recogniseByExtension();
        if (format == NOT_RECOGNISED) format = recogniseByContent();
        if (format == NOT_RECOGNISED) format = NOT_SUPPORTED;
    } else {
        throw std::runtime_error(std::string("Could not find file '") + fileName + "'");
    }

    return format;
}

FileFormat_t SfsmFromFileCreator::recogniseByExtension() {
    // get the position of the last '.' which would be directly before an extension
    auto const lastDot = fileName.find_last_of('.');
    if (lastDot != string::npos) {
        // '.' found, extract substring after that (might be an extension)
        string extension = fileName.substr(lastDot + 1);

        if (extension == "sfsm") {
            prefix = fileName.substr(0, lastDot);
            return RAW;
        } else if (extension == "csv") {
            prefix = fileName.substr(0, lastDot);
            return CSV;
        }
    }
    // no supported extension found.
    return NOT_RECOGNISED;
}

FileFormat_t SfsmFromFileCreator::recogniseByContent() {
    std::ifstream file(fileName);
    if (!file.is_open()) {
        throw std::runtime_error("Failed to open the specified file.");
    }
    std::string line;
    getline(file, line);

    if (recogniseCSV(line)) return CSV;
    if (recogniseRAW(line)) return RAW;

    return NOT_RECOGNISED;
}

bool SfsmFromFileCreator::recogniseCSV(const std::string& firstLine) {
    std::stringstream line(firstLine);
    std::string cell;

    getline(line, cell, ';');
    cell.erase(std::remove_if(cell.begin(), cell.end(), ::isspace), cell.end());
    if (!cell.empty()) {
        // The first cell of a CSV sfsm file should be empty.
        return false;
    }

    while(getline(line, cell, ';')) {
        cell.erase(remove_if(cell.begin(), cell.end(), ::isspace), cell.end());
        if (cell.empty()) {
            // all following cells should not be empty.
            return false;
        }
    }

    // The first line matches the CSV FSM file format.
    return true;
}

bool SfsmFromFileCreator::recogniseRAW(const std::string& firstLine) {
    std::stringstream line(firstLine);
    Index value;
    for (unsigned short i = 1; i <= 4; ++i) {
        line >> value;
        if (line.fail()) {
            // Could not read one an expected value.
            return false;
        }
    }

    // The first line matches the raw FSM file format.
    return true;
}

bool SfsmFromFileCreator::checkForStateNameFile(const std::string& filename) {
    return filesystem::exists(filename + ".state");
}

IdentifierDefinitions SfsmFromFileCreator::processIdentifierDefinitions(std::string const &filename, Strings const &identifierStrings, size_t lineOffset) {
    IdentifierDefinitions identifierDefinitions;
    std::transform(identifierStrings.cbegin(), identifierStrings.cend(), std::back_inserter(identifierDefinitions), [&identifierStrings,&filename,lineOffset](auto const &identifierString) {
        std::string exception;
        try {
            auto spacePosition = identifierString.find_last_of(' ');
            if (spacePosition != std::string::npos) {
                auto type = parseType(identifierString.substr(0, spacePosition));
                auto identifier = identifierString.substr(spacePosition + 1);
                if (identifier.find(' ') == std::string::npos) {
                    return IdentifierDefinition(identifier, type);
                }
            }
            exception = std::string("Invalid identifier definition: '") + identifierString + "'";
        } catch (std::runtime_error &e) {
            exception = std::string("Invalid identifier definition: '") + identifierString + "': " + e.what();
        }
        auto iter = std::find(identifierStrings.cbegin(), identifierStrings.cend(), identifierString);
        auto index = static_cast<size_t>(std::distance(identifierStrings.cbegin(), iter));
        std::string error("Error while parsing identifier in file '");
        error += filename;
        error += "', line ";
        error += std::to_string(index + lineOffset + 1);
        error += ": ";
        error += exception;
        throw std::runtime_error(error);
    });
    return identifierDefinitions;
}

Expression SfsmFromFileCreator::processInvariants(std::string const &filename, Strings const &invariantStrings, size_t lineOffset) {
    Expressions invariants;
    std::transform(invariantStrings.cbegin(), invariantStrings.cend(), std::back_inserter(invariants), [&invariantStrings,&filename,lineOffset](auto const &string) {
        try {
            return parseToExpressions(string).front();
        } catch(std::runtime_error &e) {
            auto iter = std::find(invariantStrings.cbegin(), invariantStrings.cend(), string);
            auto index = static_cast<size_t>(std::distance(invariantStrings.cbegin(), iter));
            std::string error("Error while parsing global constraint expression in file '");
            error += filename;
            error += "', line ";
            error += std::to_string(index + lineOffset + 1);
            error += ": ";
            error += e.what();
            throw std::runtime_error(error);
        }
    });
    return libfsmtest::SFSMs::makeLogicalConjunction(invariants);
}

} // namespace libfsmtest

