/*
 * Copyright 2022 Niklas Krafczyk, Jan Peleska
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#include <memory>
#include <stdexcept>
#include <fstream>
#include <vector>
#include <string>
#include <set>
#include <algorithm>
#include <iterator>

#include "sfsm/Sfsm.hpp"
#include "sfsm/State.hpp"
#include "sfsm/Transition.hpp"
#include "utils/CsvUtils.hpp"
#include "sfsm/definitions.hpp"
#include "sfsm/Expressions.hpp"

#include "SfsmFromFileCreator.hpp"
#include "SfsmFromRawCreator.hpp"

using namespace std;

namespace libfsmtest::SFSMs {

unique_ptr<SFSMs::Sfsm> SfsmFromRawCreator::createSfsm() {
    readFiles();
    
    if(not initialState.has_value()) {
        throw std::runtime_error("ERROR: SFSM does not have an initial state.");
    }
    return make_unique<SFSMs::Sfsm>(std::move(states), initialState.value(), inputAlphabet, outputAlphabet, additionalInputs, additionalOutputs, invariant, initialCondition, identifierDefinitions);
}

void SfsmFromRawCreator::readFiles() {
    std::ifstream sfsm(sfsmFileName);
    if (!sfsm.is_open()) {
        throw std::runtime_error("Failed to open sfsm input file");
    }

    auto removeTrailingWhitespace = [](std::string& s) {
        auto it = s.end();
        while (it != s.begin() && ::isspace(*--it));
        s.erase(it + 1, s.end());
    };
    Csv sfsmData = readCsv(sfsm, ' ', removeTrailingWhitespace, nullptr);

    std::vector<std::string> stateNames, inputAlphabetStrings, outputAlphabetStrings, additionalInputStrings, additionalOutputStrings, invariantStrings, initialConditionStrings, identifierStrings;

    if (statesFileName.empty()) {
        /**
         * second constructor was used, there are no state names.
         * therefore use only internal format of indices.
         */
        typedef struct Comparator {
            bool operator()(const std::string& lhs, const std::string& rhs) const {
                return stoul(lhs) < stoul(rhs);
            }
        } Comparator_t;

        // create set with preStates from sfsmData
        std::set<string> stateNamesSet(sfsmData.at(0).begin(), sfsmData.at(0).end());
        // insert possibly different postStates from sfsmData
        stateNamesSet.insert(sfsmData.at(3).begin(), sfsmData.at(3).end());
        // assign vector with data from set
        stateNames.assign(stateNamesSet.begin(), stateNamesSet.end());
        // Sort the vector data
        std::sort(stateNames.begin(), stateNames.end(), Comparator_t());

    } else {
        stateNames = readFile(statesFileName);
    }

    inputAlphabetStrings = readFile(inputFileName);
    outputAlphabetStrings = readFile(outputFileName);
    additionalInputStrings = readFile(additionalInputsFileName);
    additionalOutputStrings = readFile(additionalOutputsFileName);
    invariantStrings = readFile(invariantsFileName);
    initialConditionStrings = readFile(initialConditionsFileName);
    identifierStrings = readFile(identifierFileName);

    states.reserve(stateNames.size());
    Index idx = 0;
    for(auto const &stateName : stateNames) {
        states.emplace_back(stateName, idx++);
    }


    std::vector<Transitions> transitions(stateNames.size());
    for (size_t row = 0; row < sfsmData.at(0).size(); ++row) {
        Index preState, input, output, postState;
        try {
            preState = stoul(sfsmData.at(0).at(row));
        } catch (std::invalid_argument& invalid_argument) {
            throw std::runtime_error("ERROR: At line " + std::to_string(row) + ": prestate ID (" + sfsmData.at(0).at(row) + "is invalid: " + invalid_argument.what());
        } catch (std::out_of_range& outOfRange) {
            throw std::runtime_error("ERROR: At line " + std::to_string(row) + ": prestate ID (" + sfsmData.at(0).at(row) + "is invalid: " + outOfRange.what());
        }

        try {
            input = stoul(sfsmData.at(1).at(row));
        } catch (std::invalid_argument& invalid_argument) {
            throw runtime_error("ERROR: At line " + to_string(row) + ": input ID (" + sfsmData.at(1).at(row) + "is invalid: " + invalid_argument.what());
        } catch (std::out_of_range& outOfRange) {
            throw runtime_error("ERROR: At line " + to_string(row) + ": input ID (" + sfsmData.at(1).at(row) + "is invalid: " + outOfRange.what());
        }

        try {
            output = stoul(sfsmData.at(2).at(row));
        } catch (std::invalid_argument& invalid_argument) {
            throw std::runtime_error("ERROR: At line " + std::to_string(row) + ": output ID (" + sfsmData.at(2).at(row) + "is invalid: " + invalid_argument.what());
        } catch (std::out_of_range& outOfRange) {
            throw std::runtime_error("ERROR: At line " + std::to_string(row) + ": output ID (" + sfsmData.at(2).at(row) + "is invalid: " + outOfRange.what());
        }

        try {
            postState = stoul(sfsmData.at(3).at(row));
        } catch (std::invalid_argument& invalid_argument) {
            throw std::runtime_error("ERROR: At line " + std::to_string(row) + ": poststate ID (" + sfsmData.at(3).at(row) + "is invalid: " + invalid_argument.what());
        } catch (std::out_of_range& outOfRange) {
            throw std::runtime_error("ERROR: At line " + std::to_string(row) + ": poststate ID (" + sfsmData.at(3).at(row) + "is invalid: " + outOfRange.what());
        }

        if (not this->initialState.has_value()) {
            this->initialState = preState;
        }

        transitions.at(preState).emplace_back(states[preState].getId(), input, output, states[postState].getId());
    }
    for(size_t preStateIdx = 0; preStateIdx < stateNames.size(); ++preStateIdx) {
        states.at(preStateIdx) = states.at(preStateIdx).modify(std::nullopt, std::nullopt, transitions.at(preStateIdx));
    }


    std::transform(inputAlphabetStrings.cbegin(), inputAlphabetStrings.cend(), std::back_inserter(inputAlphabet), [&inputAlphabetStrings,&inputFileName=this->inputFileName](auto const &string) {
        try {
            return parseToExpressions(string).front();
        } catch(std::runtime_error &e) {
            auto iter = std::find(inputAlphabetStrings.cbegin(), inputAlphabetStrings.cend(), string);
            auto index = std::distance(inputAlphabetStrings.cbegin(), iter);
            std::string error("Error while parsing guard condition in file '");
            error += inputFileName;
            error += "', line ";
            error += to_string(index + 1);
            error += ": ";
            error += e.what();
            throw std::runtime_error(error);
        }
    });
    std::transform(outputAlphabetStrings.cbegin(), outputAlphabetStrings.cend(), std::back_inserter(outputAlphabet), [&outputAlphabetStrings,&outputFileName=this->outputFileName](auto const &string) {
        try {
            return parseToExpressions(string).front();
        } catch(std::runtime_error &e) {
            auto iter = std::find(outputAlphabetStrings.cbegin(), outputAlphabetStrings.cend(), string);
            auto index = std::distance(outputAlphabetStrings.cbegin(), iter);
            std::string error("Error while parsing output expression in file '");
            error += outputFileName;
            error += "', line ";
            error += to_string(index + 1);
            error += ": ";
            error += e.what();
            throw std::runtime_error(error);
        }
    });
    std::transform(additionalInputStrings.cbegin(), additionalInputStrings.cend(), std::back_inserter(additionalInputs), [&additionalInputStrings,&inputFileName=this->additionalInputsFileName](auto const &string) {
        try {
            return parseToExpressions(string).front();
        } catch(std::runtime_error &e) {
            auto iter = std::find(additionalInputStrings.cbegin(), additionalInputStrings.cend(), string);
            auto index = std::distance(additionalInputStrings.cbegin(), iter);
            std::string error("Error while parsing additional guard condition in file '");
            error += inputFileName;
            error += "', line ";
            error += to_string(index + 1);
            error += ": ";
            error += e.what();
            throw std::runtime_error(error);
        }
    });
    std::transform(additionalOutputStrings.cbegin(), additionalOutputStrings.cend(), std::back_inserter(additionalOutputs), [&additionalOutputStrings,&outputFileName=this->additionalOutputsFileName](auto const &string) {
        try {
            return parseToExpressions(string).front();
        } catch(std::runtime_error &e) {
            auto iter = std::find(additionalOutputStrings.cbegin(), additionalOutputStrings.cend(), string);
            auto index = std::distance(additionalOutputStrings.cbegin(), iter);
            std::string error("Error while parsing output expression in file '");
            error += outputFileName;
            error += "', line ";
            error += to_string(index + 1);
            error += ": ";
            error += e.what();
            throw std::runtime_error(error);
        }
    });
    this->identifierDefinitions = SfsmFromFileCreator::processIdentifierDefinitions(identifierFileName, identifierStrings, 0);
    this->invariant = SfsmFromFileCreator::processInvariants(invariantsFileName, invariantStrings, 0);
    this->initialCondition = SfsmFromFileCreator::processInvariants(initialConditionsFileName, initialConditionStrings, 0);
}

std::vector<string> SfsmFromRawCreator::readFile(const std::string &fileName) {
    std::vector<string> v;

    std::ifstream input(fileName);
    std::string line;

    if (!input.is_open()) throw std::runtime_error("Failed to open input file");

    while (getline(input, line)) {
        v.push_back(line);
    }

    return v;
}

}
