/*
 * Copyright 2022 Niklas Krafczyk, Jan Peleska
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#pragma once

#include <string>
#include <optional>
#include <vector>
#include <memory>

#include "creators/SfsmCreator.hpp"
#include "sfsm/State.hpp"
#include "sfsm/Sfsm.hpp"
#include "sfsm/Expressions.hpp"
#include "sfsm/ExpressionDefinitions.hpp"

namespace libfsmtest::SFSMs {

/**
 * Factory for reading an SFSM specification from the low-level file format.
 */
class SfsmFromRawCreator : public libfsmtest::SFSMs::SfsmCreator {
private:
    /**
     * Paths to the files of the specification.
     */
    std::string const sfsmFileName;
    std::string const statesFileName;
    std::string const inputFileName;
    std::string const outputFileName;
    std::string const additionalInputsFileName;
    std::string const additionalOutputsFileName;
    std::string const invariantsFileName;
    std::string const initialConditionsFileName;
    std::string const identifierFileName;

    /** The states of the SFSM */
    libfsmtest::SFSMs::States states;

    /** The guard conditions of the SFSM */
    std::vector<libfsmtest::SFSMs::Sfsm::GuardCondition_t> inputAlphabet;

    /** The output expressions of the SFSM */
    std::vector<libfsmtest::SFSMs::Sfsm::OutputExpression_t> outputAlphabet;

    /** The additional guard conditions of the SFSM */
    std::vector<libfsmtest::SFSMs::Sfsm::GuardCondition_t> additionalInputs;

    /** The additional output expressions of the SFSM */
    std::vector<libfsmtest::SFSMs::Sfsm::OutputExpression_t> additionalOutputs;

    /** The global constraint */
    libfsmtest::SFSMs::Expression invariant;

    /** The initial constraint */
    libfsmtest::SFSMs::Expression initialCondition;

    /** The definitions of the identifiers in the guard conditions and output
     *  expressions of the SFSM */
    libfsmtest::SFSMs::IdentifierDefinitions identifierDefinitions;

    /** The index of the initial state of the SFSM */
    std::optional<Index> initialState;

    /**
     * Reads and parses the low-level file format.
     * Fills #states, #inputAlphabet and #outputAlphabet.
     */
    void readFiles();

    /**
     * Reads the content of the given file into a vector of lines.
     * @param fileName The file to be read.
     * @return A vector containing the lines of the file.
     */
    static std::vector<std::string> readFile(const std::string& fileName);

public:
    /**
     * Constructor.
     * Won't start reading or parsing the input.
     *
     * @param sfsmFileNameParam The file containing the transitions.
     * @param statesFileNameParam The file containing the state names.
     * @param inputFileNameParam The file containing the input alphabet.
     * @param outputFileNameParam The file containing the output alphabet.
     * @param additionalInputsFileNameParam The file containing the additional inputs.
     * @param additionalOutputsFileNameParam The file containing the additional outputs.
     * @param invariantsFileNameParam The file containing the global constraint.
     * @param initialConditionsFileNameParam The file containing the global constraint.
     * @param identifierFileNameParam The file containing the identifier definitions.
     */
    SfsmFromRawCreator(std::string  sfsmFileNameParam,
                       std::string  statesFileNameParam,
                       std::string  inputFileNameParam,
                       std::string  outputFileNameParam,
                       std::string  additionalInputsFileNameParam,
                       std::string  additionalOutputsFileNameParam,
                       std::string  invariantsFileNameParam,
                       std::string  initialConditionsFileNameParam,
                       std::string  identifierFileNameParam)
                       :
    sfsmFileName(std::move(sfsmFileNameParam)),
    statesFileName(std::move(statesFileNameParam)),
    inputFileName(std::move(inputFileNameParam)),
    outputFileName(std::move(outputFileNameParam)),
    additionalInputsFileName(std::move(additionalInputsFileNameParam)),
    additionalOutputsFileName(std::move(additionalOutputsFileNameParam)),
    invariantsFileName(std::move(invariantsFileNameParam)),
    initialConditionsFileName(std::move(initialConditionsFileNameParam)),
    identifierFileName(std::move(identifierFileNameParam))  {}

    /**
     * Constructor.
     * Won't start reading or parsing the input.
     * Fills the state names with IDs from the specification in low-level
     * format.
     *
     * @param sfsmFileNameParam The file containing the transitions.
     * @param inputFileNameParam The file containing the input alphabet.
     * @param outputFileNameParam The file containing the output alphabet.
     * @param invariantsFileNameParam The file containing the global constraint.
     * @param initialConditionsFileNameParam The file containing the global constraint.
     * @param identifierFileNameParam The file containing the identifier definitions.
     */
    SfsmFromRawCreator(std::string  sfsmFileNameParam,
                       std::string  inputFileNameParam,
                       std::string  outputFileNameParam,
                       std::string  additionalInputsFileNameParam,
                       std::string  additionalOutputsFileNameParam,
                       std::string  invariantsFileNameParam,
                       std::string  initialConditionsFileNameParam,
                       std::string  identifierFileNameParam)
                       :
    sfsmFileName(std::move(sfsmFileNameParam)),
    inputFileName(std::move(inputFileNameParam)),
    outputFileName(std::move(outputFileNameParam)),
    additionalInputsFileName(std::move(additionalInputsFileNameParam)),
    additionalOutputsFileName(std::move(additionalOutputsFileNameParam)),
    invariantsFileName(std::move(invariantsFileNameParam)),
    initialConditionsFileName(std::move(initialConditionsFileNameParam)),
    identifierFileName(std::move(identifierFileNameParam)) {}

    /**
     * Creates an SFSM from the given input files. If the SFSM is
     * deterministic, a pointer to a Dfsm-instance will be returned.
     * @see readFiles()
     * @return A unique pointer to the constructed SFSM.
     */
    std::unique_ptr<Sfsm> createSfsm() override;
};

} // namespace libfsmtest



