/*
 * Copyright 2022 Niklas Krafczyk, Jan Peleska
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#pragma once

#include <cassert>

#include "creators/transformers/SfsmTransformer.hpp"
#include "utils/algorithms/FindAllSatisfiableConjunctions.hpp"
#include "sfsm/State.hpp"

namespace libfsmtest::SFSMs {

/**
 * Abstract class for all sorts of visitors transforming machines
 * into new machines
 */
class AbstractionTransformer : public SfsmTransformer {
protected:
    Expressions const abstractionExpressions;
    Expressions equivalenceClasses;

public:

    AbstractionTransformer(Sfsm const *sfsmToTransform, Expressions const &paramAbstractionExpressions) :
    SfsmTransformer(sfsmToTransform), abstractionExpressions(paramAbstractionExpressions) {
        assert(not abstractionExpressions.empty());
        typedef std::pair<std::vector<ssize_t>, Expression> ConjunctionResult;
        std::vector<ConjunctionResult> const propositionEquivalenceClasses 
            = findAllSatisfiableConjunctions(sfsmToTransform->getIdentifierDefinitions(), sfsmToTransform->getInvariant(), abstractionExpressions);
        std::transform(propositionEquivalenceClasses.cbegin(), propositionEquivalenceClasses.cend(), std::back_inserter(equivalenceClasses), [](auto const &elem) {
            return std::get<1>(elem);
        });
    }
    
    ~AbstractionTransformer() override = default;

    void visit(Sfsm const& /* fsm */) override {
        this->transformedSfsm = this->sfsmReadOnly->clone();
        std::vector<State> &newStates = this->transformedSfsm->getStates();
        newStates.clear();
        Expressions newGuardConditions;
        Expressions newOutputExpressions;
        for(auto const &state : this->sfsmReadOnly->getStates()) {
            std::vector<Transition> newTransitions;
            for(auto const &transition : state.getTransitions()) {
                std::vector<size_t> satisfyingExpressions;
                Expression const &guardCondition = sfsmReadOnly->getUsedGuardConditions().at(transition.getInputId());
                Expression const &outputExpression = sfsmReadOnly->getUsedOutputExpressions().at(transition.getOutputId());
                for(size_t index = 0; index < equivalenceClasses.size(); ++index) {
                    auto const &equivalenceClass = equivalenceClasses.at(index);
                    auto expressionToCheck = makeLogicalConjunction(this->sfsmReadOnly->getInvariant(), guardCondition, outputExpression, equivalenceClass);
                    if(isSatisfiable(expressionToCheck, this->sfsmReadOnly->getIdentifierDefinitions()).has_value()) {
                        satisfyingExpressions.push_back(index);
                    }
                }
                for(auto const &expressionIndex : satisfyingExpressions) {
                    auto const &expression = equivalenceClasses.at(expressionIndex);
                    auto splitExpression = splitFunctionCall(expression).value_or(Expressions{expression});
                    if(splitExpression.size() > 1) {
                        Expressions operands(splitExpression.cbegin() + 1, splitExpression.cend());
                        splitExpression = operands;
                    }
                    Expressions guardConditions;
                    Expressions outputExpressions;
                    for(auto const &operand : splitExpression) {
                        if(containsIdentifiersWithLocation(this->sfsmReadOnly->getIdentifierDefinitions(), operand, IdentifierType::LocationInModel::OUTPUT)) {
                            outputExpressions.push_back(operand);
                        } else {
                            guardConditions.push_back(operand);
                        }
                    }
                    if(guardConditions.empty()) {
                        guardConditions.push_back(makeIdentifier("true"));
                    }
                    if(outputExpressions.empty()) {
                        outputExpressions.push_back(makeIdentifier("true"));
                    }
                    std::sort(guardConditions.begin(), guardConditions.end());
                    std::sort(outputExpressions.begin(), outputExpressions.end());
                    auto newGuardCondition = guardConditions.empty() ? guardConditions.front() : makeLogicalConjunction(guardConditions);
                    auto newOutputExpression = outputExpressions.empty() ? outputExpressions.front() : makeLogicalConjunction(outputExpressions);
                    size_t guardConditionIndex;
                    size_t outputExpressionIndex;
                    if(auto iter = std::find(newGuardConditions.cbegin(), newGuardConditions.cend(), newGuardCondition);
                        iter != newGuardConditions.cend()) {
                        guardConditionIndex = static_cast<size_t>(iter - newGuardConditions.cbegin());
                    } else {
                        guardConditionIndex = newGuardConditions.size();
                        newGuardConditions.push_back(newGuardCondition);
                    }
                    if(auto iter = std::find(newOutputExpressions.cbegin(), newOutputExpressions.cend(), newOutputExpression);
                        iter != newOutputExpressions.cend()) {
                        outputExpressionIndex = static_cast<size_t>(iter - newOutputExpressions.cbegin());
                    } else {
                        outputExpressionIndex = newOutputExpressions.size();
                        newOutputExpressions.push_back(newOutputExpression);
                    }
                    newTransitions.emplace_back(transition.getSourceId(), guardConditionIndex, outputExpressionIndex, transition.getTargetId());
                }
            }
            newStates.emplace_back(state.modify(std::nullopt, std::nullopt, newTransitions));
        }
        this->transformedSfsm = this->transformedSfsm->modify(std::nullopt,
                                                              std::nullopt,
                                                              newGuardConditions,
                                                              newOutputExpressions,
                                                              std::nullopt,
                                                              std::nullopt,
                                                              std::nullopt,
                                                              std::nullopt,
                                                              std::nullopt);
    }
};

} // namespace libfsmtest::SFSMs
