/*
 * Copyright 2022 Niklas Krafczyk, Jan Peleska
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#include <iostream>
#include <optional>
#include <tuple>
#include <istream>
#include <fstream>
#include <string>
#include <vector>
#include <cstddef>
#include <stdexcept>

#include "version.hpp"
#include "sfsm/Sfsm.hpp"
#include "creators/input/SfsmFromFileCreator.hpp"
#include "sfsm/Expressions.hpp"
#include "testgeneration/SfsmMethod.hpp"
#include "processing/checkers/TraceInclusionChecker.hpp"

using namespace libfsmtest;

static void printUsage(char const *name) {
    std::cout << name << " <Model file> [Executions file]" << std::endl;
}

static std::optional<
         std::tuple<
                    std::string,
                    std::ifstream*> //Returning stream as variant as ifstream cannot be converted to istream
       >
parseArguments(int argc, char *argv[]) {
    if(argc < 2 or argc > 3) {
        return std::nullopt;
    }

    std::string modelFile = argv[1];
    std::ifstream *inFile = nullptr;
    if(argc == 3) {
        inFile = new std::ifstream(argv[2]);
    }

    return std::make_tuple(modelFile, inFile);
}

static int checkTestExecution(libfsmtest::SFSMs::Sfsm const &sfsm, std::istream &executions) {
    size_t constexpr bufferSize = 512;
    char buffer[bufferSize];
    libfsmtest::SFSMs::SymbolicTrace testExecutionTrace;
    std::string currentLine;
    libfsmtest::SFSMs::TraceInclusionChecker checker(sfsm);

    while(not executions.eof()) {
        executions.getline(buffer, bufferSize);
        if(executions.fail() and executions.bad()) {
            throw std::runtime_error("Error when reading execution log.");
        } else {
            currentLine += buffer;
            if(not executions.fail()) {
                if(not currentLine.empty()) {
                    std::vector<libfsmtest::SFSMs::Expression> inputOutput = SFSMs::parseToExpressions(currentLine);
                    if(inputOutput.size() != 2) {
                        throw std::runtime_error(std::string("Could not parse the following line as an execution line:\n") + currentLine);
                    }
                    auto const &input = inputOutput.front();
                    auto const &output = inputOutput.back();
                    if(not checker(std::make_pair(input, output))) {
                        std::cout << '1' << std::endl;
                        return 1;
                    } else {
                        std::cout << '0' << std::endl;
                    }
                    currentLine.clear();
                }
            }
        }
    }

    return 0;
}

int main(int argc, char *argv[]) {
    if(auto arguments = parseArguments(argc, argv);
        arguments.has_value()) {
        auto const &[modelFile, executionFile] = arguments.value();
        std::istream *executionFileStreamPtr = &std::cin;
        if(executionFile != nullptr) {
            executionFileStreamPtr = executionFile;
        }
        std::unique_ptr<SFSMs::Sfsm> sfsm = SFSMs::createSfsm<SFSMs::SfsmFromFileCreator>(modelFile);

        return checkTestExecution(*sfsm, *executionFileStreamPtr);
    } else {
        printUsage(argv[0]);
        return -1;
    }
}
