/*
 * Copyright 2022 Niklas Krafczyk, Jan Peleska
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <string>
#include <memory>
#include <fstream>
#include <iostream>
#include <cstdlib>
#include <optional>
#include <tuple>
#include <charconv>

#include "version.hpp"
#include "sfsm/Sfsm.hpp"
#include "creators/input/SfsmFromFileCreator.hpp"
#include "sfsm/Expressions.hpp"
#include "creators/transformers/AbstractionTransformer.hpp"
#include "visitors/output/ToDotFileVisitor.hpp"
#include "testgeneration/SfsmMethod.hpp"

using namespace libfsmtest;

static void SfsmMethod(size_t additionalStates, std::string const &pathToCSV, std::string const &propositionsFile) {
    std::unique_ptr<SFSMs::Sfsm> sfsm = SFSMs::createSfsm<SFSMs::SfsmFromFileCreator>(pathToCSV);
    SFSMs::Expressions props;
    if(not propositionsFile.empty()) {
        std::string apData;
        std::ifstream apFile(propositionsFile);
        std::string line;
        while(std::getline(apFile, line)) {
            apData += line + "\n";
        }
        std::cout << apData << std::endl;
        props = SFSMs::parseToExpressions(apData);
    }
    std::unique_ptr<SFSMs::Sfsm> abstraction;
    if(not props.empty()) {
        abstraction = SFSMs::transformSfsm<SFSMs::AbstractionTransformer>(sfsm.get(), props);
    }
    if(auto modelToDot_envvar = getenv("LIBSFSMTEST_MODEL_DOT_OUTPUT");
        modelToDot_envvar != nullptr) {
        SFSMs::ToDotFileVisitor model(modelToDot_envvar);
        sfsm->accept(model);
        model.writeToFile();
    }
    if(auto abstractionToDot_envvar = getenv("LIBSFSMTEST_ABSTRACTION_DOT_OUTPUT");
        abstraction != nullptr and abstractionToDot_envvar != nullptr) {
        SFSMs::ToDotFileVisitor abstractionVisitor(abstractionToDot_envvar);
        abstraction->accept(abstractionVisitor);
        abstractionVisitor.writeToFile();
    }
    auto result = SFSMs::sfsmMethod(*sfsm, additionalStates, props);
    std::cout << "Test suite (size: " << result.size() << "):" << std::endl;
    for(auto const &trace : result) {
        std::cout << std::to_string(trace) << std::endl;
    }
}

static void printUsage(char const *name) {
    std::cout << name << " <Number of additional states> <Model file> [Atomic propositions file]" << std::endl;
}

static std::optional<
         std::tuple<size_t,
                    std::string,
                    std::string>
       >
parseArguments(int argc, char *argv[]) {
    if(argc < 3 or argc > 4) {
        return std::nullopt;
    }

    size_t additionalStates = 0;
    std::string additionalStatesString = argv[1];
    char const *first = additionalStatesString.c_str();
    char const *last = first + additionalStatesString.length();
    auto fromCharsResult = std::from_chars(first, last, additionalStates);
    if(fromCharsResult.ptr != last
        or fromCharsResult.ec == std::errc::invalid_argument
        or fromCharsResult.ec == std::errc::result_out_of_range) {
        return std::nullopt;
    }

    std::string modelFile = argv[2];
    std::string propFile;
    if(argc == 4) {
        propFile = argv[3];
    }

    return std::make_tuple(additionalStates, modelFile, propFile);
}

int main(int argc, char *argv[]) {
    if(auto arguments = parseArguments(argc, argv);
        arguments.has_value()) {
        auto const &[addStates, modelFile, propositionsFile] = arguments.value();
        SfsmMethod(addStates, modelFile, propositionsFile);
        return 0;
    } else {
        printUsage(argv[0]);
        return -1;
    }
}
