#include "processing/RestrictedSfsmDistinguishingInputCalculation.hpp"

#include <vector>
#include <future>

#include "utils/concurrency.hpp"
#include "sfsm/SMTTools.hpp"
#include "sfsm/Variable.hpp"

namespace libfsmtest::SFSMs {

static RestrictedSfsmDistinguishingInputCalculator::result_type calculateDistinguishingInputMap(Expressions const &guardConditions, Expressions const &outputExpressions, Expression const &invariant, IdentifierDefinitions const &identifierDefinitions) {
    std::vector<std::tuple<size_t, size_t, std::shared_future<SatisfyingValuationResult>>> resultPreparation;
    // Filter identifierDefinitions for outputs
    IdentifierDefinitions outputDefinitions;
    std::copy_if(identifierDefinitions.begin(), identifierDefinitions.end(), std::back_inserter(outputDefinitions), [](auto const &idDef) {
        auto const &[id, type] = idDef;
        return type.getLocation() == IdentifierType::LocationInModel::OUTPUT;
    });
    // Filter identifierDefinitions for inputs
    IdentifierDefinitions inputDefinitions;
    std::copy_if(identifierDefinitions.begin(), identifierDefinitions.end(), std::back_inserter(inputDefinitions), [](auto const &idDef) {
        auto const &[id, type] = idDef;
        return type.getLocation() == IdentifierType::LocationInModel::INPUT;
    });
    // First, calculate the set of guard condition/output expression pairs that
    // could label a valid transition (i.e. that are satisfiable under conjunction)
    std::vector<std::tuple<size_t, size_t, std::shared_future<SatisfyingValuationResult>>> inOutResults;
    for(size_t idxGuardCondition = 0; idxGuardCondition < guardConditions.size(); ++idxGuardCondition) {
        for(size_t idxOutputExpression = 0; idxOutputExpression < outputExpressions.size(); ++idxOutputExpression) {
            auto const &guardCondition = guardConditions.at(idxGuardCondition);
            auto const &outputExpression = outputExpressions.at(idxOutputExpression);
            auto const conjunction = makeLogicalConjunction(guardCondition, outputExpression);
            inOutResults.emplace_back(idxGuardCondition, idxOutputExpression, async(isSatisfiable(makeLogicalConjunction(invariant, conjunction), identifierDefinitions)));
        }
    }
    
    // Then collect all output expressions for each input, where the
    // conjunction of an output expression with that input is satisfiable.
    std::vector<std::tuple<size_t, size_t, size_t, std::future<SatisfyingValuationResult>>> equivalenceResults;
    std::map<size_t, std::vector<size_t>> satisfiableOutputExpressionsPerInput;
    for(auto &[idxInput, idxOutput, future] : inOutResults) {
        if(future.get().has_value()) {
            satisfiableOutputExpressionsPerInput[idxInput].push_back(idxOutput);
        }
    }
    // For each guard condition, check all pairs of output expressions, that
    // individually are satisfiable under conjunction with said guard
    // condition, for equivalence.
    for(auto &[idxInput, outputIndices] : satisfiableOutputExpressionsPerInput) {
        auto const &guardCondition = guardConditions.at(idxInput);
        for(size_t idx1 = 0; idx1 < outputIndices.size(); ++idx1) {
            auto const &outputExpression1 = outputExpressions.at(outputIndices.at(idx1));
            for(size_t idx2 = idx1 + 1; idx2 < outputIndices.size(); ++idx2) {
                auto const &outputExpression2 = outputExpressions.at(outputIndices.at(idx2));
                auto const problem = makeLogicalConjunction(invariant, guardCondition, makeComparisonInequality(outputExpression1, outputExpression2));
                equivalenceResults.emplace_back(idxInput, outputIndices.at(idx1), outputIndices.at(idx2), async(isSatisfiable(problem, identifierDefinitions)));
            }
        }
    }
    // Collect all unequal pairs of output expressions for each guard condition
    std::map<size_t, std::multimap<size_t, size_t>> unequalOutputs;
    for(auto &[idxInput, idxOutput1, idxOutput2, distinguishable] : equivalenceResults) {
        if(distinguishable.get().has_value()) {
            unequalOutputs[idxInput].insert({idxOutput1, idxOutput2});
            unequalOutputs[idxInput].insert({idxOutput2, idxOutput1});
        }
    }
    auto getUnequalOutputs = [&unequalOutputs,&outputExpressions](size_t idxInput, size_t idxOutput) {
        Expressions result;
        auto [begin, end] = unequalOutputs.at(idxInput).equal_range(idxOutput);
        for(auto unequalOutputsIter = begin; unequalOutputsIter != end; ++unequalOutputsIter) {
            auto [_, value] = *unequalOutputsIter;
            result.push_back(outputExpressions.at(value));
        }
        return result;
    };
    // Check for each feasible guard condition/output expression pair whether
    // there is an output valuation that distinguishes the output expression
    // from all other output expressions that are
    //  * also feasible with this guard condition
    //  * are not equal to the output expression under discussion
    std::vector<std::tuple<size_t, size_t, Expressions, std::future<SatisfyingValuationResult>>> finalResults;
    for(auto &[idxInput, idxOutput, future] : inOutResults) {
        if(future.get().has_value()) {
            Expressions otherOutputs = getUnequalOutputs(idxInput, idxOutput);
            Expression const &input = guardConditions.at(idxInput);
            Expression const &output = outputExpressions.at(idxOutput);
            if(otherOutputs.empty()) {
                auto &idxInputLambda = idxInput;
                auto &idxOutputLambda = idxOutput;
                auto inOutResult = std::find_if(inOutResults.begin(), inOutResults.end(), [idxInput=idxInputLambda, idxOutput=idxOutputLambda](auto const &elem) {
                    auto const &[in, out, _] = elem;
                    if(in == idxInput and out == idxOutput) {
                        return true;
                    }
                    return false;
                });
                resultPreparation.emplace_back(idxInput, idxOutput, std::get<std::shared_future<SatisfyingValuationResult>>(*inOutResult));
            } else {
                // forall o_1, o_2, o_3, ... , o_n. out1(o_1, o_2, o_3, ... , o_n) implies not(others(o_1, o_2, o_3, ... , o_n))
                // Implication: Outputs that fulfill `output` do not fulfill any of the other outputs
                auto const implication = makeLogicalImplication(output, makeLogicalNegation(makeLogicalDisjunction(otherOutputs)));
                // The implication has to hold only together with `input` being fulfilled by some input valuation
                // We reuse the output identifiers. For every output variable there is now a variable symbol bound by the universal quantifier
                auto const conjunction = makeLogicalConjunction(input, makeUniversallyQuantifiedFormula(outputDefinitions, implication));
                // Is there a result in within the invariant
                resultPreparation.emplace_back(idxInput, idxOutput, async(isSatisfiable(makeLogicalConjunction(invariant, conjunction), identifierDefinitions)));
            }
        }
    }
    std::map<std::tuple<size_t, size_t>, Valuation> returnValue;
    for(auto &[idxInput, idxOutput, future] : resultPreparation) {
        if(auto &singleResult = future.get(); singleResult.has_value()) {
            Valuation inputValuation{Valuation(singleResult.value()).getValuations(inputDefinitions)};
            returnValue.insert({{idxInput, idxOutput}, inputValuation});
        } else {
            // Diagnose which outputs cannot be distinguished
            std::vector<std::tuple<Expression, Expression, Expression, std::future<SatisfyingValuationResult>>> diagnosticResults;
            for(auto &[input, output, result] : resultPreparation) {
                if(not result.get().has_value()) {
                    Expression const &inputExpression = guardConditions.at(input);
                    Expression const &outputExpression = outputExpressions.at(output);
                    for(auto const &otherOutput : getUnequalOutputs(input, output)) {
                        auto const implication = makeLogicalImplication(outputExpression, makeLogicalNegation(otherOutput));
                        auto const conjunction = makeLogicalConjunction(inputExpression, makeUniversallyQuantifiedFormula(outputDefinitions, implication));
                        auto const problem = makeLogicalConjunction(invariant, conjunction);
                        diagnosticResults.emplace_back(inputExpression, outputExpression, otherOutput, async(isSatisfiable(makeLogicalConjunction(invariant, conjunction), identifierDefinitions)));
                    }
                }
            }
            std::string errorMessage;
            for(auto &[input, output, otherOutput, result] : diagnosticResults) {
                if(not result.get().has_value()) {
                    errorMessage += "Output expressions ";
                    errorMessage += '`' + std::to_string(output) + "`";
                    errorMessage += " and ";
                    errorMessage += '`' + std::to_string(otherOutput) + "`";
                    errorMessage += " are not distinct for guard condition ";
                    errorMessage += '`' + std::to_string(input) + "`.\n";
                }
            }
            throw std::runtime_error("Cannot pick distinguishing inputs:\n" + errorMessage);
        }
    }
    return returnValue;
}

RestrictedSfsmDistinguishingInputCalculator::result_type RestrictedSfsmDistinguishingInputCalculator::operator()(Sfsm const &sfsm) {
    return calculateDistinguishingInputMap(sfsm.getGuardConditions(), sfsm.getOutputExpressions(), sfsm.getInvariant(), sfsm.getIdentifierDefinitions());
}

}

