#include "processing/checkers/RestrictednessChecker.hpp"

#include <vector>
#include <future>
#include <iostream>

#include "processing/checkers/MinimalityChecker.hpp"
#include "processing/checkers/ObservabilityChecker.hpp"
#include "utils/concurrency.hpp"
#include "sfsm/SMTTools.hpp"
#include "sfsm/Variable.hpp"

namespace libfsmtest::SFSMs {

RestrictednessChecker::RestrictednessChecker(Expressions const &atomicPropositions)
 : APs(atomicPropositions) { }

bool inputSpacePartitioned(Expressions const &guardConditions, Expression const &invariant, IdentifierDefinitions const &identifierDefinitions) {
    auto notGuardConditions = makeLogicalNegation(makeLogicalDisjunction(guardConditions));
    if(isSatisfiable(makeLogicalConjunction(invariant, notGuardConditions), identifierDefinitions).has_value()) {
        std::cerr << "The guard conditions do not cover the whole input domain" << std::endl;
        return false;
    }
    // Check for overlapping
    std::vector<std::future<SatisfyingValuationResult>> overlapResults;
    for(size_t idx1 = 0; idx1 < guardConditions.size(); ++idx1) {
        for(size_t idx2 = idx1 + 1; idx2 < guardConditions.size(); ++idx2) {
            overlapResults.push_back(async(isSatisfiable(makeLogicalConjunction(invariant, guardConditions.at(idx1), guardConditions.at(idx2)), identifierDefinitions)));
        }
    }
    for(auto &result : overlapResults) {
        if(result.get().has_value()) {
            std::cerr << "Guard conditions do overlap" << std::endl;
            return false;
        }
    }
    return true;
}

bool alphabetsImplyAPs(Expressions const &guardConditions, Expressions const &outputExpressions, Expressions const &APs, Expression const &invariant, IdentifierDefinitions const &identifierDefinitions) {
    std::vector<std::tuple<Expression, std::future<SatisfyingValuationResult>>> inOutResults;
    for(auto const &guardCondition : guardConditions) {
        for(auto const &outputExpression : outputExpressions) {
            auto const conjunction = makeLogicalConjunction(guardCondition, outputExpression);
            inOutResults.emplace_back(conjunction, async(isSatisfiable(makeLogicalConjunction(invariant, conjunction), identifierDefinitions)));
        }
    }
    std::vector<std::tuple<Expression, Expression, std::future<SatisfyingValuationResult>>> apResults;
    for(auto &[conjunction, future] : inOutResults) {
        if(future.get().has_value()) {
            auto const &conjCopyForLambda = conjunction;
            for(auto const &ap : APs) {
                apResults.emplace_back(conjunction, ap, async(isSatisfiable(makeLogicalConjunction(invariant, conjCopyForLambda, ap), identifierDefinitions)));
            }
        }
    }
    std::vector<std::future<SatisfyingValuationResult>> finalResults;
    for(auto &[conjunction, ap, future] : apResults) {
        if(future.get().has_value()) {
            auto const &conjCopyForLambda = conjunction;
            auto const &apCopyForLambda = ap;
            finalResults.push_back(async(isSatisfiable(makeLogicalConjunction(invariant, conjCopyForLambda, makeLogicalNegation(apCopyForLambda)), identifierDefinitions)));
        }
    }
    for(auto &result : finalResults) {
        if(result.get().has_value()) {
            return false;
        }
    }
    return true;
}

bool outputsDistinct(Expressions const &guardConditions, Expressions const &outputExpressions, Expression const &invariant, IdentifierDefinitions const &identifierDefinitions) {
    // Filter identifierDefinitions for outputs
    IdentifierDefinitions outputDefinitions;
    std::copy_if(identifierDefinitions.begin(), identifierDefinitions.end(), std::back_inserter(outputDefinitions), [](auto const &idDef) {
        auto const &[id, type] = idDef;
        return type.getLocation() == IdentifierType::LocationInModel::OUTPUT;
    });
    // First, calculate the set of guard condition/output expression pairs that
    // could label a valid transition (i.e. that are satisfiable under conjunction)
    std::vector<std::tuple<size_t, size_t, std::shared_future<SatisfyingValuationResult>>> inOutResults;
    for(size_t idxGuardCondition = 0; idxGuardCondition < guardConditions.size(); ++idxGuardCondition) {
        for(size_t idxOutputExpression = 0; idxOutputExpression < outputExpressions.size(); ++idxOutputExpression) {
            auto const &guardCondition = guardConditions.at(idxGuardCondition);
            auto const &outputExpression = outputExpressions.at(idxOutputExpression);
            auto const conjunction = makeLogicalConjunction(guardCondition, outputExpression);
            inOutResults.emplace_back(idxGuardCondition, idxOutputExpression, async(isSatisfiable(makeLogicalConjunction(invariant, conjunction), identifierDefinitions)));
        }
    }
    
    // Then collect all output expressions for each input, where the
    // conjunction of an output expression with that input is satisfiable.
    std::vector<std::tuple<size_t, size_t, size_t, std::future<SatisfyingValuationResult>>> equivalenceResults;
    std::map<size_t, std::vector<size_t>> satisfiableOutputExpressionsPerInput;
    for(auto &[idxInput, idxOutput, future] : inOutResults) {
        if(future.get().has_value()) {
            satisfiableOutputExpressionsPerInput[idxInput].push_back(idxOutput);
        }
    }
    // For each guard condition, check all pairs of output expressions, that
    // individually are satisfiable under conjunction with said guard
    // condition, for equivalence.
    for(auto &[idxInput, outputIndices] : satisfiableOutputExpressionsPerInput) {
        auto const &guardCondition = guardConditions.at(idxInput);
        for(size_t idx1 = 0; idx1 < outputIndices.size(); ++idx1) {
            auto const &outputExpression1 = outputExpressions.at(outputIndices.at(idx1));
            for(size_t idx2 = idx1 + 1; idx2 < outputIndices.size(); ++idx2) {
                auto const &outputExpression2 = outputExpressions.at(outputIndices.at(idx2));
                auto const problem = makeLogicalConjunction(invariant, guardCondition, makeComparisonInequality(outputExpression1, outputExpression2));
                equivalenceResults.emplace_back(idxInput, outputIndices.at(idx1), outputIndices.at(idx2), async(isSatisfiable(problem, identifierDefinitions)));
            }
        }
    }
    // Collect all unequal pairs of output expressions for each guard condition
    std::map<size_t, std::multimap<size_t, size_t>> unequalOutputs;
    for(auto &[idxInput, idxOutput1, idxOutput2, distinguishable] : equivalenceResults) {
        if(distinguishable.get().has_value()) {
            unequalOutputs[idxInput].insert({idxOutput1, idxOutput2});
            unequalOutputs[idxInput].insert({idxOutput2, idxOutput1});
        }
    }
    // Check for each feasible guard condition/output expression pair whether
    // there is an output valuation that distinguishes the output expression
    // from all other output expressions that are
    //  * also feasible with this guard condition
    //  * are not equal to the output expression under discussion
    std::vector<std::tuple<Expression, Expression, Expressions, std::future<SatisfyingValuationResult>>> finalResults;
    for(auto &[idxInput, idxOutput, future] : inOutResults) {
        if(future.get().has_value()) {
            Expressions otherOutputs;
            auto [begin, end] = unequalOutputs.at(idxInput).equal_range(idxOutput);
            for(auto unequalOutputsIter = begin; unequalOutputsIter != end; ++unequalOutputsIter) {
                auto [_, value] = *unequalOutputsIter;
                otherOutputs.push_back(outputExpressions.at(value));
            }
            if(not otherOutputs.empty()) {
                Expression const &input = guardConditions.at(idxInput);
                Expression const &output = outputExpressions.at(idxOutput);
                // forall o_1, o_2, o_3, ... , o_n. out1(o_1, o_2, o_3, ... , o_n) implies not(others(o_1, o_2, o_3, ... , o_n))
                // Implication: Outputs that fulfill `output` do not fulfill any of the other outputs
                auto const implication = makeLogicalImplication(output, makeLogicalNegation(makeLogicalDisjunction(otherOutputs)));
                // The implication has to hold only together with `input` being fulfilled by some input valuation
                // We reuse the output identifiers. For every output variable there is now a variable symbol bound by the universal quantifier
                auto const conjunction = makeLogicalConjunction(input, makeUniversallyQuantifiedFormula(outputDefinitions, implication));
                // Is there a result in within the invariant
                finalResults.emplace_back(input, output, otherOutputs, async(isSatisfiable(makeLogicalConjunction(invariant, conjunction), identifierDefinitions)));
            }
        }
    }
    std::vector<std::tuple<Expression, Expression, Expression, std::future<SatisfyingValuationResult>>> diagnosticResults;
    for(auto &[input, output, otherOutputs, result] : finalResults) {
        if(not result.get().has_value()) {
            for(auto const &otherOutput : otherOutputs) {
                auto const implication = makeLogicalImplication(output, makeLogicalNegation(otherOutput));
                auto const conjunction = makeLogicalConjunction(input, makeUniversallyQuantifiedFormula(outputDefinitions, implication));
                diagnosticResults.emplace_back(input, output, otherOutput, async(isSatisfiable(makeLogicalConjunction(invariant, conjunction), identifierDefinitions)));
            }
        }
    }
    for(auto &[input, output, otherOutput, result] : diagnosticResults) {
        if(not result.get().has_value()) {
            std::cerr << "Output expressions ";
            std::cerr << '`' << std::to_string(output) << "`";
            std::cerr << " and ";
            std::cerr << '`' << std::to_string(otherOutput) << "`";
            std::cerr << " are not distinct for guard condition ";
            std::cerr << '`' << std::to_string(input) << "`." << std::endl;
        }
    }
    return diagnosticResults.empty();
}

bool RestrictednessChecker::operator()(Sfsm const &sfsm) {
    if(not outputsDistinct(sfsm.getGuardConditions(), sfsm.getOutputExpressions(), sfsm.getInvariant(), sfsm.getIdentifierDefinitions())) {
        std::cerr << "Some outputs are not distinguishable" << std::endl;
        return false;
    }
    if(not applyProcessor<MinimalityChecker>(sfsm)) {
        std::cerr << "Not minimal" << std::endl;
        return false;
    }
    if(not applyProcessor<ObservabilityChecker>(sfsm)) {
        std::cerr << "Not observable" << std::endl;
        return false;
    }
    if(not inputSpacePartitioned(sfsm.getGuardConditions(), sfsm.getInvariant(), sfsm.getIdentifierDefinitions())) {
        std::cerr << "Input space not partitioned" << std::endl;
        return false;
    }
    if(not alphabetsImplyAPs(sfsm.getGuardConditions(), sfsm.getOutputExpressions(), APs, sfsm.getInvariant(), sfsm.getIdentifierDefinitions())) {
        std::cerr << "Alphabets not fine grained enough for APs" << std::endl;
        return false;
    }
    return true;
}

}

