#include "sfsm/Expressions.hpp"

#include <libexpressions/expressions/expression_node.hpp>
#include <regex>
#include <unordered_set>
#include <string>
#include <libexpressions/expressions/expression_factory.hpp>
#include <vector>
#include <algorithm>
#include <libexpressions/parsers/s-expressions/interface.hpp>
#include <stdexcept>
#include <iterator>
#include <libexpressions/utils/tree-visit.hpp>
#include <libexpressions/utils/expression-tree-visit.hpp>
#include <libexpressions/expressions/operator.hpp>
#include <libexpressions/expressions/atom.hpp>
#include <libexpressions/expressions/expression_replacement.hpp>
#include <unordered_map>
#include <functional>
#include <libexpressions/matchers/matchers.hpp>
#include <libexpressions/matchers/expressions/PredicateMatcher.hpp>
#include <libexpressions/parsers/ASTConversion.hpp>
#include <libexpressions/evaluators/evaluator.hpp>
#include <charconv>
#include <cmath>

#include "sfsm/ExpressionDefinitions.hpp"
#include "sfsm/Variable.hpp"
#include "sfsm/typeparser/TypeParser.hpp"

namespace libfsmtest::SFSMs {

class Expression::ExpressionImpl {
private:
    libexpressions::ExpressionNodePtr ptr;
public:
    explicit ExpressionImpl(libexpressions::ExpressionNodePtr paramPtr) : ptr(std::move(paramPtr)) { }
    ExpressionImpl(ExpressionImpl const &) = default;
    ExpressionImpl(ExpressionImpl &&) = default;
    ~ExpressionImpl() = default;

    ExpressionImpl &operator=(ExpressionImpl const &) = default;
    ExpressionImpl &operator=(ExpressionImpl &&) = default;

    [[nodiscard]] libexpressions::ExpressionNodePtr const &getPtr() const {
        return this->ptr;
    }

    void setPtr(libexpressions::ExpressionNodePtr paramPtr) {
        this->ptr = std::move(paramPtr);
    }
};

bool operator<(Expression const &expression1, Expression const &expression2) {
    return expression1.impl->getPtr() < expression2.impl->getPtr();
}
bool operator==(Expression const &expression1, Expression const &expression2) {
    return expression1.impl->getPtr() == expression2.impl->getPtr();
}

std::size_t Expression::getHash() const {
    return this->impl->getPtr()->hash();
}

bool checkIdentifierForValidity(Identifier const &id) {
    // Allowed identifier patterns
    static std::regex checker = std::regex("([_a-z-][[:alnum:]-]*)|(-?[01-9]+.?[01-9]*)", std::regex::icase | std::regex::optimize | std::regex::ECMAScript);
    // SMT LIB2 reserved keywords and operators
    static std::unordered_set<std::string> keywords = {"not", "distinct",
    "and", "or", "as", "exists", "forall", "let", "match", "par", "assert",
    "check-sat", "check-sat-assuming", "declare-const", "declare-datatype",
    "declare-datatypes", "declare-fun", "declare-sort", "define-fun",
    "define-fun-rec", "define-sort", "echo", "exit", "get-assertions",
    "get-assignment", "get-info", "get-model", "get-option", "get-proof",
    "get-unsat-assumptions", "get-unsat-core", "get-value", "pop", "push",
    "reset", "reset-assertions", "set-info", "set-logic", "set-option",
    "store", "select", "BINARY", "DECIMAL", "HEXADECIMAL", "NUMERAL",
    "STRING"};
    return (std::regex_match(id, checker))
            and keywords.count(id) == 0;
}

Expression makeIdentifier(Identifier const &id) {
    return Expression(new Expression::ExpressionImpl(libexpressions::ExpressionFactory::get()->makeIdentifier(id)));
}
Expression makeNewIdentifier(Identifier const &prefix) {
    return Expression(new Expression::ExpressionImpl(libexpressions::ExpressionFactory::get()->makeNewIdentifier(prefix)));
}
Expression makeFunctionCall(Expression const &identifier, Expressions const &arguments) {
    Expressions pack(1, identifier);
    pack.insert(pack.end(), arguments.begin(), arguments.end());
    return makeFunctionCall(pack);
}
Expression makeFunctionCall(Expressions const &identifierAndArguments) {
    std::vector<libexpressions::ExpressionNodePtr> unpacked;
    std::transform(identifierAndArguments.begin(), identifierAndArguments.end(), std::back_inserter(unpacked), [](auto const &expression) {
        return expression.getImpl()->getPtr();
    });
    return Expression(new Expression::ExpressionImpl(libexpressions::ExpressionFactory::get()->makeExpression(unpacked)));
}
static Expression makeUnaryOperator(std::string const &operatr,
                                     Expression const &expression) {
    return Expression(new Expression::ExpressionImpl(libexpressions::ExpressionFactory::get()->makeExpression(
                                                      operatr,
                                                      expression.getImpl()->getPtr())));
}
static Expression makeBinaryOperator(std::string const &operatr,
                                     Expression const &expression1,
                                     Expression const &expression2) {
    return Expression(new Expression::ExpressionImpl(libexpressions::ExpressionFactory::get()->makeExpression(
                                                      operatr,
                                                      expression1.getImpl()->getPtr(),
                                                      expression2.getImpl()->getPtr())));
}
static Expression makeNaryOperator(std::string const &operatr,
                                     Expressions const &expressions) {
    return makeFunctionCall(makeIdentifier(operatr), expressions);
}

Expressions parseToExpressions(std::string const &text) {
    libexpressions::parsers::SExpressionRepresentationInterface interface;
    try {
        auto expressions = libexpressions::parsers::generateExpressionsFromString(libexpressions::ExpressionFactory::get(), &interface, text);
        Expressions result;
        std::transform(expressions.begin(), expressions.end(), std::back_inserter(result), [](auto &expression) {
            return Expression(new Expression::ExpressionImpl(expression));
        });
        return result;
    } catch(std::runtime_error &error) {
        throw std::runtime_error(std::string(error.what()) + "\nReferring to input:\n" + text + "\n");
    }
}
Expression makeLogicalConjunction(Expression const &expression1, Expression const &expression2) {
    return makeBinaryOperator("and", expression1, expression2);
}
[[nodiscard]] Expression makeLogicalConjunction(Expressions const &expressions) {
    return makeNaryOperator("and", expressions);
}
Expression makeLogicalNegation(Expression const &expression) {
    return makeUnaryOperator("not", expression);
}
Expression makeComparisonEquality(Expression const &expression1, Expression const &expression2) {
    return makeBinaryOperator("=", expression1, expression2);
}
[[nodiscard]] Expression makeComparisonEquality(Expressions const &expressions) {
    return makeNaryOperator("=", expressions);
}
Expression makeComparisonLess(Expression const &expression1, Expression const &expression2) {
    return makeBinaryOperator("<", expression1, expression2);
}
Expression makeArithmeticAddition(Expression const &expression1, Expression const &expression2) {
    return makeBinaryOperator("+", expression1, expression2);
}
[[nodiscard]] Expression makeArithmeticAddition(Expressions const &expressions) {
    return makeNaryOperator("+", expressions);
}
Expression makeArithmeticNegation(Expression const &expression) {
    return makeUnaryOperator("-", expression);
}
Expression makeArithmeticMultiplication(Expression const &expression1, Expression const &expression2) {
    return makeBinaryOperator("*", expression1, expression2);
}
[[nodiscard]] Expression makeArithmeticMultiplication(Expressions const &expressions) {
    return makeNaryOperator("*", expressions);
}
Expression makeArithmeticDivision(Expression const &expression1, Expression const &expression2) {
    return makeBinaryOperator("/", expression1, expression2);
}
Expression makeArithmeticIntegerDivision(Expression const &expression1, Expression const &expression2) {
    return makeBinaryOperator("div", expression1, expression2);
}
Expression makeArithmeticModulus(Expression const &expression1, Expression const &expression2) {
    return makeBinaryOperator("mod", expression1, expression2);
}

Expression makeLogicalDisjunction(Expression const &expression1, Expression const &expression2) {
    return makeBinaryOperator("or", expression1, expression2);
}
Expression makeLogicalDisjunction(Expressions const &expressions) {
    return makeNaryOperator("or", expressions);
}
Expression makeLogicalImplication(Expression const &expression1, Expression const &expression2) {
    return makeBinaryOperator("=>", expression1, expression2);
}
Expression makeComparisonInequality(Expression const &expression1, Expression const &expression2) {
    return makeBinaryOperator("distinct", expression1, expression2);
}
Expression makeComparisonInequality(Expressions const &expressions) {
    return makeNaryOperator("distinct", expressions);
}
Expression makeComparisonLessOrEquals(Expression const &expression1, Expression const &expression2) {
    return makeBinaryOperator("<=", expression1, expression2);
}
Expression makeComparisonGreater(Expression const &expression1, Expression const &expression2) {
    return makeBinaryOperator(">", expression1, expression2);
}
Expression makeComparisonGreaterOrEquals(Expression const &expression1, Expression const &expression2) {
    return makeBinaryOperator(">=", expression1, expression2);
}
Expression makeArithmeticSubtraction(Expression const &expression1, Expression const &expression2) {
    return makeBinaryOperator("-", expression1, expression2);
}

Expressions makeLogicalNegation(Expressions const &expressions) {
    Expressions result;
    std::transform(expressions.cbegin(), expressions.cend(), std::back_inserter(result), [](auto const &exp) {
        return makeLogicalNegation(exp);
    });
    return result;
}

Expressions makeArithmeticNegation(Expressions const &expressions) {
    Expressions result;
    std::transform(expressions.cbegin(), expressions.cend(), std::back_inserter(result), [](auto const &exp) {
        return makeArithmeticNegation(exp);
    });
    return result;
}

Expression makeIfThenElse(Expression const &condition, Expression const &consequence_true, Expression const &consequence_false) {
    return makeFunctionCall(makeIdentifier("ite"), {condition, consequence_true, consequence_false});
}

Expression::~Expression() {
    delete this->impl;
}
Expression::Expression(Expression const &other) {
    this->impl = new Expression::ExpressionImpl(*other.impl);
}
Expression::Expression(Expression &&other) noexcept {
    this->impl = other.impl;
    other.impl = nullptr;
}
Expression &Expression::operator=(Expression const &other) {
    if(this != &other) {
        delete this->impl;
        this->impl = new Expression::ExpressionImpl(*other.impl);
    }
    return *this;
}
Expression &Expression::operator=(Expression &&other) noexcept {
    delete this->impl;
    this->impl = other.impl;
    other.impl = nullptr;
    return *this;
}

Expressions getMaximumSubexpressionsWithoutForbiddenIdentifiers(Identifiers const &forbiddenIdentifiers, Expression const &expression) {
    std::unordered_multiset<libexpressions::ExpressionNodePtr> withoutForbiddenOperator;
    libexpressions::traverseTree<libexpressions::TreeTraversalOrder::POSTFIX_TRAVERSAL>(libexpressions::getChildrenIteratorsForExpressionNode,
        [&forbiddenIdentifiers, &withoutForbiddenOperator](auto const &node, auto const &/*path*/) {
        if(auto op = dynamic_cast<libexpressions::Operator const*>(node.get()); op != nullptr) {
            if(std::all_of(op->cbegin(), op->cend(), [&withoutForbiddenOperator](auto const &operand) {
                return withoutForbiddenOperator.count(operand) > 0;
            })) {
                withoutForbiddenOperator.insert(node);
                for(auto const &operand : *op) {
                    withoutForbiddenOperator.erase(operand);
                }
            }
        } else if(auto atom = dynamic_cast<libexpressions::Atom const*>(node.get()); atom != nullptr) {
            if(std::find(forbiddenIdentifiers.cbegin(), forbiddenIdentifiers.cend(), atom->getSymbol()) == forbiddenIdentifiers.cend()) {
                withoutForbiddenOperator.insert(node);
            }
        }
    }, expression.getImpl()->getPtr());

    Expressions result;
    std::transform(withoutForbiddenOperator.cbegin(),
                   withoutForbiddenOperator.cend(),
                   std::back_inserter(result),
                   [](auto const &node) {
        return Expression(new Expression::ExpressionImpl(node));
    });

    return result;
}

IdentifierType parseType(std::string const &typeString) {
    return parseIdentifierType(typeString);
}

Expression replaceSubexpressions(Expression const &expressionToReplaceIn, std::vector<std::pair<Expression, Expression>> const &replacementMap) {
    std::unordered_map<libexpressions::ExpressionNodePtr, libexpressions::ExpressionNodePtr> internalReplacementMap;
    std::transform(replacementMap.cbegin(), replacementMap.cend(), std::inserter(internalReplacementMap, internalReplacementMap.end()),
        [](auto const &replacement) {
        auto const &[left, right] = replacement;
        return make_pair(left.getImpl()->getPtr(), right.getImpl()->getPtr());
    });
    return Expression(new Expression::ExpressionImpl(libexpressions::replaceSubexpressionsByEquality(libexpressions::ExpressionFactory::get(),
                                            expressionToReplaceIn.getImpl()->getPtr(),
                                            internalReplacementMap)));
}

bool isFunctionCall(Expression const &expression) {
    auto const expressionPointer = expression.getImpl()->getPtr();
    return dynamic_cast<libexpressions::Operator const*>(expressionPointer.get());
}

std::optional<Expressions> splitFunctionCall(Expression const &expression) {
    auto const expressionPointer = expression.getImpl()->getPtr();
    auto const op = dynamic_cast<libexpressions::Operator const*>(expressionPointer.get());
    if(op != nullptr) {
        Expressions result;
        auto const &operands = op->getOperands();
        std::transform(operands.cbegin(), operands.cend(), std::back_inserter(result), [](auto const &operand) {
            return Expression(new Expression::ExpressionImpl(operand));
        });
        return result;
    } else {
        return std::nullopt;
    }
}

bool containsIdentifiersWithLocation(IdentifierDefinitions const &identifiers, Expression const &expression, IdentifierType::LocationInModel const &location) {
    bool result = false;
    callOnSubexpressionsIf(expression, 
        [&location,&identifiers](auto const &subexpression) {
            if(not isFunctionCall(subexpression)) {
                auto const strName = std::to_string(subexpression);
                for(auto const &[name, type] : identifiers) {
                    if(name == strName) {
                        if(type.getLocation() == location) {
                            return true;
                        }
                    }
                }
            }
            return false;
        },
        [&result](auto const &/*expression*/) {
            result = true;
        });
    return result;
}

void callOnSubexpressionsIf(Expression const &expression, std::function<bool(Expression const&)> const &predicate, std::function<void(Expression const&)> const &fn) {
    auto expressionInternal = expression.getImpl()->getPtr();
    auto predicateWrap = [&predicate](auto const &exprPtr) {
        return predicate(Expression(new Expression::ExpressionImpl(exprPtr)));
    };
    auto fnWrap = [&fn](auto const &exprPtr, auto const &/*path*/) {
        fn(Expression(new Expression::ExpressionImpl(exprPtr)));
    };
    libexpressions::invokeUsingMatcher<libexpressions::TreeTraversalOrder::POSTFIX_TRAVERSAL>(expressionInternal, libexpressions::Matchers::SatisfiesPredicate(predicateWrap), fnWrap);
}

std::size_t combine_hashes(size_t previousHash, size_t hash) {
    return previousHash ^ (hash + static_cast<std::size_t>(0x701bbd9a57630f95) + (previousHash >> 1) + (previousHash << 2));
}

std::variant<bool,ssize_t,double> getValueInType(IdentifierValuation const &valuation) {
    auto valuationExpressions = parseToExpressions(valuation.second);
    assert(valuationExpressions.size() == 1);
    auto content = valuationExpressions.front();
    enum class OperatorTypes {
        DIVISION,
        MULTIPLICATION,
        ADDITION,
        SUBTRACTION,
        REMAINDER,
        INT_DIVISION,
        ROOT,
        POWER
    };
    enum class ValueTypes {
        BOOLEAN,
        INTEGER,
        FLOAT,
        IDENTIFIER
    };
    typedef std::variant<bool, ssize_t, double, std::string> ValueType;
    typedef std::variant<OperatorTypes, ValueTypes> TypeType;
    class BasicMathSemantics : public libexpressions::Semantics<TypeType, ValueType> {
    public:
        EvaluationState evaluateNonOperatorTerminal(std::string const &c) override {
            std::string const boolean{R"(true|false)"};
            std::string const integer{R"(-?[0-9]*)"};
            std::string const floating{R"(-?[0-9]*.[0-9]*)"};
            std::string const all = (((std::string("(") += boolean) += std::string(")|(") += integer) += std::string(")|(") += floating) += std::string(")");
            std::smatch matchResults;
            std::regex allRegex{all, std::regex::icase | std::regex::optimize | std::regex::ECMAScript};
            if(not std::regex_match(c, matchResults, allRegex)) {
                return {ValueTypes::IDENTIFIER, c};
            } else if(matchResults[1].matched) {
                // Boolean match
                std::regex booleanRegex{"(true)|(false)", std::regex::icase | std::regex::optimize | std::regex::ECMAScript};
                bool matchResultBool = std::regex_match(c, matchResults, booleanRegex);
                assert(matchResultBool);
                assert(matchResults[1].matched || matchResults[2].matched);
                return {ValueTypes::BOOLEAN, matchResults[1].matched};
            } else if(matchResults[2].matched) {
                // Integer match
                ssize_t intResult = 0;
                auto conversionResult = std::from_chars(c.c_str(), c.c_str() + c.size(), intResult);
                assert(conversionResult.ptr == c.c_str()+c.size());
                return {ValueTypes::INTEGER, intResult};
            } else if(matchResults[3].matched) {
                // Some compilers do not support from_chars for double. Using stod until full support.
                /*double floatResult = 0.0;
                auto conversionResult = std::from_chars(c.c_str(), c.c_str() + c.size(), floatResult);
                assert(conversionResult.ptr == c.c_str()+c.size());*/
                size_t pos = 0;
                auto floatResult = stod(c, &pos);
                assert(pos = c.size());
                return {ValueTypes::FLOAT, floatResult};
            } else {
                assert(false);
            }
        }
        EvaluationState evaluateOperatorTerminal(std::string const &c) override {
            std::regex operators{
                    R"((/)|(\*)|(\+)|(-)|(mod)|(div)|(root-obj)|(^))",
                    std::regex::icase | std::regex::optimize | std::regex::ECMAScript
            };
            std::smatch matchResults;
            if(not std::regex_match(c, matchResults, operators)) {
                throw std::runtime_error((std::string("Operator ") += c) += " not known.");
            } else if(matchResults[1].matched) {
                return {OperatorTypes::DIVISION, c};
            } else if(matchResults[2].matched) {
                return {OperatorTypes::MULTIPLICATION, c};
            } else if(matchResults[3].matched) {
                return {OperatorTypes::ADDITION, c};
            } else if(matchResults[4].matched) {
                return {OperatorTypes::SUBTRACTION, c};
            } else if(matchResults[5].matched) {
                return {OperatorTypes::REMAINDER, c};
            } else if(matchResults[6].matched) {
                return {OperatorTypes::INT_DIVISION, c};
            } else if(matchResults[7].matched) {
                return {OperatorTypes::ROOT, c};
            } else if(matchResults[8].matched) {
                return {OperatorTypes::POWER, c};
            } else {
                assert(false);
            }
        }
        EvaluationState evaluateOperator(EvaluationState const &op, std::vector<EvaluationState> const &operands) override {
            auto const &[operVariant, _] = op;
            auto const &opValue = std::get<OperatorTypes>(operVariant);
            if(operands.size() == 2) {
                auto const &[typeVar1, value1] = operands.front();
                auto const &[typeVar2, value2] = operands.back();
                auto const &type1 = std::get<ValueTypes>(typeVar1);
                auto const &type2 = std::get<ValueTypes>(typeVar2);
                class Modulus {
                public:
                    decltype(auto) operator()(ssize_t val1, ssize_t val2) {
                        return val1 % val2;
                    }
                    decltype(auto) operator()(double val1, double val2) {
                        return fmod(val1, val2);
                    }
                    decltype(auto) operator()(double val1, ssize_t val2) {
                        return fmod(val1, static_cast<double>(val2));
                    }
                    decltype(auto) operator()(ssize_t val1, double val2) {
                        return fmod(static_cast<double>(val1), val2);
                    }
                };
                class IntDivision {
                public:
                    decltype(auto) operator()(ssize_t val1, ssize_t val2) {
                        return val1/val2;
                    }
                    decltype(auto) operator()(double val1, double val2) {
                        return trunc(val1/val2);
                    }
                    decltype(auto) operator()(ssize_t val1, double val2) {
                        return trunc(static_cast<double>(val1)/val2);
                    }
                    decltype(auto) operator()(double val1, ssize_t val2) {
                        return trunc(val1/static_cast<double>(val2));
                    }
                };
                auto eval2 = [&opValue](auto const &val1, auto const &val2) {
                    switch(opValue) {
                        case OperatorTypes::DIVISION:
                            return val1 / val2;
                        case OperatorTypes::MULTIPLICATION:
                            return val1 * val2;
                        case OperatorTypes::ADDITION:
                            return val1 + val2;
                        case OperatorTypes::SUBTRACTION:
                            return val1 - val2;
                        case OperatorTypes::REMAINDER:
                            return Modulus{}(val1, val2);
                        case OperatorTypes::INT_DIVISION:
                            return IntDivision{}(val1, val2);
                        case OperatorTypes::ROOT:
                            throw std::runtime_error("root-obj evaluation not implemented");
                            //return Root{}(val1, val2);
                        case OperatorTypes::POWER:
                            throw std::runtime_error("power function evaluation not implemented");
                    }
                };
                if (type1 == ValueTypes::FLOAT) {
                    auto val1 = std::get<double>(value1);
                    if (type2 == ValueTypes::FLOAT) {
                        auto val2 = std::get<double>(value2);
                        return {ValueTypes::FLOAT, eval2(val1, val2)};
                    } else if (type2 == ValueTypes::INTEGER) {
                        auto val2 = std::get<ssize_t>(value2);
                        return {ValueTypes::FLOAT, eval2(val1, val2)};
                    } else {
                        throw std::runtime_error("Unsupported type for basic arithmetic evaluation.");
                    }
                } else if (type1 == ValueTypes::INTEGER) {
                    auto val1 = std::get<ssize_t>(value1);
                    if (type2 == ValueTypes::FLOAT) {
                        auto val2 = std::get<double>(value2);
                        return {ValueTypes::FLOAT, eval2(val1, val2)};
                    } else if (type2 == ValueTypes::INTEGER) {
                        auto val2 = std::get<ssize_t>(value2);
                        return {ValueTypes::INTEGER, eval2(val1, val2)};
                    } else {
                        throw std::runtime_error("Unsupported type for basic arithmetic evaluation.");
                    }
                } else {
                    throw std::runtime_error("Unsupported type for basic arithmetic evaluation.");
                }
            } else {
               throw std::runtime_error("Unsupported number of operands for evaluation.");
           }
        }
    };
    BasicMathSemantics semantics{};
    auto [evaluatedType, evaluatedValue] = libexpressions::evaluateExpression(content.getImpl()->getPtr(), semantics);
    return std::visit([&valuation](auto const &var) {
        if constexpr(std::is_same_v<std::string, std::decay_t<decltype(var)>>) {
            throw std::runtime_error("Cannot evaluate '" + var + "'");
            return std::variant<bool, ssize_t, double>{};
        } else {
            auto const &[name, identifierType] = valuation.first;
            auto &type = identifierType.getType();
            if(std::holds_alternative<IdentifierType::ComplexType>(type)) {
                throw std::runtime_error("Cannot evaluate expressions for complex types. Not implemented.");
            } else {
                auto simpleType = std::get<IdentifierType::SimpleType>(type);
                switch(simpleType.getBaseType()) {
                case IdentifierType::SimpleBaseType::BOOLEAN:
                    return std::variant<bool, ssize_t, double>{static_cast<bool>(var)};
                case IdentifierType::SimpleBaseType::INTEGER:
                    return std::variant<bool, ssize_t, double>{static_cast<ssize_t>(var)};
                case IdentifierType::SimpleBaseType::REAL:
                    return std::variant<bool, ssize_t, double>{static_cast<double>(var)};
                default:
                    throw std::runtime_error("Not implemented.");
                }
            }
        }
        }, evaluatedValue);
}

Valuation getValuationFromExpression(Expression const &content, IdentifierDefinitions const &identifiers) {
    // 1.) Check structure
    auto const &expressionPtr = content.getImpl()->getPtr();
    // Required structures:
    //  * (= <var> <anything>)
    //  * (and (= <var> <anything>) (= ...))
    //  * (and (and ...))

    //  (= <var> <anything>)
    auto equality = libexpressions::Matchers::IsEqual(libexpressions::ExpressionFactory::get()->makeIdentifier("="));
    auto isIdentifier = libexpressions::Matchers::SatisfiesPredicate([&identifiers](auto const &node) {
        auto *atom = dynamic_cast<libexpressions::Atom const *>(node.get());
        if(atom != nullptr) {
            auto symbol = atom->getSymbol();
            for(auto const &[name, type] : identifiers) {
                if(name == symbol) {
                    return true;
                }
            }
        }
        return false;
    });
    std::vector<std::unique_ptr<libexpressions::MatcherImpl>> baseCaseCriteria;
    baseCaseCriteria.push_back(std::move(libexpressions::Matchers::HasNthChild(0)->operator()(std::move(equality))));
    baseCaseCriteria.push_back(std::move(libexpressions::Matchers::HasNthChild(1)->operator()(std::move(isIdentifier))));
    auto baseCaseMatcher = libexpressions::Matchers::And(baseCaseCriteria);

    //  (and (= <var> <anything>) (= ...))
    std::vector<std::unique_ptr<libexpressions::MatcherImpl>> simpleAndCriteria;
    auto andOperatorAtom = libexpressions::Matchers::IsEqual(libexpressions::ExpressionFactory::get()->makeIdentifier("and"));
    simpleAndCriteria.push_back(libexpressions::Matchers::HasNthChild(0)->operator()(andOperatorAtom->clone()));
    simpleAndCriteria.push_back(libexpressions::Matchers::AllChildrenExceptNth(0)->operator()(baseCaseMatcher->clone()));
    auto simpleAndMatcher = libexpressions::Matchers::And(simpleAndCriteria);

    //  (and (and ...))
    std::vector<std::unique_ptr<libexpressions::MatcherImpl>> recursionBaseMatchers;
    recursionBaseMatchers.push_back(baseCaseMatcher->clone());
    recursionBaseMatchers.push_back(andOperatorAtom->clone());
    auto recursionBaseCase = libexpressions::Matchers::Or(recursionBaseMatchers);
    std::vector<std::unique_ptr<libexpressions::MatcherImpl>> recursionAndNotBaseCase;
    recursionAndNotBaseCase.push_back(libexpressions::Matchers::RecursiveUntilProperty(libexpressions::Matchers::HasNthChild(0)->operator()(andOperatorAtom->clone()))->operator()(recursionBaseCase->clone()));
    recursionAndNotBaseCase.push_back(libexpressions::Matchers::Not(recursionBaseCase->clone()));
    std::vector<std::unique_ptr<libexpressions::MatcherImpl>> validExpressionMatchers;
    validExpressionMatchers.push_back(libexpressions::Matchers::And(recursionAndNotBaseCase));
    validExpressionMatchers.push_back(baseCaseMatcher->clone());
    libexpressions::Matcher baseMatcher(std::move(baseCaseMatcher));
    libexpressions::Matcher validExpression(libexpressions::Matchers::Or(validExpressionMatchers));

    if(not validExpression(expressionPtr)) {
        throw std::runtime_error("Expression in wrong format to extract valuation.");
    }
    Valuation result;
    libexpressions::invokeUsingMatcher<libexpressions::TreeTraversalOrder::POSTFIX_TRAVERSAL>(content.getImpl()->getPtr(), baseMatcher, [&result,&identifiers](auto const &expr, auto const &path) {
        auto op = dynamic_cast<libexpressions::Operator const*>(expr.get());
        auto identifierName = op->getOperands().at(1)->toString();
        auto value = op->getOperands().at(2)->toString();
        auto &valuationFunction = result.getValuations();
        bool found = false;
        for(auto const &definition : identifiers) {
            auto const &[name, type] = definition;
            if(name == identifierName) {
                if(valuationFunction.count(definition) > 0) {
                    throw std::runtime_error("Multiple definitions for value of " + name);
                }
                result = result.modify(definition, value);
                found = true;
                break;
            }
        }
        if(not found) {
            throw std::runtime_error("Could not find definition for identifier name " + identifierName);
        }
    });
    return result;
}

}

namespace std {
std::string to_string(libfsmtest::SFSMs::Expression const &expression) {
    auto ast = libexpressions::parsers::generateASTFromExpressions({expression.impl->getPtr()});
    libexpressions::parsers::SExpressionRepresentationInterface interface;
    return interface.expressionListToString(ast);
}
}

