/*
 * Copyright 2022 Niklas Krafczyk, Jan Peleska
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#pragma once

#include <string>
#include <type_traits>
#include <optional>
#include <utility>
#include <vector>
#include <utility>
#include <functional>
#include <variant>
#include <algorithm>
#include <stdexcept>

#include "sfsm/ExpressionDefinitions.hpp"

namespace std {
[[nodiscard]] std::string to_string(libfsmtest::SFSMs::Expression const &expression);

// to_string for IdentifierModels and ValuationFunction
template<typename Container, typename = std::enable_if_t<std::is_same_v<typename Container::value_type, libfsmtest::SFSMs::IdentifierValuation>>>
[[nodiscard]] std::string to_string(Container &&container) {
    std::string result = "(and ";
    std::string openParen = "(";
    std::string closeParen = ")";
    std::string equals = "=";
    std::string space = " ";
    for(auto const&[identifier, value] : container) {
        result += openParen += equals += identifier += space += value += closeParen;
    }
    return result + closeParen;
}
}

namespace libfsmtest::SFSMs {

[[nodiscard]] Expressions parseToExpressions(std::string const &text);

[[nodiscard]] bool checkIdentifierForValidity(Identifier const &id);
[[nodiscard]] Expression makeIdentifier(Identifier const &id);
[[nodiscard]] Expression makeNewIdentifier(Identifier const &prefix);
[[nodiscard]] Expression makeFunctionCall(Expression const &identifier, Expressions const &arguments);
[[nodiscard]] Expression makeFunctionCall(Expressions const &identifierAndArguments);

[[nodiscard]] Expression makeLogicalConjunction(Expression const &expression1, Expression const &expression2);
[[nodiscard]] Expression makeLogicalNegation(Expression const &expression);
[[nodiscard]] Expression makeComparisonEquality(Expression const &expression1, Expression const &expression2);
[[nodiscard]] Expression makeComparisonLess(Expression const &expression1, Expression const &expression2);
[[nodiscard]] Expression makeArithmeticAddition(Expression const &expression1, Expression const &expression2);
[[nodiscard]] Expression makeArithmeticNegation(Expression const &expression);
[[nodiscard]] Expression makeArithmeticMultiplication(Expression const &expression1, Expression const &expression2);
[[nodiscard]] Expression makeArithmeticDivision(Expression const &expression1, Expression const &expression2);
[[nodiscard]] Expression makeArithmeticIntegerDivision(Expression const &expression1, Expression const &expression2);
[[nodiscard]] Expression makeArithmeticModulus(Expression const &expression1, Expression const &expression2);

[[nodiscard]] Expression makeLogicalDisjunction(Expression const &expression1, Expression const &expression2);
[[nodiscard]] Expression makeLogicalImplication(Expression const &expression1, Expression const &expression2);
[[nodiscard]] Expression makeComparisonInequality(Expression const &expression1, Expression const &expression2);
[[nodiscard]] Expression makeComparisonLessOrEquals(Expression const &expression1, Expression const &expression2);
[[nodiscard]] Expression makeComparisonGreater(Expression const &expression1, Expression const &expression2);
[[nodiscard]] Expression makeComparisonGreaterOrEquals(Expression const &expression1, Expression const &expression2);
[[nodiscard]] Expression makeArithmeticSubtraction(Expression const &expression1, Expression const &expression2);

// Expression generation with n-ary operators
[[nodiscard]] Expression makeArithmeticAddition(Expressions const &expressions);
[[nodiscard]] Expression makeArithmeticMultiplication(Expressions const &expressions);
[[nodiscard]] Expression makeComparisonEquality(Expressions const &expressions);
[[nodiscard]] Expression makeComparisonInequality(Expressions const &expressions);
[[nodiscard]] Expression makeLogicalConjunction(Expressions const &expressions);
[[nodiscard]] Expression makeLogicalDisjunction(Expressions const &expressions);

[[nodiscard]] Expression makeIfThenElse(Expression const &condition, Expression const &consequence_true, Expression const &consequence_false);

// Convenience functions to apply unary operators on sets of `Expression`s
[[nodiscard]] Expressions makeLogicalNegation(Expressions const &expressions);
[[nodiscard]] Expressions makeArithmeticNegation(Expressions const &expressions);

// Expression generation with n-ary operators from variable number of individual arguments
template<typename... Args>
[[nodiscard]] Expression makeArithmeticAddition(Args&& ...args);
template<typename... Args>
[[nodiscard]] Expression makeArithmeticMultiplication(Args&& ...args);
template<typename... Args>
[[nodiscard]] Expression makeComparisonEquality(Args&& ...args);
template<typename... Args>
[[nodiscard]] Expression makeComparisonInequality(Args&& ...args);
template<typename... Args>
[[nodiscard]] Expression makeLogicalConjunction(Args&& ...args);
template<typename... Args>
[[nodiscard]] Expression makeLogicalDisjunction(Args&& ...args);

[[nodiscard]] Expressions getMaximumSubexpressionsWithoutForbiddenIdentifiers(Identifiers const &forbiddenIdentifiers, Expression const &expression);

// Replaces subexpressions by other expressions. The latter replacement takes
// precendence, if there are multiple replacements defined for the same
// expression.
[[nodiscard]] Expression replaceSubexpressions(Expression const &expressionToReplaceIn, std::vector<std::pair<Expression, Expression>> const &replacementMap);
[[nodiscard]] Expression makeFunctionDefinition(Identifier const &functionName, IdentifierType const &returnType, IdentifierDefinitions const &parameters, Expression const &content);

// Convenience functions for common checks
[[nodiscard]] bool areEquivalent(Expression const &exp1, Expression const &exp2, IdentifierDefinitions const &identifierDefinitions);
[[nodiscard]] bool isImplication(Expression const &antecedent, Expression const &consequent, IdentifierDefinitions const &identifierDefinitions);

[[nodiscard]] bool isValuationExpression(Expression const &exp1);
[[nodiscard]] IdentifierValuations translateToValuation(Expression const &exp1);

void callOnSubexpressionsIf(Expression const &expression, std::function<bool(Expression const&)> const &predicate, std::function<void(Expression const&)> const &fn);



class Expression {
public:
    class ExpressionImpl;
private:
    ExpressionImpl *impl;
public:
    explicit Expression(ExpressionImpl *paramImpl = nullptr) : impl(paramImpl) { }
    Expression(Expression const &);
    Expression(Expression &&) noexcept;
    ~Expression();

    [[nodiscard]] ExpressionImpl *getImpl() const { return this->impl; }

    Expression &operator=(Expression const &);
    Expression &operator=(Expression &&) noexcept;

    friend bool operator<(Expression const &expression1, Expression const &expression2);
    friend bool operator==(Expression const &expression1, Expression const &expression2);

    [[nodiscard]] size_t getHash() const;
    friend std::string std::to_string(Expression const &expression);
};

class IdentifierType {
public:
    class SimpleType;
    class ComplexType;
    typedef std::variant<SimpleType,ComplexType> Type;
    enum class SimpleBaseType {
        BOOLEAN,
        INTEGER,
        REAL
    };
    class SimpleType {
    private:
        SimpleBaseType baseType;
    public:
        explicit SimpleType(SimpleBaseType base) : baseType(base) {}
        SimpleType(SimpleType const &) = default;
        ~SimpleType() = default;
        SimpleType &operator=(SimpleType const&) = default;
        [[nodiscard]] SimpleBaseType getBaseType() const { return this->baseType; }
        bool operator==(SimpleType const &other) const {
            return this->baseType == other.baseType;
        }
        bool operator<(SimpleType const &other) const {
            return (this->baseType == SimpleBaseType::BOOLEAN and (other.baseType == SimpleBaseType::INTEGER or other.baseType == SimpleBaseType::REAL))
                or (this->baseType == SimpleBaseType::INTEGER and other.baseType == SimpleBaseType::REAL);
        }

    };
    enum class ComplexBaseType {
        ARRAY,
        FUNCTION
    };
    class ComplexType {
    private:
        ComplexBaseType baseType;
        std::vector<Type> params;
    public:
        ComplexType(ComplexBaseType base, std::vector<Type> paramTypes)
         : baseType(base), params(std::move(paramTypes)) { }
        ComplexType(ComplexType const &) = default;
        ~ComplexType() = default;
        ComplexType &operator=(ComplexType const&) = default;
        [[nodiscard]] ComplexBaseType getBaseType() const { return this->baseType; }
        [[nodiscard]] std::vector<Type> getParameters() const { return this->params; }
        bool operator==(ComplexType const &other) const {
            return this->baseType == other.baseType and this->params == other.params;
        }
        bool operator<(ComplexType const &other) const {
            return (this->baseType == ComplexBaseType::ARRAY and other.baseType == ComplexBaseType::FUNCTION)
                or std::lexicographical_compare(this->params.cbegin(), this->params.cend(), other.params.cbegin(), other.params.cend());
        }
    };
    enum class LocationInModel {
        INPUT,
        OUTPUT,
        OTHER
    };
private:
    Type type;
    LocationInModel location;
public:
    IdentifierType(Type t, LocationInModel l) : type(std::move(t)), location(l) { }
    //IdentifierType(SimpleType const &t, LocationInModel l) : type(t), location(l) { }
    //IdentifierType(ComplexType const &t, LocationInModel l) : type(t), location(l) { }
    IdentifierType(IdentifierType const &other) = default;
    ~IdentifierType() = default;

    IdentifierType &operator=(IdentifierType const &other) = default;

    [[nodiscard]] Type const &getType() const { return this->type; }
    [[nodiscard]] LocationInModel getLocation() const { return this->location; }

    bool operator==(IdentifierType const &other) const {
        return this->type == other.type and this->location == other.location;
    }
    bool operator<(IdentifierType const &other) const {
        bool const locationLess = (this->location == LocationInModel::INPUT and (other.location == LocationInModel::OUTPUT or other.location == LocationInModel::OTHER))
                         or (this->location == LocationInModel::OUTPUT and other.location == LocationInModel::OTHER);
        return this->type < other.type or ((not (other.type < this->type)) and locationLess);
    }
};

IdentifierType parseType(std::string const &typeString);

template<typename... Args>
[[nodiscard]] Expression makeArithmeticAddition(Args&& ...args) {
    return makeArithmeticAddition({std::forward<Args>(args)...});
}
template<typename... Args>
[[nodiscard]] Expression makeArithmeticMultiplication(Args&& ...args) {
    return makeArithmeticMultiplication({args...});
}
template<typename... Args>
[[nodiscard]] Expression makeComparisonEquality(Args&& ...args) {
    return makeComparisonEquality({args...});
}
template<typename... Args>
[[nodiscard]] Expression makeComparisonInequality(Args&& ...args) {
    return makeComparisonInequality({args...});
}
template<typename... Args>
[[nodiscard]] Expression makeLogicalConjunction(Args&& ...args) {
    return makeLogicalConjunction({args...});
}
template<typename... Args>
[[nodiscard]] Expression makeLogicalDisjunction(Args&& ...args) {
    return makeLogicalDisjunction({args...});
}

[[nodiscard]] bool isFunctionCall(Expression const &expression);
[[nodiscard]] std::optional<Expressions> splitFunctionCall(Expression const &expression);

bool containsIdentifiersWithLocation(IdentifierDefinitions const &identifiers, Expression const &expression, IdentifierType::LocationInModel const &location);

template<typename Container, typename = std::enable_if_t<
            std::disjunction_v<
                std::is_convertible<std::decay_t<decltype(*std::declval<Container>().begin())>, libfsmtest::SFSMs::Expression>,
                std::is_convertible<std::decay_t<decltype(*std::declval<Container>().begin())>, libfsmtest::SFSMs::Value>,
                std::is_convertible<std::decay_t<decltype(*std::declval<Container>().begin())>, libfsmtest::SFSMs::IdentifierDefinition>,
                std::is_convertible<std::decay_t<decltype(*std::declval<Container>().begin())>, libfsmtest::SFSMs::IdentifierValuation>
            >>>
std::size_t hash_container(Container &&container) {
    std::size_t seed = 0;
    for(auto const &element : container) {
        seed = libfsmtest::SFSMs::combine_hashes(seed, std::hash<std::decay_t<decltype(element)>>{}(element));
    }
    return seed;
}

}

namespace std {
template<>
struct hash<libfsmtest::SFSMs::Expression> {
    std::size_t operator()(libfsmtest::SFSMs::Expression const &expression) const {
        return expression.getHash();
    }
};
template<>
struct hash<libfsmtest::SFSMs::IdentifierType::SimpleType> {
    std::size_t operator()(libfsmtest::SFSMs::IdentifierType::SimpleType const &obj) {
        switch(obj.getBaseType()) {
        case libfsmtest::SFSMs::IdentifierType::SimpleBaseType::BOOLEAN:
            return 1;
        case libfsmtest::SFSMs::IdentifierType::SimpleBaseType::INTEGER:
            return 2;
        case libfsmtest::SFSMs::IdentifierType::SimpleBaseType::REAL:
            return 4;
        default:
            throw std::runtime_error("Invalid type for hashing. Sorry! Forgot to implement that.");
        }
    }
};
template<>
struct hash<libfsmtest::SFSMs::IdentifierType::ComplexType> {
    std::size_t operator()(libfsmtest::SFSMs::IdentifierType::ComplexType const &obj) {
        size_t result = 0;
        for(auto const &paramType : obj.getParameters()) {
            result = libfsmtest::SFSMs::combine_hashes(result, std::hash<libfsmtest::SFSMs::IdentifierType::Type>{}(paramType));
        }
        switch(obj.getBaseType()) {
        case libfsmtest::SFSMs::IdentifierType::ComplexBaseType::ARRAY:
            return libfsmtest::SFSMs::combine_hashes(result, 1);
        case libfsmtest::SFSMs::IdentifierType::ComplexBaseType::FUNCTION:
            return libfsmtest::SFSMs::combine_hashes(result, 2);
        default:
            throw std::runtime_error("Invalid type for hashing. Sorry! Forgot to implement that.");
        }
    }
};
template<>
struct hash<libfsmtest::SFSMs::IdentifierType> {
    std::size_t operator()(libfsmtest::SFSMs::IdentifierType const &obj) const {
        std::size_t locationHash = 0;
        switch(obj.getLocation()) {
        case libfsmtest::SFSMs::IdentifierType::LocationInModel::INPUT:
            locationHash = 1;
            break;
        case libfsmtest::SFSMs::IdentifierType::LocationInModel::OUTPUT:
            locationHash = 2;
            break;
        case libfsmtest::SFSMs::IdentifierType::LocationInModel::OTHER:
            locationHash = 4;
            break;
        default:
            throw std::runtime_error("Invalid type for hashing. Sorry! Forgot to implement that.");
            break;
        }
        return libfsmtest::SFSMs::combine_hashes(std::hash<libfsmtest::SFSMs::IdentifierType::Type>{}(obj.getType()), locationHash);
    }
};
template<>
struct hash<libfsmtest::SFSMs::IdentifierDefinition> {
    std::size_t operator()(libfsmtest::SFSMs::IdentifierDefinition const &obj) const {
        auto const &[identifier, type] = obj;
        return libfsmtest::SFSMs::combine_hashes(std::hash<libfsmtest::SFSMs::Identifier>{}(identifier), std::hash<libfsmtest::SFSMs::IdentifierType>{}(type));
    }
};
template<>
struct hash<libfsmtest::SFSMs::IdentifierValuation> {
    std::size_t operator()(libfsmtest::SFSMs::IdentifierValuation const &obj) const {
        auto const &[identifier, value] = obj;
        return libfsmtest::SFSMs::combine_hashes(std::hash<libfsmtest::SFSMs::IdentifierDefinition>{}(identifier), std::hash<libfsmtest::SFSMs::Value>{}(value));
    }
};
}
