#include "sfsm/SMTTools.hpp"

#include <stdexcept>
#include <variant>
#include <algorithm>
#include <unistd.h>
#include <system_error>
#include <fcntl.h>
#include <string>
#include <vector>
#include <cstdlib>
#include <err.h>
#include <iostream>
#include <sys/wait.h>
#include <optional>
#include <unordered_map>
#include <cassert>
#include <iterator>
#include <sstream>
//#include <threadpool/ThreadPool.hpp>
//#include <threadpool/ThreadPoolPolicies.hpp>

#include "sfsm/Expressions.hpp"

namespace libfsmtest::SFSMs {

Expression getSMTTypeExpression(IdentifierType::Type const &type) {
    class TypeToExpressionVisitor {
    public:
        Expression operator()(IdentifierType::SimpleType const &type) {
            switch(type.getBaseType()) {
                case IdentifierType::SimpleBaseType::BOOLEAN:
                    return makeIdentifier("Bool");
                case IdentifierType::SimpleBaseType::INTEGER:
                    return makeIdentifier("Int");
                case IdentifierType::SimpleBaseType::REAL:
                    return makeIdentifier("Real");
                default:
                    throw std::runtime_error("Could not map type to SMT type");
            }
        }
        Expression operator()(IdentifierType::ComplexType const &type) {
            auto const &complexType = static_cast<IdentifierType::ComplexType const&>(type);
            Expressions operands;
            switch(type.getBaseType()) {
                case IdentifierType::ComplexBaseType::ARRAY:
                    operands.push_back(makeIdentifier("Array"));
                    break;
                case IdentifierType::ComplexBaseType::FUNCTION:
                    break;
                default:
                    throw std::runtime_error("Could not map type to SMT type");
            }
            for(auto const &paramType : complexType.getParameters()) {
                operands.push_back(getSMTTypeExpression(paramType));
            }
            return makeFunctionCall(operands);
        }
    };
    return std::visit(TypeToExpressionVisitor{}, type);
}

Expression getSMTTypeExpression(IdentifierType const &type) {
    return getSMTTypeExpression(type.getType());
}

template<IdentifierType::SimpleBaseType TypeToDetect>
class SimpleTypeDetector {
public:
    bool operator()(IdentifierType::SimpleType const &type) {
        return type.getBaseType() == TypeToDetect;
    }
    bool operator()(IdentifierType::ComplexType const &type) {
        auto const &params = type.getParameters();
        return std::any_of(params.begin(), params.end(), [this](auto const &param) {
            return std::visit(*this, param);
        });
    }
};
template<IdentifierType::ComplexBaseType TypeToDetect>
class ComplexTypeDetector {
public:
    bool operator()(IdentifierType::SimpleType const &/*type*/) {
        return false;
    }
    bool operator()(IdentifierType::ComplexType const &type) {
        auto const &params = type.getParameters();
        return type.getBaseType() == TypeToDetect 
            or std::any_of(params.begin(), params.end(), [this](auto const &param) {
            return std::visit(*this, param);
        });
    }
};

/*static Expression deduceAndGenerateLogicExpression(Expressions const &expressions*//*) {*/
    // Are there integer functions?
    // Are there real functions?
    /*bool ints = std::any_of(identifierDefinitions.begin(), identifierDefinitions.end(), [](auto const &definition) {
        auto const &[_, type] = definition;
        return std::visit(SimpleTypeDetector<IdentifierType::SimpleBaseType::INTEGER>{}, type.getType());
    });
    bool reals = std::any_of(identifierDefinitions.begin(), identifierDefinitions.end(), [](auto const &definition) {
        auto const &[_, type] = definition;
        return std::visit(SimpleTypeDetector<IdentifierType::SimpleBaseType::REAL>{}, type.getType());
    });
    bool arrays = std::any_of(identifierDefinitions.begin(), identifierDefinitions.end(), [](auto const &definition) {
        auto const &[_, type] = definition;
        return std::visit(ComplexTypeDetector<IdentifierType::ComplexBaseType::ARRAY>{}, type.getType());
    });*/
    //std::string logic = "UFNIRA";
    /*if (ints) {
        if(reals) {
            // ints and reals and (arrays or not arrays)
            logic = "AUFNIRA";
        } else {
            if(arrays) {
                // ints and not reals and arrays
                logic = "AUFNIRA";
            } else {
                logic = "QF_UFNIA";
            }
        }
    } else {
        if(reals) {
            if(arrays) {
                // not ints and reals and arrays
                logic = "AUFNIRA";
            } else {
                // not ints and reals and not arrays
                logic = "QF_UFNRA";
            }
        } else {
            if(arrays) {
                // not ints and not reals and arrays
                logic = "QF_AUFLIA";
            } else {
                // not ints and not reals and not arrays
                logic = "QF_UF";
            }
        }
    }*//*
    return makeFunctionCall(makeIdentifier("set-logic"), {makeIdentifier(logic)});
}*/

static void writeToPipe(int pipe, std::string input) {
    auto dataToBeTransfered = input.size();
    while(dataToBeTransfered > 0) {
        auto result = write(pipe, input.data() + (input.size() - dataToBeTransfered), dataToBeTransfered);
        if(result == -1) {
            throw std::system_error(errno, std::system_category());
        }
        dataToBeTransfered -= static_cast<size_t>(result);
    }
}

static int getCurrentStateOfPipe(int pipe) {
    auto pipeFlags = fcntl(pipe, F_GETFL);
    if(pipeFlags == -1) {
            throw std::system_error(errno, std::system_category());
    }
    return pipeFlags;
}

static void setStateOfPipe(int pipe, int flags) {
    if(fcntl(pipe, F_SETFL, flags) == -1) {
            throw std::system_error(errno, std::system_category());
    }
}

static void blockPipe(int pipe, int flags) {
    if(fcntl(pipe, F_SETFL, flags &= ~O_NONBLOCK) == -1) {
            throw std::system_error(errno, std::system_category());
    }
}

static void unblockPipe(int pipe, int flags) {
    if(fcntl(pipe, F_SETFL, flags |= O_NONBLOCK) == -1) {
            throw std::system_error(errno, std::system_category());
    }
}

static std::string readFromPipe(int pipe) {
    std::string result;
    ssize_t returnValue;
    char buffer[512];
    auto pipeFlags = getCurrentStateOfPipe(pipe);
    if(pipeFlags & O_NONBLOCK) {
        blockPipe(pipe, pipeFlags);
    }
    do {
        returnValue = read(pipe, buffer, 512);
        if(returnValue >= 0) {
            unblockPipe(pipe, pipeFlags);
            result.append(buffer, static_cast<size_t>(returnValue));
            if(*(buffer+returnValue-1) == '\n' and
                std::count(result.cbegin(), result.cend(), '(') == std::count(result.cbegin(), result.cend(), ')')) {
                break;
            }
        } else {
            if(errno != EAGAIN) {
                throw std::system_error(errno, std::system_category());
            }
        }
    } while(true);
    setStateOfPipe(pipe, pipeFlags);
    return result;
}

static std::string callSolver(std::string const &input, std::vector<std::string> const &queries) {
    std::string outputOfSolver;
#ifndef LIBSFSMTEST_PLATFORM_WINDOWS
#ifndef NDEBUG
    auto const verboseSolver_envvar = getenv("LIBSFSMTEST_VERBOSE_SOLVER");
    bool const verboseSolver = verboseSolver_envvar != nullptr and (std::string("1") == verboseSolver_envvar);
#else //NDEBUG
    bool const verboseSolver = false;
#endif
    int pipeLibToSMT[2];
    int pipeSMTToLib[2];
    if(not pipe(pipeLibToSMT)) {
        if(not pipe(pipeSMTToLib)) {
            auto pid = fork(); //TBD: Might want to use alternative here. vfork/clone

            if(not pid) {
                close(pipeLibToSMT[1]);
                close(pipeSMTToLib[0]);

                if(dup2(pipeLibToSMT[0], STDIN_FILENO) >= 0) {
                    if(dup2(pipeSMTToLib[1], STDOUT_FILENO) >= 0) {
                        close(pipeLibToSMT[0]);
                        close(pipeSMTToLib[1]);
                        execlp("z3", "-smt2", "-in", "-model", nullptr);
                    }
                }
                err(-1, "Solver call returned");
            } else {
                if(verboseSolver) {
                    std::cout << input << std::endl << std::endl;
                }
                writeToPipe(pipeLibToSMT[1], input+'\n');
                if(not input.empty()) {
                    outputOfSolver = readFromPipe(pipeSMTToLib[0]);
                }
                if(outputOfSolver.substr(0, 3) == "sat") {
                    for(auto const &query : queries) {
                        writeToPipe(pipeLibToSMT[1], query);
                        outputOfSolver += readFromPipe(pipeSMTToLib[0]);
                    }
                } else if(outputOfSolver.substr(0, 5) != "unsat") {
                    throw std::runtime_error(std::string("Unexpected output from SMT solver: ") + outputOfSolver + "\nReferring to this problem:\n" + input + "\n");
                }
                writeToPipe(pipeLibToSMT[1], "(exit)\n");
                if(verboseSolver) {
                    std::cout << outputOfSolver << std::endl << std::endl;
                }
                while(waitpid(pid, nullptr, 0) == -1) {
                    //Waiting may be interrupted. We just wait again. We're patient.
                    if(errno != EINTR) {
                        throw std::system_error(errno, std::system_category());
                    }
                }
            }
            close(pipeSMTToLib[0]);
            close(pipeSMTToLib[1]);
        }
        close(pipeLibToSMT[0]);
        close(pipeLibToSMT[1]);
    }
#endif //ifndef LIBSFSMTEST_PLATFORM_WINDOWS
    return outputOfSolver;
}

std::optional<std::unordered_map<Expression, Expression>> isSatisfiable(Expressions const &expressions, Expressions const &queries) {
    //auto logicExpression = deduceAndGenerateLogicExpression(expressions);
    auto checkSat = makeFunctionCall({makeIdentifier("check-sat")});

    std::ostringstream output;
    std::vector<std::string> queryList;
    //output << std::to_string(logicExpression) << std::endl;
    for(auto const &expression : expressions) {
        output << std::to_string(expression) << std::endl;
    }
    output << std::to_string(checkSat) << std::endl;
    for(auto const &query : queries) {
        queryList.push_back(std::to_string(query) + "\n");
    }

    auto solverOutput = callSolver(output.str(), queryList);
    if(solverOutput.substr(0, 3) == "sat") {
        std::unordered_map<Expression, Expression> result;
        Expressions queryResults = parseToExpressions(solverOutput.substr(3));
        assert(queryResults.size() == queryList.size());
        for(size_t index = 0; index < queryList.size(); ++index) {
            result[queries.at(index)] = queryResults.at(index);
        }
        return result;
    } else {
        if(solverOutput.substr(0, 5) != "unsat") {
            throw std::runtime_error("Solver returned neither `sat` nor `unsat`.\nSolver output: " + solverOutput);
        }
        return {};
    }
}

std::optional<IdentifierValuations> isSatisfiable(Expression const &expression, IdentifierDefinitions const &identifierDefinitions) {
    auto getModelCall = makeFunctionCall({makeIdentifier("get-model")});
    std::vector<Expression> problemStatements;
    std::transform(identifierDefinitions.cbegin(), identifierDefinitions.cend(), std::back_inserter(problemStatements), [](auto const &identifierDefinition) {
        return makeIdentifierDefinition(identifierDefinition);
    });
    problemStatements.push_back(makeFunctionCall({makeIdentifier("assert"), expression}));
    auto query = isSatisfiable(problemStatements, {getModelCall});
    if(query.has_value()) {
        IdentifierValuations models;
        auto modelExpression = query->at(getModelCall);
        auto functionDefinitionMatcher = [](Expression const &paramExpression) {
            if(isFunctionCall(paramExpression)) {
                auto splitCall = splitFunctionCall(paramExpression).value();
                return (not splitCall.empty()) and (std::to_string(splitCall.front()) == "define-fun");
            }
            return false;
        };
        auto functionExtractor = [&models,&identifierDefinitions](Expression const &node) {
            auto splitCall = splitFunctionCall(node).value();
            auto const name = std::to_string(splitCall.at(1));
            auto const idIter = std::find_if(identifierDefinitions.cbegin(), identifierDefinitions.cend(), [&name](auto const &definition) {
                return std::get<std::string>(definition) == name;
            });
#ifndef NDEBUG
            if(idIter == identifierDefinitions.cend()) {
                std::cout << "Unknown function: " << std::to_string(node) << std::endl;
            }
#endif
            assert(idIter != identifierDefinitions.cend());
            IdentifierDefinition const &def = *idIter;
            models.emplace_back(def, std::to_string(splitCall.at(4)));
        };
        callOnSubexpressionsIf(modelExpression, functionDefinitionMatcher, functionExtractor);
        return models;
    } else {
        return std::nullopt;
    }
}

/*std::future<std::optional<IdentifierValuations>> asyncIsSatisfiable(Expression const &expression, IdentifierDefinitions const &identifierDefinitions) {
 *    auto &threadPool = ThreadPool<policies::PolicyCollection<policies::DependenciesNotRespected_t, policies::PassiveWaiting_t>>::getDefaultInstance();
 *    std::optional<IdentifierValuations> (*isSatisfiableFunction)(Expression const&, IdentifierDefinitions const&) = libfsmtest::SFSMs::isSatisfiable;
 *    return threadPool.addTask(isSatisfiableFunction, expression, identifierDefinitions).future;
 *}
 */
/*std::future<std::optional<std::unordered_map<Expression, Expression>>> asyncIsSatisfiable(Expressions const &expressions, Expressions const &queries) {
 *    auto &threadPool = ThreadPool<policies::PolicyCollection<policies::DependenciesNotRespected_t, policies::PassiveWaiting_t>>::getDefaultInstance();
 *    std::optional<std::unordered_map<Expression, Expression>> (*isSatisfiableFunction)(Expressions const &, Expressions const &) = libfsmtest::SFSMs::isSatisfiable;
 *    return threadPool.addTask(isSatisfiableFunction, expressions, queries).future;
 *}
 */

Expression makeIdentifierDefinition(IdentifierDefinition const &definition) {
    auto const &[identifier, type] = definition;
    auto const typeExpression = getSMTTypeExpression(type);
    auto const paramListExpression = makeFunctionCall({});
    return makeFunctionCall(makeIdentifier("declare-fun"), {makeIdentifier(identifier), paramListExpression, typeExpression});
}

static Expression makeFunctionDefinition_internal(Expression const &definitionExpression, std::optional<Expression> const &functionName, std::optional<IdentifierType> const &returnType, IdentifierDefinitions const &parameters, Expression const &content) {
    Expressions resultPrep;
    if(functionName.has_value()) {
        resultPrep.push_back(functionName.value());
    }
    Expressions parameterExpressions;
    std::transform(parameters.cbegin(), parameters.cend(), std::back_inserter(parameterExpressions), [](auto const &definition) {
        auto const &[name, type] = definition;
        return makeFunctionCall(makeIdentifier(name), {getSMTTypeExpression(type)});
    });
    resultPrep.push_back(makeFunctionCall(parameterExpressions));
    if(returnType.has_value()) {
        resultPrep.push_back(getSMTTypeExpression(returnType.value()));
    }
    resultPrep.push_back(content);
    return makeFunctionCall(definitionExpression, resultPrep);
}

Expression makeUniversallyQuantifiedFormula(IdentifierDefinitions const &quantifiedIdentifiers, Expression const &content) {
    return makeFunctionDefinition_internal(makeIdentifier("forall"), std::nullopt, std::nullopt, quantifiedIdentifiers, content);
}

Expression makeFunctionDefinition(Expression const &functionName, IdentifierType const &returnType, IdentifierDefinitions const &parameters, Expression const &content) {
    return makeFunctionDefinition_internal(makeIdentifier("define-fun"), functionName, returnType, parameters, content);
}

Expression makeRecursiveFunctionDefinition(Expression const &functionName, IdentifierType const &returnType, IdentifierDefinitions const &parameters, Expression const &content) {
    return makeFunctionDefinition_internal(makeIdentifier("define-fun-rec"), functionName, returnType, parameters, content);
}

Expression makeFunctionDeclaration(IdentifierDefinition const &functionDefinition) {
    auto const &[name, idType] = functionDefinition;
    IdentifierType::Type const &type = idType.getType();
    Expressions resultPrep;
    resultPrep.push_back(makeIdentifier(name));

    if(std::holds_alternative<IdentifierType::ComplexType>(type)) {
        auto const &complexType = std::get<IdentifierType::ComplexType>(type);
        Expressions parameterExpressions;
        auto const &parameters = complexType.getParameters();
        switch(complexType.getBaseType()) {
        case IdentifierType::ComplexBaseType::ARRAY:
            std::transform(parameters.cbegin(), parameters.cend(), std::back_inserter(parameterExpressions), [](auto const &typeOfParameter) {
                return getSMTTypeExpression(typeOfParameter);
            });
            resultPrep.push_back(makeFunctionCall({}));
            resultPrep.push_back(makeFunctionCall(makeIdentifier("Array"), parameterExpressions));
            break;
        case IdentifierType::ComplexBaseType::FUNCTION:
            std::transform(parameters.cbegin(), parameters.cend()-1, std::back_inserter(parameterExpressions), [](auto const &typeOfParameter) {
                return getSMTTypeExpression(typeOfParameter);
            });
            resultPrep.push_back(makeFunctionCall(parameterExpressions));
            resultPrep.push_back(getSMTTypeExpression(parameters.back()));
            break;
        default:
            throw std::runtime_error("Cannot generate function declaration.");
        }
    } else {
        auto const &simpleType = std::get<IdentifierType::SimpleType>(type);
        resultPrep.push_back(makeFunctionCall({}));
        resultPrep.push_back(getSMTTypeExpression(simpleType));
    }
    return makeFunctionCall(makeIdentifier("declare-fun"), resultPrep);
}

bool areEquivalent(Expression const &exp1, Expression const &exp2, IdentifierDefinitions const &identifierDefinitions) {
    auto assertion = makeComparisonInequality(exp1, exp2);
    return not isSatisfiable(assertion, identifierDefinitions).has_value();
}

bool isImplication(Expression const &antecedent, Expression const &consequent, IdentifierDefinitions const &identifierDefinitions) {
    auto assertion = makeLogicalConjunction(antecedent, makeLogicalNegation(consequent));
    return not isSatisfiable(assertion, identifierDefinitions).has_value();
}

}


