/*
 * Copyright 2022 Niklas Krafczyk, Jan Peleska
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#include "Sfsm.hpp"

#include <set>
#include <deque>
#include <iostream>
#include <utility>
#include <algorithm>

#include "visitors/SfsmVisitor.hpp"
#include "sfsm/State.hpp"
#include "sfsm/TransitionRelationGenerator.hpp"

using namespace std;

namespace libfsmtest::SFSMs {

Sfsm::Sfsm(std::vector<State> stateVec,
           Index initialState,
           vector<GuardCondition_t> paramInputAlphabet,
           vector<OutputExpression_t> paramOutputAlphabet,
           vector<GuardCondition_t> paramAdditionalGuards,
           vector<OutputExpression_t> paramAdditionalOutputs,
           Expression paramInvariant,
           Expression paramInitialCondition,
           IdentifierDefinitions paramIdentifiers) :
states(std::move(stateVec)),
initialStateId(initialState),
inputAlphabet(std::move(paramInputAlphabet)),
outputAlphabet(std::move(paramOutputAlphabet)),
additionalGuards(std::move(paramAdditionalGuards)),
additionalOutputs(std::move(paramAdditionalOutputs)),
invariant(std::move(paramInvariant)),
initialCondition(std::move(paramInitialCondition)),
identifiers(std::move(paramIdentifiers)) {
    if ( initialState >= states.size() )
        throw runtime_error("ERROR: Initial state number too large.");
    
}

Index Sfsm::getInitialStateId() const {
    return initialStateId;
}

IdentifierDefinitions const &Sfsm::getIdentifierDefinitions() const {
    return this->identifiers;
}

bool Sfsm::isObservable() const {
    return std::all_of(states.begin(), states.end(), [](const auto& s){
        return s.isObservable();
    });
}

bool Sfsm::isDeterministic() const {
    return std::all_of(states.begin(), states.end(), [](const auto& s){
        return s.isDeterministic();
    });
}

bool Sfsm::isCompletelySpecified() const {
    return std::all_of(states.begin(), states.end(), [&identifiers=this->identifiers, &invariant=this->invariant, &guardConditions=this->inputAlphabet](const auto& s){
        return s.isCompletelySpecified(identifiers, invariant, guardConditions);
    });
}

size_t Sfsm::size() const {
    return states.size();
}

unique_ptr<Sfsm> Sfsm::clone() const {
    return make_unique<Sfsm>(*this);
}

void Sfsm::accept(SfsmVisitor &v) const {
    v.visit(*this);
}
void Sfsm::accept(SfsmVisitor &v) {
    v.visit(*this);
}

bool Sfsm::isInitiallyConnected() const {
    
    set<Index> initiallyConnectedStates;
    
    // Initial state is always reachable
    initiallyConnectedStates.insert( initialStateId );
    
    // Breadth-first-search queue
    deque<Index> bfsQueue;
    bfsQueue.push_back(initialStateId);
    
    while ( ! bfsQueue.empty() ) {
        
        Index currentStateId = bfsQueue.front();
        bfsQueue.pop_front();
        const State& thisState = states.at(currentStateId);
        
        for ( const auto& tr : thisState.getTransitions() ) {
            Index targetStateId = tr.getTargetId();
            if ( initiallyConnectedStates.find(targetStateId) == initiallyConnectedStates.end() ) {
                initiallyConnectedStates.insert(targetStateId);
                bfsQueue.push_back(targetStateId);
            }
        }
        
    }
    
    // Initially connected iff cardinality of initiallyConnectedNodes
    // equals number of states
    return ( initiallyConnectedStates.size() == size() );
}


std::set<Index> Sfsm::getTargets(const SymbolicInputTrace& inputTrace) const {
    return this->getTargets(inputTrace, this->getInitialStateId());
}
std::set<Index> Sfsm::getTargets(const ConcreteTrace& inputTrace) const {
    return this->getTargets(inputTrace, this->getInitialStateId());
}
std::set<Index> Sfsm::getTargets(const SymbolicInputTrace& inputTrace, Index fromState) const {
    SymbolicTrace trace;
    std::transform(inputTrace.cbegin(), inputTrace.cend(), std::back_inserter(trace), [](auto const &input) -> SymbolicTrace::value_type {
        return {input, makeIdentifier("true")};
    });
    auto targets = getTargetStatesForSfsm(*this, {fromState}, trace);
    return {targets.begin(), targets.end()};
}
std::set<Index> Sfsm::getTargets(const ConcreteTrace& inputTrace, Index fromState) const {
    auto targets = getTargetStatesForSfsm(*this, {fromState}, inputTrace);
    return {targets.begin(), targets.end()};
}

std::unique_ptr<Sfsm> Sfsm::modify(std::optional<States> newStates,
                                 std::optional<Index> newInitialStateId,
                                 std::optional<std::vector<GuardCondition_t>> inputs,
                                 std::optional<std::vector<OutputExpression_t>> outputs,
                                 std::optional<std::vector<GuardCondition_t>> additionalGuardConditions,
                                 std::optional<std::vector<OutputExpression_t>> additionalOutputExpressions,
                                 std::optional<Expression> newInvariant,
                                 std::optional<Expression> newInitialCondition,
                                 std::optional<IdentifierDefinitions> newIdentifierDefinitions) const {
    if(not newStates.has_value()) {
        newStates = this->states;
    }
    if(not newInitialStateId.has_value()) {
        newInitialStateId = this->initialStateId;
    }
    if(not inputs.has_value()) {
        inputs = this->inputAlphabet;
    }
    if(not outputs.has_value()) {
        outputs = this->outputAlphabet;
    }
    if(not additionalGuardConditions.has_value()) {
        additionalGuardConditions = this->additionalGuards;
    }
    if(not additionalOutputExpressions.has_value()) {
        additionalOutputExpressions = this->additionalOutputs;
    }
    if(not newInvariant.has_value()) {
        newInvariant = this->invariant;
    }
    if(not newInitialCondition.has_value()) {
        newInitialCondition = this->initialCondition;
    }
    if(not newIdentifierDefinitions.has_value()) {
        newIdentifierDefinitions = this->identifiers;
    }
    return make_unique<Sfsm>(newStates.value(), newInitialStateId.value(), inputs.value(), outputs.value(), additionalGuardConditions.value(), additionalOutputExpressions.value(), newInvariant.value(), newInitialCondition.value(), newIdentifierDefinitions.value());
}

} // namespace libfsmtest

namespace std {
std::string to_string(libfsmtest::SFSMs::SymbolicTrace const &symbolicTrace) {
    if(symbolicTrace.empty()) {
        return "ε";
    }
    std::string result;
    bool isFirst = true;
    for(auto const &[input, output] : symbolicTrace) {
        if(not isFirst) {
            result += '.';
        }
        result += std::to_string(input);
        result += "/";
        result += std::to_string(output);
        isFirst = false;
    }
    return result;
}
std::string to_string(libfsmtest::SFSMs::SymbolicInputTrace const &symbolicTrace) {
    if(symbolicTrace.empty()) {
        return "ε";
    }
    std::string result;
    bool isFirst = true;
    for(auto const &expression : symbolicTrace) {
        if(not isFirst) {
            result += '.';
        }
        result += std::to_string(expression);
        isFirst = false;
    }
    return result;
}

std::string to_string(libfsmtest::SFSMs::StateTrace const &trace) {
    if(trace.empty()) {
        return "ε";
    }
    std::string result;
    bool isFirst = true;
    for(auto const &state : trace) {
        if(not isFirst) {
            result += '.';
        }
        result += std::to_string(state);
        isFirst = false;
    }
    return result;
}
}
