/*
 * Copyright 2022 Niklas Krafczyk, Jan Peleska
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#pragma once

#include <vector>
#include <utility>
#include <set>
#include <memory>
#include <optional>
#include <string>

#include "sfsm/Expressions.hpp"
#include "sfsm/definitions.hpp"
#include "sfsm/ExpressionDefinitions.hpp"

namespace libfsmtest::SFSMs {
class State;
class Valuation;
typedef std::vector<std::pair<libfsmtest::SFSMs::Expression, libfsmtest::SFSMs::Expression>> SymbolicTrace;
typedef std::vector<libfsmtest::SFSMs::Expression> SymbolicInputTrace;
typedef std::vector<libfsmtest::SFSMs::Expression> SymbolicOutputTrace;
typedef std::vector<Valuation> ConcreteTrace;

class SfsmVisitor;

/**
 * Most general class of symbolic finite state machines (based on Mealy machines)
 */
class Sfsm {
public:
    typedef libfsmtest::SFSMs::Expression GuardCondition_t;
    typedef libfsmtest::SFSMs::Expression OutputExpression_t;
protected:
    std::vector<State> states;
    Index initialStateId;
    std::vector<GuardCondition_t> inputAlphabet;
    std::vector<OutputExpression_t> outputAlphabet;
    std::vector<GuardCondition_t> additionalGuards;
    std::vector<OutputExpression_t> additionalOutputs;
    Expression invariant;
    Expression initialCondition;
    libfsmtest::SFSMs::IdentifierDefinitions identifiers;
public:
    Sfsm(std::vector<State> states,
         Index initialStateId,
         std::vector<GuardCondition_t> inputAlphabet,
         std::vector<OutputExpression_t> outputAlphabet,
         std::vector<GuardCondition_t> additionalGuards,
         std::vector<OutputExpression_t> additionalOutputs,
         libfsmtest::SFSMs::Expression invariant,
         libfsmtest::SFSMs::Expression initialCondition,
         libfsmtest::SFSMs::IdentifierDefinitions identifiers);

    /**
     * Copy constructor
     */
    Sfsm(const Sfsm& fsm) = default;

    /**
     * Assignment operators
     */
    Sfsm& operator=(const Sfsm& other) = default;
    Sfsm& operator=(Sfsm&& other) = default;

    /** Virtual destructor */
    virtual ~Sfsm() = default;
    
    std::vector<State>& getStates() { return states; }
    [[nodiscard]] std::vector<State> const& getStates() const { return states; }
    
    [[nodiscard]] Index getInitialStateId() const;

    [[nodiscard]] IdentifierDefinitions const &getIdentifierDefinitions() const;

    [[nodiscard]] bool isObservable() const;
    
    [[nodiscard]] bool isDeterministic() const;
    
    [[nodiscard]] bool isCompletelySpecified() const;
    
    /**
     * The size of an FSM is the number of its states.
     */
    [[nodiscard]] size_t size() const;

    /**
     * Accept any kind of FsmVisitor
     */
    virtual void accept(SfsmVisitor& v) const;
    virtual void accept(SfsmVisitor& v);

    /**
     *  This operation throws an exception, because it requires
     *  observable FSMs (Ofsm instances)
     */
    [[nodiscard]] virtual bool isInitiallyConnected() const;

    // Get symbolic traces for concrete trace
    // Get symbolic traces for symbolic trace (subset or in union?)
    // ... from state
    // ... include undefined
    
    /**
     * Collect all targets that can be reached by a given input trace.
     * @param inputTrace The input trace.
     * @return The IDs of all states that can be reached by the input trace.
     */
    [[nodiscard]] std::set<Index> getTargets(const SymbolicInputTrace& inputTrace) const;
    [[nodiscard]] std::set<Index> getTargets(const ConcreteTrace& inputTrace) const;
    
    /**
     * Collect all targets that can be reached by a given input trace
     * and a given start state.
     * @param inputTrace The input trace.
     * @param fromState The ID of the state to apply the input trace to
     * @return The IDs of all states that can be reached by the input trace.
     */
    [[nodiscard]] std::set<Index> getTargets(const SymbolicInputTrace& inputTrace, Index fromState) const;
    [[nodiscard]] std::set<Index> getTargets(const ConcreteTrace& inputTrace, Index fromState) const;

    /**
     * Return the size of the input alphabet of this fsm.
     */
    [[nodiscard]] size_t getInputAlphabetSize() const {
        return this->inputAlphabet.size() + this->additionalGuards.size();
    }

    /**
     * Return the size of the output alphabet of this fsm.
     */
    [[nodiscard]] size_t getOutputAlphabetSize() const {
        return this->outputAlphabet.size() + this->additionalOutputs.size();
    }

    [[nodiscard]] std::vector<GuardCondition_t> getGuardConditions() const {
        std::vector<GuardCondition_t> result = this->inputAlphabet;
        result.insert(result.end(), this->additionalGuards.cbegin(), this->additionalGuards.cend());
        return result;
    }

    [[nodiscard]] std::vector<OutputExpression_t> getOutputExpressions() const {
        std::vector<OutputExpression_t> result = this->outputAlphabet;
        result.insert(result.end(), this->additionalOutputs.cbegin(), this->additionalOutputs.cend());
        return result;
    }

    [[nodiscard]] std::vector<GuardCondition_t> const &getUsedGuardConditions() const {
        return this->inputAlphabet;
    }

    [[nodiscard]] std::vector<OutputExpression_t> const &getUsedOutputExpressions() const {
        return this->outputAlphabet;
    }

    [[nodiscard]] std::vector<GuardCondition_t> const &getAdditionalGuardConditions() const {
        return this->additionalGuards;
    }

    [[nodiscard]] std::vector<OutputExpression_t> const &getAdditionalOutputExpressions() const {
        return this->additionalOutputs;
    }

    [[nodiscard]] Expression getInvariant() const {
        return this->invariant;
    }

    [[nodiscard]] Expression getInitialCondition() const {
        return this->initialCondition;
    }

    /**
     * Creates a copy of the object.
     * @return A unique_ptr pointing to the created clone
     */
    [[nodiscard]] virtual std::unique_ptr<Sfsm> clone() const;

    /**
     * Creates an object of the same type as this one but with different
     * attributes.
     * @param states Optional parameter. If set, the created SFSM will have the
     *  set of states defined by the parameter as its state set. If not set, the
     *  created FSM will have the same state space as this one.
     * @param initialStateId Optional parameter. If set, the created SFSM will
     *  have the initial state set according to the given argument. If not set,
     *  the created FSM will have the same initial state ID as this one.
     * @param inputAlphabet Optional parameter. If set, the created SFSM will
     *  have the input alphabet set according to the given argument. If not set,
     *  the created SFSM will have the same input alphabet as this one.
     * @param outputAlphabet Optional parameter. If set, the created SFSM will
     *  have the output alphabet set according to the given argument. If not
     *  set, the created SFSM will have the same output alphabet as this one.
     * @param invariant Optional parameter. If set, the created SFSM will
     *  have the invariant set according to the given argument. If not
     *  set, the created SFSM will have the same invariant as this one.
     * @param initialCondition Optional parameter. If set, the created SFSM will
     *  have the initialCondition set according to the given argument. If not
     *  set, the created SFSM will have the same initialCondition as this one.
     * @param identifierDefinitions Optional parameter. If set, the created SFSM will
     *  have the identifier definitions set according to the given argument. If not
     *  set, the created SFSM will have the same identifier definitions as this one.
     */
    [[nodiscard]] virtual std::unique_ptr<Sfsm> modify(std::optional<std::vector<State>> states,
                                         std::optional<Index> initialStateId,
                                         std::optional<std::vector<GuardCondition_t>> inputAlphabet,
                                         std::optional<std::vector<OutputExpression_t>> outputAlphabet,
                                         std::optional<std::vector<GuardCondition_t>> additionalGuards,
                                         std::optional<std::vector<OutputExpression_t>> additionalOutputs,
                                         std::optional<Expression> invariant,
                                         std::optional<Expression> initialCondition,
                                         std::optional<IdentifierDefinitions> identifierDefinitions) const;
};

}

namespace std {
[[nodiscard]] std::string to_string(libfsmtest::SFSMs::SymbolicTrace const &symbolicTrace);
[[nodiscard]] std::string to_string(libfsmtest::SFSMs::SymbolicInputTrace const &symbolicTrace);
[[nodiscard]] std::string to_string(libfsmtest::SFSMs::SymbolicOutputTrace const &symbolicTrace);
[[nodiscard]] std::string to_string(libfsmtest::SFSMs::StateTrace const &trace);
}
