/*
 * Copyright 2022 Niklas Krafczyk, Jan Peleska
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#include "sfsm/State.hpp"

#include <string>
#include <set>
#include <vector>
#include <algorithm>
#include <iterator>
#include <utility>
#include <unordered_set>
#include <optional>

#include "sfsm/definitions.hpp"
#include "sfsm/Transition.hpp"
#include "sfsm/Sfsm.hpp"
#include "sfsm/Expressions.hpp"
#include "sfsm/SMTTools.hpp"
#include "visitors/SfsmVisitor.hpp"

using namespace std;
namespace libfsmtest::SFSMs {

State::State(string aStateName, Index theId, Transitions paramTransitions) :
 id(theId), name(std::move(aStateName)), transitions(std::move(paramTransitions)) { }

Transitions const& State::getTransitions() const {
    return transitions;
}

bool State::isDeterministic() const {
    
    set<Index> inputSet;
    
    // Check if more than one outgoing transition
    //  is labelled with the same input value
    for (auto& tr : transitions) {
        // Recall that inputSet.insert(tr.getInputId()).second
        // is false iff tr.getInputId() is already in the set
        if (!inputSet.insert(tr.getInputId()).second) return false;
    }
    return true;
}


bool State::isCompletelySpecified(IdentifierDefinitions const &identifiers, Expression const &invariant, std::vector<Sfsm::GuardCondition_t> const &guardConditions) const {
    vector<Expression> inputs;
    std::transform(transitions.cbegin(), transitions.cend(), std::back_inserter(inputs), [&guardConditions](auto const &tr) {
        return guardConditions.at(tr.getInputId());
    });
    auto assertion = makeLogicalConjunction(invariant, makeLogicalNegation(makeLogicalDisjunction(inputs)));
    
    return not isSatisfiable(assertion, identifiers).has_value();
}



bool State::isObservable() const {
    
    // We check that the input/output labels of outgoing transitions are unique
    set< pair<Index,Index> > labelSet;
    
    for (auto& tr : transitions) {
        // Return false if this I/O label has already been attached
        // to a previous transition
        if (!labelSet.insert(make_pair(tr.getInputId(),tr.getOutputId())).second)
            return false;
    }
    
    return true;
}

void State::accept(SfsmVisitor &v) {
    v.visit(*this);
}

std::unordered_set<Index> State::getDefinedInputs() const
{
    std::unordered_set<Index> result;
    for (auto& t : transitions)
    {
        result.insert(t.getInputId());
    }
        
    return result;
}

std::unordered_set<Index> State::getResponsesToInput(const Index x) const {
    std::unordered_set<Index> result;
    for (auto& t : transitions)
    {
        if (t.getInputId() == x)
            result.insert(t.getOutputId());
    }
        
    return result;
}

State State::modify(std::optional<std::string> newName,
                    std::optional<Index> newId,
                    std::optional<Transitions> newTransitions) const {
    if(not newName.has_value()) {
        newName = this->getName();
    }
    if(not newId.has_value()) {
        newId = this->getId();
    }
    if(not newTransitions.has_value()) {
        newTransitions = this->getTransitions();
    }
    return {*newName, *newId, *newTransitions};
}

}
