/*
 * Copyright 2022 Niklas Krafczyk, Jan Peleska
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#pragma once

#include <string>
#include <vector>
#include <unordered_set>
#include <optional>

#include "sfsm/definitions.hpp"
#include "sfsm/ExpressionDefinitions.hpp"
#include "sfsm/Expressions.hpp"
#include "sfsm/Transition.hpp"

namespace libfsmtest::SFSMs {
class SfsmVisitor;

/**
 * Class representing FSM states
 */
class State {
    
private:
    
    /** Internal state number */
    Index id;
    
    /** State name */
    std::string name;
    
    /** vector of outgoing transitions */
    std::vector<Transition> transitions;
    
    
    
public:
    
    State(std::string  aStateName, Index theId, std::vector<Transition> transitions = {});
    
    /**
     * Copy constructor
     */
    State(const State& state) = default;
    State &operator=(State const &) = default;
    
    [[nodiscard]] std::vector<Transition> const& getTransitions() const;
    
    [[nodiscard]] Index getId() const { return id; }
    
    [[nodiscard]] std::string getName() const { return name; }
    
    [[nodiscard]] bool isDeterministic() const;
    
    /**
     * Return true iff vor every input of the specified alphabet,
     * there exists a transition triggered by this input.
     * @param identifiers Identifier definitions used for satisfiability checks
     * @param invariant Invariant specifying the set of allowed valuations,
     *  potentially restricting the set of input valuations that must be
     *  covered by the guard conditions of the outgoing transitions for this
     *  state to be completely specified.
     * @param guardConditions The set of guard conditions the guard condition
     *  IDs in this state's transitions refer to.
     * @note Events of the input alphabet are internally numbered by 0..lastInputId
     */
    [[nodiscard]] bool isCompletelySpecified(IdentifierDefinitions const &identifiers, Expression const &invariant, Expressions const &guardConditions) const;
    
    /**
     * Return true if the transitions of this state are
     * observable. This means that each input/output pair
     * uniquely determines the post-state.
     */
    [[nodiscard]] bool isObservable() const;
    
    /** Accept FsmVisitors */
    void accept(SfsmVisitor& v);

    /**
     * Collect the defined inputs of this state,
     * i.e., all inputs x such that the state has 
     * an outgoing transition labelled with input x.
     */
    [[nodiscard]] std::unordered_set<Index> getDefinedInputs() const;

    /**
     * Collect all outputs y such that this state
     * has a transition with input x and output y.
     */    
    [[nodiscard]] std::unordered_set<Index> getResponsesToInput(Index x) const;

    [[nodiscard]] State modify(std::optional<std::string> aStateName = std::nullopt,
                 std::optional<Index> theId = std::nullopt,
                 std::optional<std::vector<Transition>> transitions = std::nullopt) const;
};

using States = std::vector<State>;

} // namespace libfsmtest::SFSMs

