/*
 * Copyright 2022 Niklas Krafczyk, Jan Peleska
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#pragma once

#include <vector>
#include <tuple>
#include <type_traits>
#include <cstdio>

#include "sfsm/definitions.hpp"


namespace libfsmtest::SFSMs {
class SfsmVisitor;

class Transition {
private:
    Index sourceId;
    Index inputIndex;
    Index outputIndex;
    Index targetId;
    
public:
    Transition(Index sourceId,
               Index inputIndex,
               Index outputIndex,
               Index targetId);

    Transition(Transition const &/*other*/) = default;
    Transition &operator=(Transition const &/*other*/) = default;

    [[nodiscard]] Index getSourceId() const { return sourceId; }
    [[nodiscard]] Index getTargetId() const { return targetId; }
    
    [[nodiscard]] Index getInputId() const { return inputIndex; }
    [[nodiscard]] Index getOutputId() const { return outputIndex; }
    
    /** Accept SfsmVisitors */
    void accept(SfsmVisitor& v) const;

    friend bool operator==(Transition const & t1, Transition const & t2)
    {
        return t1.inputIndex == t2.inputIndex
                && t1.outputIndex == t2.outputIndex 
                && t1.sourceId == t2.sourceId
                && t1.targetId == t2.targetId;
    }

    friend bool operator<(Transition const & t1, Transition const & t2)
    {
        if (t1.inputIndex < t2.inputIndex) return true;
        if (t1.inputIndex == t2.inputIndex) {
            if (t1.outputIndex < t2.outputIndex) return true;
            if (t1.outputIndex == t2.outputIndex) {
                if (t1.sourceId < t2.sourceId) return true;
                if (t1.sourceId == t2.sourceId) {
                    return (t1.targetId < t2.targetId);
                }
            }
        }

        return false;
    }
    
};


using Transitions = std::vector<Transition>;

template<size_t I>
size_t get(Transition const &transition) {
    static_assert(I < 4);
    if constexpr(I == 0) {
        return transition.getSourceId();
    } else if constexpr(I == 1) {
        return transition.getInputId();
    } else if constexpr(I == 2) {
        return transition.getOutputId();
    } else {
        return transition.getTargetId();
    }
}

} // namespace libfsmtest::SFSMs

namespace std {
    template<>
    struct tuple_size<libfsmtest::SFSMs::Transition>
        : std::integral_constant<size_t, 4> { };

    template<size_t Index>
    struct tuple_element<Index, libfsmtest::SFSMs::Transition>
        : tuple_element<Index, std::tuple<size_t, size_t, size_t, size_t>> { };
}

