#include "sfsm/TransitionRelationGenerator.hpp"

#include <string>
#include <algorithm>
#include <iterator>
#include <tuple>
#include <stdexcept>
#include <optional>
#include <vector>
#include <map>
#include <unordered_map>
#include <set>
#include <cassert>
#include <utility>
#include <cstdlib>
#include <charconv>

#include "sfsm/Expressions.hpp"
#include "sfsm/ExpressionDefinitions.hpp"
#include "sfsm/Transition.hpp"
#include "sfsm/SMTTools.hpp"
#include "sfsm/State.hpp"
#include "sfsm/Sfsm.hpp"
#include "sfsm/Variable.hpp"
#include "sfsm/definitions.hpp"


namespace libfsmtest::SFSMs {

std::string const PRE_STATE_IDENTIFIER = "__sfsmState";
std::string const POST_STATE_IDENTIFIER = "__sfsmState_post";
std::string const TRANSITION_RELATION_IDENTIFIER = "__transRel";
std::string const STEP_IDENTIFIER = "__n";
std::string const OUTPUT_IDENTIFIER_PRE_STATE_PREFIX = "__preStateOf_";

static IdentifierType::Type makeTraceType(IdentifierType::Type const &baseType) {
    return IdentifierType::ComplexType(IdentifierType::ComplexBaseType::FUNCTION, {IdentifierType::SimpleType(IdentifierType::SimpleBaseType::INTEGER), baseType});
}

static IdentifierDefinitions getIdentifiersByLocation(IdentifierDefinitions const &definitions, IdentifierType::LocationInModel const &location) {
    IdentifierDefinitions result;
    std::copy_if(definitions.cbegin(), definitions.cend(), std::back_inserter(result), [&location](auto const &definition) {
        return std::get<IdentifierType>(definition).getLocation() == location;
    });
    return result;
}

static void checkIdentifierNameValidity(IdentifierDefinitions const &definitions) {
    if(auto iter = std::find_if(definitions.cbegin(), definitions.cend(), [](auto const &definition) {
        auto const &[name, type] = definition;
        return name == PRE_STATE_IDENTIFIER
            or name == POST_STATE_IDENTIFIER
            or name == TRANSITION_RELATION_IDENTIFIER
            or name == STEP_IDENTIFIER
            or name == "null"
            or name.substr(0, std::string(OUTPUT_IDENTIFIER_PRE_STATE_PREFIX).size()) == OUTPUT_IDENTIFIER_PRE_STATE_PREFIX;
    }); iter != definitions.cend()) {
        std::string error = "Forbidden identifier '";
        error += std::get<0>(*iter) + "' found.";
        throw std::runtime_error(error);
    }
}

static Expression makeNullOutputExpression(IdentifierDefinitions const &definitions) {
    auto const outputs = getIdentifiersByLocation(definitions, IdentifierType::LocationInModel::OUTPUT);
    Expressions equalToPreStateExpressions;
    std::transform(outputs.cbegin(), outputs.cend(), std::back_inserter(equalToPreStateExpressions), [](auto const &outputIdentifier) {
        auto const &[identifier, _] = outputIdentifier;
        return makeComparisonEquality(makeIdentifier(identifier), makeIdentifier(OUTPUT_IDENTIFIER_PRE_STATE_PREFIX + identifier));
    });
    return makeLogicalConjunction(equalToPreStateExpressions);
}

Expression generateTransitionExpression(size_t preState, Expression const &guardCondition, Expression const &outputExpression, size_t postState) {
    auto prestateId = makeIdentifier(PRE_STATE_IDENTIFIER);
    auto poststateId = makeIdentifier(POST_STATE_IDENTIFIER);
    auto sourceIdExp = makeComparisonEquality(prestateId, makeIdentifier(std::to_string(preState)));
    auto targetIdExp = makeComparisonEquality(poststateId, makeIdentifier(std::to_string(postState)));
    return makeLogicalConjunction(sourceIdExp, guardCondition, outputExpression, targetIdExp);
}

Expression generateAbstractTransitionExpression(size_t preState, Expression const &abstractExpression, size_t postState) {
    auto prestateId = makeIdentifier(PRE_STATE_IDENTIFIER);
    auto poststateId = makeIdentifier(POST_STATE_IDENTIFIER);
    auto sourceIdExp = makeComparisonEquality(prestateId, makeIdentifier(std::to_string(preState)));
    auto targetIdExp = makeComparisonEquality(poststateId, makeIdentifier(std::to_string(postState)));
    return makeLogicalConjunction(sourceIdExp, abstractExpression, targetIdExp);
}

Expression generateTransitionRelationForTransition(IdentifierDefinitions const &definitions, Transition const &transition, Expressions const &guardConditions, Expressions const &outputExpressions) {
    checkIdentifierNameValidity(definitions);

    auto const &guardCondition = guardConditions.at(transition.getInputId());
    auto const &outputExpression = outputExpressions.at(transition.getOutputId());

    return generateTransitionExpression(transition.getSourceId(), guardCondition, outputExpression, transition.getTargetId());
}

Expressions generateAbstractTransitionRelationsForTransition(IdentifierDefinitions const &definitions, Transition const &transition, Expression const &invariant, Expressions const &guardConditions, Expressions const &outputExpressions, Expressions const &propositionEquivalenceClasses) {
    checkIdentifierNameValidity(definitions);
    Expressions abstractTransitionExpressions;

    auto guardCondition = guardConditions.at(transition.getInputId());
    auto outputExpression = outputExpressions.at(transition.getOutputId());

    for(auto const &propositionEquivalenceClass : propositionEquivalenceClasses) {
        if(isSatisfiable(makeLogicalConjunction(guardCondition, outputExpression, propositionEquivalenceClass, invariant), definitions).has_value()) {
            abstractTransitionExpressions.push_back(generateAbstractTransitionExpression(transition.getSourceId(), propositionEquivalenceClass, transition.getTargetId()));
        }
    }

    return abstractTransitionExpressions;
}

Expression generateTransitionRelationForState(IdentifierDefinitions const &definitions, State const &state, Expressions const &guardConditions, Expressions const &outputExpressions) {
    Expressions relsForTransitions;
    Expressions guardConditionsInState;
    auto const &transitions = state.getTransitions();
    std::transform(transitions.cbegin(), transitions.cend(), std::back_inserter(relsForTransitions), [&definitions,&guardConditions,&outputExpressions](auto const &transition) {
        return generateTransitionRelationForTransition(definitions, transition, guardConditions, outputExpressions);
    });
    std::transform(transitions.cbegin(), transitions.cend(), std::back_inserter(guardConditionsInState), [&guardConditions](auto const &transition) {
        return guardConditions.at(transition.getInputId());
    });
    auto const notAnyGuard = makeLogicalNegation(makeLogicalDisjunction(guardConditionsInState));
    if(isSatisfiable(notAnyGuard, definitions).has_value()) {
        auto const stateId = state.getId();
        relsForTransitions.push_back(generateTransitionExpression(stateId, notAnyGuard, makeNullOutputExpression(definitions), stateId));
    }
    return makeLogicalDisjunction(relsForTransitions);
}

Expression generateAbstractTransitionRelationForState(IdentifierDefinitions const &definitions, State const &state, Expression const &invariant, Expressions const &guardConditions, Expressions const &outputExpressions, Expressions const &propositionEquivalenceClasses) {
    Expressions relsForTransitions;
    Expressions guardConditionsInState;
    auto const &transitions = state.getTransitions();
    auto const transitionAbstractor = [&definitions,&guardConditions,&outputExpressions,&propositionEquivalenceClasses,&invariant](auto const &transition) {
        return generateAbstractTransitionRelationsForTransition(definitions, transition, invariant, guardConditions, outputExpressions, propositionEquivalenceClasses);
    };
    for(auto const &transition : transitions) {
        auto const abstractedTransitions = transitionAbstractor(transition);
        relsForTransitions.insert(relsForTransitions.end(), abstractedTransitions.begin(), abstractedTransitions.end());
    }
    std::transform(transitions.cbegin(), transitions.cend(), std::back_inserter(guardConditionsInState), [&guardConditions](auto const &transition) {
        return guardConditions.at(transition.getInputId());
    });
    auto const notAnyGuard = makeLogicalNegation(makeLogicalDisjunction(guardConditionsInState));
    if(isSatisfiable(notAnyGuard, definitions).has_value()) {
        auto const stateId = state.getId();
        relsForTransitions.push_back(generateTransitionExpression(stateId, notAnyGuard, makeNullOutputExpression(definitions), stateId));
    }
    return makeLogicalDisjunction(relsForTransitions);
}

Expression generateTransitionRelationForSfsm(Sfsm const &sfsm) {
    Expressions relsForStates;
    IdentifierDefinitions const &definitions = sfsm.getIdentifierDefinitions();
    auto const &states = sfsm.getStates();
    auto const &guardConditions = sfsm.getGuardConditions();
    auto const &outputExpressions = sfsm.getOutputExpressions();
    std::transform(states.cbegin(), states.cend(), std::back_inserter(relsForStates), [&definitions,&guardConditions,&outputExpressions](auto const &state) {
        return generateTransitionRelationForState(definitions, state, guardConditions, outputExpressions);
    });
    return makeLogicalConjunction(makeLogicalDisjunction(relsForStates), sfsm.getInvariant());
}

Expression generateAbstractTransitionRelationForSfsm(Sfsm const &sfsm, Expressions const &propositionEquivalenceClasses) {
    Expressions relsForStates;
    IdentifierDefinitions const &definitions = sfsm.getIdentifierDefinitions();
    auto const &states = sfsm.getStates();
    auto const &guardConditions = sfsm.getGuardConditions();
    auto const &outputExpressions = sfsm.getOutputExpressions();
    auto const &invariant = sfsm.getInvariant();
    std::transform(states.cbegin(), states.cend(), std::back_inserter(relsForStates), [&definitions,&guardConditions,&outputExpressions,&propositionEquivalenceClasses,&invariant](auto const &state) {
        return generateAbstractTransitionRelationForState(definitions, state, invariant, guardConditions, outputExpressions, propositionEquivalenceClasses);
    });
    return makeLogicalConjunction(makeLogicalDisjunction(relsForStates), sfsm.getInvariant());
}

Expression generateTransitionRelationCall(Sfsm const &sfsm, Expressions const &arguments) {
    auto const numIdentifiers = sfsm.getIdentifierDefinitions().size();
    auto const numOutputs = getIdentifiersByLocation(sfsm.getIdentifierDefinitions(), IdentifierType::LocationInModel::OUTPUT).size();
    if(arguments.size() !=  numIdentifiers + numOutputs + 2) {
        std::string error = "Incorrect number of arguments for transition relation. Must be exactly the number of identifiers in the SFSM plus two for step number and state trace. Expected: ";
        error += std::to_string(sfsm.getIdentifierDefinitions().size() + 2);
        error += ", Received: ";
        error += std::to_string(arguments.size());
        throw std::runtime_error(error);
    }
    return makeFunctionCall(makeIdentifier(TRANSITION_RELATION_IDENTIFIER), arguments);
}

Expression generateTransitionRelationDefinitionForSfsm(Sfsm const &sfsm) {
    auto const relation = generateTransitionRelationForSfsm(sfsm);
    IdentifierDefinitions parameters = {
        {PRE_STATE_IDENTIFIER,
            {IdentifierType::SimpleType{IdentifierType::SimpleBaseType::INTEGER},
             IdentifierType::LocationInModel::OTHER}},
        {POST_STATE_IDENTIFIER,
            {IdentifierType::SimpleType{IdentifierType::SimpleBaseType::INTEGER},
             IdentifierType::LocationInModel::OTHER}}};
    auto const &sfsmIdentifiers = sfsm.getIdentifierDefinitions();
    for(auto const &identifier : sfsmIdentifiers) {
        auto const &[name, type] = identifier;
        parameters.insert(parameters.end(), identifier);
        if(type.getLocation() == IdentifierType::LocationInModel::OUTPUT) {
            auto const newIdentifierForPreState = IdentifierDefinition(std::string("__preStateOf_") + name, type);
            parameters.insert(parameters.end(), newIdentifierForPreState);
        }
    }

    return makeFunctionDefinition(makeIdentifier(TRANSITION_RELATION_IDENTIFIER), 
    {IdentifierType::SimpleType(IdentifierType::SimpleBaseType::BOOLEAN), IdentifierType::LocationInModel::OTHER}, parameters, relation);
}

Expression generateAbstractTransitionRelationDefinitionForSfsm(Sfsm const &sfsm, Expressions const &propositionEquivalenceClasses) {
    auto const relation = generateAbstractTransitionRelationForSfsm(sfsm, propositionEquivalenceClasses);
    IdentifierDefinitions parameters = {
        {PRE_STATE_IDENTIFIER,
            {IdentifierType::SimpleType{IdentifierType::SimpleBaseType::INTEGER},
             IdentifierType::LocationInModel::OTHER}},
        {POST_STATE_IDENTIFIER,
            {IdentifierType::SimpleType{IdentifierType::SimpleBaseType::INTEGER},
             IdentifierType::LocationInModel::OTHER}}};
    auto const &sfsmIdentifiers = sfsm.getIdentifierDefinitions();
    for(auto const &identifier : sfsmIdentifiers) {
        auto const &[name, type] = identifier;
        parameters.insert(parameters.end(), identifier);
        if(type.getLocation() == IdentifierType::LocationInModel::OUTPUT) {
            auto const newIdentifierForPreState = IdentifierDefinition(std::string("__preStateOf_") + name, type);
            parameters.insert(parameters.end(), newIdentifierForPreState);
        }
    }

    return makeFunctionDefinition(makeIdentifier(TRANSITION_RELATION_IDENTIFIER), 
    {IdentifierType::SimpleType(IdentifierType::SimpleBaseType::BOOLEAN), IdentifierType::LocationInModel::OTHER}, parameters, relation);
}

static std::map<Identifier, Expression> allocateTransitionRelationArgumentIdentifierInstantiations(Sfsm const &sfsm, std::map<Identifier, Expression> hints) {
    auto const &definitions = sfsm.getIdentifierDefinitions();
    for(auto const &[identifier, _] : definitions) {
        if(hints.count(identifier) == 0) {
            // Using hashes as identifiers so the chance of randomly hitting
            // an identifier existing in some model is minor, even if they
            // employ the same naming scheme.
            hints[identifier] = makeNewIdentifier(identifier);
        }
    }
    return hints;
}

std::pair<std::unordered_map<IdentifierDefinition, IdentifierDefinition>, Expressions> findDistinguishingTraceProblem(Sfsm const &sfsm, Expression const &transitionRelationDefinition, Index state1, Index state2) {
    //TODO: Wir brauchen die zweite Trace in der Transition relation nicht. Es reicht, eine der beiden zu haben und beim jeweils anderen state zu fordern, dass er das nicht kann.
    std::set<Expression> declarations;
    Expressions problemStatements;
    auto const &definitions = sfsm.getIdentifierDefinitions();
    size_t const upperSearchBound = sfsm.size();

    // Get inputs and outputs
    IdentifierDefinitions inputs = getIdentifiersByLocation(definitions, IdentifierType::LocationInModel::INPUT);
    IdentifierDefinitions outputs = getIdentifiersByLocation(definitions, IdentifierType::LocationInModel::OUTPUT);
    //TODO: Make this a usercheck
    assert(not inputs.empty() and not outputs.empty());

    // Allocate distinct identifiers for all identifiers in the model
    auto traceIdentifiers1 = allocateTransitionRelationArgumentIdentifierInstantiations(sfsm, {});

    // Allocate two distinct identifiers for state functions
    auto stateFunction1 = makeNewIdentifier("state1");
    auto stateFunction2 = makeNewIdentifier("state2");
    declarations.insert(declarations.end(), makeFunctionDefinition(stateFunction1, {IdentifierType::SimpleType(IdentifierType::SimpleBaseType::INTEGER), IdentifierType::LocationInModel::OTHER}, {}, makeIdentifier(std::to_string(state1))));
    declarations.insert(declarations.end(), makeFunctionDefinition(stateFunction2, {IdentifierType::SimpleType(IdentifierType::SimpleBaseType::INTEGER), IdentifierType::LocationInModel::OTHER}, {}, makeIdentifier(std::to_string(state2))));

    auto stateFunctionGeneral = makeNewIdentifier("state");
    std::map<Expression, IdentifierType> argumentIdentifiers1 = { 
        {stateFunctionGeneral, {makeTraceType(IdentifierType::SimpleType(IdentifierType::SimpleBaseType::INTEGER)), IdentifierType::LocationInModel::OTHER}} };
    std::transform(definitions.cbegin(), definitions.cend(), std::inserter(argumentIdentifiers1, argumentIdentifiers1.end()), [&traceIdentifiers1](auto const &definition) {
        auto const &[identifier, type] = definition;
        return std::make_pair(traceIdentifiers1.at(identifier), IdentifierType(makeTraceType(type.getType()), type.getLocation()));
    });

    // Collect all identifiers allocated for this problem and create a declaration expression
    std::map<Expression, IdentifierType> toDeclare = argumentIdentifiers1;
    //toDeclare.insert(argumentIdentifiers2.cbegin(), argumentIdentifiers2.cend());

    std::transform(toDeclare.cbegin(), toDeclare.cend(), std::inserter(declarations, declarations.end()), [](auto const &declare) {
        auto const &[expr, type] = declare;
        return makeFunctionDeclaration({std::to_string(expr), type});
    });
    declarations.insert(transitionRelationDefinition);
    // Create step identifier declaration
    declarations.insert(makeFunctionDeclaration({STEP_IDENTIFIER, 
        {IdentifierType::SimpleType(IdentifierType::SimpleBaseType::INTEGER), IdentifierType::LocationInModel::OTHER}}));

    // Initial constraints
    auto const initialState = makeFunctionCall(stateFunctionGeneral, {makeIdentifier("0")});
    problemStatements.push_back(makeLogicalDisjunction(makeComparisonEquality(initialState, stateFunction1),
                                                       makeComparisonEquality(initialState, stateFunction2)));

    // Pre- and post states for state1 and state2
    auto const preStateGeneral = makeFunctionCall({stateFunctionGeneral, makeIdentifier(STEP_IDENTIFIER)});
    auto const postStateGeneral = makeFunctionCall({stateFunctionGeneral, makeArithmeticAddition(makeIdentifier(STEP_IDENTIFIER), makeIdentifier("1"))});

    // Create a transition relation call with all identifiers allocated for the first trace
    Expressions transRelCallArguments = {preStateGeneral, postStateGeneral};

    for(auto const &definition : definitions) {
        auto const &[identifier, type] = definition;
        auto const mappedIdentifier = traceIdentifiers1.at(identifier);
        transRelCallArguments.push_back(makeFunctionCall({mappedIdentifier, makeIdentifier(STEP_IDENTIFIER)}));
        if(type.getLocation() == IdentifierType::LocationInModel::OUTPUT) {
            transRelCallArguments.push_back(makeFunctionCall({mappedIdentifier, makeArithmeticSubtraction(makeIdentifier(STEP_IDENTIFIER), makeIdentifier("1"))}));
        }
    }

    auto transRelCall = generateTransitionRelationCall(sfsm, transRelCallArguments);

    Expressions greaterThanStepConditions;
    for(size_t step = 0; step < upperSearchBound; ++step) {
        greaterThanStepConditions.push_back(makeComparisonGreaterOrEquals(makeIdentifier(STEP_IDENTIFIER), makeIdentifier(std::to_string(step))));
    }
    
    // Call transition relation k times, where k is the number of states
    // TODO: Instead of iterating up to sfsm.getSize(), iterate up to longest shortest trace reaching a state.
    for(size_t callIndex = 0; callIndex < sfsm.size(); ++callIndex) {
        // (n >= callIndex) => (transRel (state n) (state (+ n 1)) ...
        auto condition = greaterThanStepConditions.at(callIndex);
        auto relation = replaceSubexpressions(transRelCall, {{makeIdentifier(STEP_IDENTIFIER), makeIdentifier(std::to_string(callIndex))}});
        problemStatements.push_back(makeLogicalImplication(condition, relation));
    }

    // Trace for one starting state shall not be possible for any target state for the other starting state
    // universally quantified state identifiers
    Expressions universallyQuantifiedStateIdentifiers;
    IdentifierDefinitions uqualStateIdentifierDefinitions;
    for(size_t step = 0; step < upperSearchBound; ++step) {
        universallyQuantifiedStateIdentifiers.push_back(makeNewIdentifier(std::string("__uq_state") + std::to_string(step)));
    }
    std::transform(universallyQuantifiedStateIdentifiers.begin(),
                   universallyQuantifiedStateIdentifiers.end(),
                   std::back_inserter(uqualStateIdentifierDefinitions),
                   [](auto const &identifier) {
        return IdentifierDefinition(std::to_string(identifier), IdentifierType(IdentifierType::SimpleType{IdentifierType::SimpleBaseType::INTEGER}, IdentifierType::LocationInModel::OTHER));
    });

    // Not possible from state1
    Expressions possibleInStepNFromOtherState;
    for(size_t step = 0; step < upperSearchBound; ++step) {
        Expression preState;
        Expression postState = universallyQuantifiedStateIdentifiers.at(step);
        if(step == 0) {
            preState = makeIfThenElse(makeComparisonEquality(makeFunctionCall({stateFunctionGeneral, makeIdentifier("0")}), stateFunction1), stateFunction2, stateFunction1);
        } else {
            preState = universallyQuantifiedStateIdentifiers.at(step - 1);
        }
        auto transRelCallWithChangedStates = replaceSubexpressions(transRelCall, {{preStateGeneral, preState}, {postStateGeneral, postState},
                                                                                   {makeIdentifier(STEP_IDENTIFIER), makeIdentifier(std::to_string(step))}});
        possibleInStepNFromOtherState.push_back(makeLogicalImplication(greaterThanStepConditions.at(step), transRelCallWithChangedStates));
    }
    auto traceNotPossibleInState = makeUniversallyQuantifiedFormula(uqualStateIdentifierDefinitions,
                                                                     makeLogicalNegation(makeLogicalConjunction(possibleInStepNFromOtherState)));
    problemStatements.push_back(traceNotPossibleInState);
    
    // As we are guaranteed to be able to reach all reachable states of an SFSM
    // with n states in at most n steps, we can specify an upper bound
    // TODO: Use longest shortest trace length from above
    problemStatements.push_back(makeComparisonLessOrEquals(makeIdentifier(STEP_IDENTIFIER), makeIdentifier(std::to_string(sfsm.getStates().size()))));
    problemStatements.push_back(makeComparisonGreaterOrEquals(makeIdentifier(STEP_IDENTIFIER), makeIdentifier(std::to_string(0))));

    auto problemStatement = makeFunctionCall(makeIdentifier("assert"), {makeLogicalConjunction(problemStatements)});
    std::unordered_map<IdentifierDefinition, IdentifierDefinition> inputIdentifiers;
    std::transform(inputs.cbegin(), inputs.cend(), std::inserter(inputIdentifiers, inputIdentifiers.end()), [&traceIdentifiers1](auto const &input) {
        auto const &[name, type] = input;
        return std::pair<IdentifierDefinition, IdentifierDefinition>(input, IdentifierDefinition(std::to_string(traceIdentifiers1.at(name)), type));
    });
    std::transform(outputs.cbegin(), outputs.cend(), std::inserter(inputIdentifiers, inputIdentifiers.end()), [&traceIdentifiers1](auto const &input) {
        auto const &[name, type] = input;
        return std::pair<IdentifierDefinition, IdentifierDefinition>(input, IdentifierDefinition(std::to_string(traceIdentifiers1.at(name)), type));
    });

    // return the results
    Expressions result(declarations.cbegin(), declarations.cend());
    result.push_back(problemStatement);
    return {inputIdentifiers, result};
}

std::pair<std::unordered_map<IdentifierDefinition, IdentifierDefinition>, Expressions> findDistinguishingTraceProblem(Sfsm const &sfsm, Expression const &transitionRelationDefinition, std::vector<Index> const &stateSet1, std::vector<Index> const &stateSet2) {
    std::set<Expression> declarations;
    Expressions problemStatements;
    auto const &definitions = sfsm.getIdentifierDefinitions();
    size_t const upperSearchBound = sfsm.size();

    // Get inputs and outputs
    IdentifierDefinitions inputs = getIdentifiersByLocation(definitions, IdentifierType::LocationInModel::INPUT);
    IdentifierDefinitions outputs = getIdentifiersByLocation(definitions, IdentifierType::LocationInModel::OUTPUT);
    //TODO: Make this a usercheck
    assert(not inputs.empty() and not outputs.empty());

    // Allocate distinct identifiers for all identifiers in the model
    auto traceIdentifiers1 = allocateTransitionRelationArgumentIdentifierInstantiations(sfsm, {});

    // Allocate two distinct identifiers for state functions
    auto stateFunction1 = makeNewIdentifier("state1");
    auto stateFunction2 = makeNewIdentifier("state2");
    Expression stateFunction1Definition = makeIdentifier(std::to_string(stateSet1.at(0)));
    for(size_t stateIndex = 1; stateIndex < stateSet1.size(); ++stateIndex) {
        auto condition = makeComparisonEquality(makeIdentifier("idx"), makeIdentifier(std::to_string(stateIndex)));
        stateFunction1Definition = makeIfThenElse(condition, makeIdentifier(std::to_string(stateSet1.at(stateIndex))), stateFunction1Definition);
    }
    Expression stateFunction2Definition = makeIdentifier(std::to_string(stateSet2.at(0)));
    for(size_t stateIndex = 1; stateIndex < stateSet2.size(); ++stateIndex) {
        auto condition = makeComparisonEquality(makeIdentifier("idx"), makeIdentifier(std::to_string(stateIndex)));
        stateFunction1Definition = makeIfThenElse(condition, makeIdentifier(std::to_string(stateSet2.at(stateIndex))), stateFunction2Definition);
    }
    IdentifierDefinition stateFunctionIndexIdentifier{"idx", {IdentifierType::SimpleType(IdentifierType::SimpleBaseType::INTEGER), IdentifierType::LocationInModel::OTHER}};
    declarations.insert(declarations.end(), makeFunctionDefinition(stateFunction1, {IdentifierType::SimpleType(IdentifierType::SimpleBaseType::INTEGER), IdentifierType::LocationInModel::OTHER}, {stateFunctionIndexIdentifier}, stateFunction1Definition));
    declarations.insert(declarations.end(), makeFunctionDefinition(stateFunction2, {IdentifierType::SimpleType(IdentifierType::SimpleBaseType::INTEGER), IdentifierType::LocationInModel::OTHER}, {stateFunctionIndexIdentifier}, stateFunction2Definition));
    
    auto stateFunction1Index = makeNewIdentifier("state1Index");
    auto stateFunction2Index = makeNewIdentifier("state2Index");
    auto stateSetSelection = makeNewIdentifier("stateSet");

    auto stateFunctionGeneral = makeNewIdentifier("state");
    std::map<Expression, IdentifierType> argumentIdentifiers1 = { 
        {stateFunctionGeneral, {makeTraceType(IdentifierType::SimpleType(IdentifierType::SimpleBaseType::INTEGER)), IdentifierType::LocationInModel::OTHER}} };
    std::transform(definitions.cbegin(), definitions.cend(), std::inserter(argumentIdentifiers1, argumentIdentifiers1.end()), [&traceIdentifiers1](auto const &definition) {
        auto const &[identifier, type] = definition;
        return std::make_pair(traceIdentifiers1.at(identifier), IdentifierType(makeTraceType(type.getType()), type.getLocation()));
    });

    // Collect all identifiers allocated for this problem and create a declaration expression
    std::map<Expression, IdentifierType> toDeclare = argumentIdentifiers1;

    assert(toDeclare.count(stateFunction1Index) == 0);
    assert(toDeclare.count(stateFunction2Index) == 0);
    assert(toDeclare.count(stateSetSelection) == 0);
    toDeclare.insert(toDeclare.end(), std::make_pair(stateFunction1Index, IdentifierType{IdentifierType::SimpleType(IdentifierType::SimpleBaseType::INTEGER), IdentifierType::LocationInModel::OTHER}));
    toDeclare.insert(toDeclare.end(), std::make_pair(stateFunction2Index, IdentifierType{IdentifierType::SimpleType(IdentifierType::SimpleBaseType::INTEGER), IdentifierType::LocationInModel::OTHER}));
    toDeclare.insert(toDeclare.end(), std::make_pair(stateSetSelection, IdentifierType{IdentifierType::SimpleType(IdentifierType::SimpleBaseType::INTEGER), IdentifierType::LocationInModel::OTHER}));

    std::transform(toDeclare.cbegin(), toDeclare.cend(), std::inserter(declarations, declarations.end()), [](auto const &declare) {
        auto const &[expr, type] = declare;
        return makeFunctionDeclaration({std::to_string(expr), type});
    });
    declarations.insert(transitionRelationDefinition);
    // Create step identifier declaration
    declarations.insert(makeFunctionDeclaration({STEP_IDENTIFIER, 
        {IdentifierType::SimpleType(IdentifierType::SimpleBaseType::INTEGER), IdentifierType::LocationInModel::OTHER}}));
    declarations.insert(makeFunctionDeclaration({STEP_IDENTIFIER, 
        {IdentifierType::SimpleType(IdentifierType::SimpleBaseType::INTEGER), IdentifierType::LocationInModel::OTHER}}));

    // Initial constraints
    auto const initialState = makeFunctionCall(stateFunctionGeneral, {makeIdentifier("0")});
    auto selectedState1 = makeFunctionCall({stateFunction1, stateFunction1Index});
    auto selectedState2 = makeFunctionCall({stateFunction2, stateFunction2Index});
    problemStatements.push_back(makeIfThenElse(makeComparisonEquality(stateSetSelection, makeIdentifier("0")),
                                                makeComparisonEquality(initialState, selectedState1),
                                                makeComparisonEquality(initialState, selectedState2)));

    // Pre- and post states for state1 and state2
    auto const preStateGeneral = makeFunctionCall({stateFunctionGeneral, makeIdentifier(STEP_IDENTIFIER)});
    auto const postStateGeneral = makeFunctionCall({stateFunctionGeneral, makeArithmeticAddition(makeIdentifier(STEP_IDENTIFIER), makeIdentifier("1"))});

    // Create a transition relation call with all identifiers allocated for the first trace
    Expressions transRelCallArguments = {preStateGeneral, postStateGeneral};

    for(auto const &definition : definitions) {
        auto const &[identifier, type] = definition;
        auto const mappedIdentifier = traceIdentifiers1.at(identifier);
        transRelCallArguments.push_back(makeFunctionCall({mappedIdentifier, makeIdentifier(STEP_IDENTIFIER)}));
        if(type.getLocation() == IdentifierType::LocationInModel::OUTPUT) {
            transRelCallArguments.push_back(makeFunctionCall({mappedIdentifier, makeArithmeticSubtraction(makeIdentifier(STEP_IDENTIFIER), makeIdentifier("1"))}));
        }
    }

    auto transRelCall = generateTransitionRelationCall(sfsm, transRelCallArguments);

    Expressions greaterThanStepConditions;
    for(size_t step = 0; step < upperSearchBound; ++step) {
        greaterThanStepConditions.push_back(makeComparisonGreaterOrEquals(makeIdentifier(STEP_IDENTIFIER), makeIdentifier(std::to_string(step))));
    }
    
    // Call transition relation k times, where k is the number of states
    // TODO: Instead of iterating up to sfsm.getSize(), iterate up to longest shortest trace reaching a state.
    for(size_t callIndex = 0; callIndex < sfsm.size(); ++callIndex) {
        // (n >= callIndex) => (transRel (state n) (state (+ n 1)) ...
        auto condition = greaterThanStepConditions.at(callIndex);
        auto relation = replaceSubexpressions(transRelCall, {{makeIdentifier(STEP_IDENTIFIER), makeIdentifier(std::to_string(callIndex))}});
        problemStatements.push_back(makeLogicalImplication(condition, relation));
    }

    // Trace for one starting state shall not be possible for any target state for the other starting state
    // universally quantified state identifiers
    Expressions universallyQuantifiedStateIdentifiers;
    IdentifierDefinitions uqualStateIdentifierDefinitions;
    for(size_t step = 0; step < upperSearchBound; ++step) {
        universallyQuantifiedStateIdentifiers.push_back(makeNewIdentifier(std::string("__uq_state") + std::to_string(step)));
    }
    universallyQuantifiedStateIdentifiers.push_back(makeNewIdentifier(std::string("__uq_initStateIndex")));
    std::transform(universallyQuantifiedStateIdentifiers.begin(),
                   universallyQuantifiedStateIdentifiers.end(),
                   std::back_inserter(uqualStateIdentifierDefinitions),
                   [](auto const &identifier) {
        return IdentifierDefinition(std::to_string(identifier), IdentifierType(IdentifierType::SimpleType{IdentifierType::SimpleBaseType::INTEGER}, IdentifierType::LocationInModel::OTHER));
    });

    // Not possible from state1
    Expressions possibleInStepNFromOtherState;
    for(size_t step = 0; step < upperSearchBound; ++step) {
        Expression preState;
        Expression postState = universallyQuantifiedStateIdentifiers.at(step);
        if(step == 0) {
            auto const condition = makeComparisonEquality(stateSetSelection, makeIdentifier("0"));
            auto const stateIndex = universallyQuantifiedStateIdentifiers.back();
            preState = makeIfThenElse(condition, makeFunctionCall({stateFunction2, stateIndex}), makeFunctionCall({stateFunction1, stateIndex}));
        } else {
            preState = universallyQuantifiedStateIdentifiers.at(step - 1);
        }
        auto transRelCallWithChangedStates = replaceSubexpressions(transRelCall, {{preStateGeneral, preState}, {postStateGeneral, postState},
                                                                                   {makeIdentifier(STEP_IDENTIFIER), makeIdentifier(std::to_string(step))}});
        possibleInStepNFromOtherState.push_back(makeLogicalImplication(greaterThanStepConditions.at(step), transRelCallWithChangedStates));
    }
    auto traceNotPossibleInState = makeUniversallyQuantifiedFormula(uqualStateIdentifierDefinitions,
                                                                     makeLogicalNegation(makeLogicalConjunction(possibleInStepNFromOtherState)));
    problemStatements.push_back(traceNotPossibleInState);
    
    // As we are guaranteed to be able to reach all reachable states of an SFSM
    // with n states in at most n steps, we can specify an upper bound
    // TODO: Use longest shortest trace length from above
    problemStatements.push_back(makeComparisonLessOrEquals(makeIdentifier(STEP_IDENTIFIER), makeIdentifier(std::to_string(sfsm.getStates().size()))));
    problemStatements.push_back(makeComparisonGreaterOrEquals(makeIdentifier(STEP_IDENTIFIER), makeIdentifier(std::to_string(0))));

    auto problemStatement = makeFunctionCall(makeIdentifier("assert"), {makeLogicalConjunction(problemStatements)});
    std::unordered_map<IdentifierDefinition, IdentifierDefinition> inputIdentifiers;
    std::transform(inputs.cbegin(), inputs.cend(), std::inserter(inputIdentifiers, inputIdentifiers.end()), [&traceIdentifiers1](auto const &input) {
        auto const &[name, type] = input;
        return std::pair<IdentifierDefinition, IdentifierDefinition>(input, IdentifierDefinition(std::to_string(traceIdentifiers1.at(name)), type));
    });
    std::transform(outputs.cbegin(), outputs.cend(), std::inserter(inputIdentifiers, inputIdentifiers.end()), [&traceIdentifiers1](auto const &input) {
        auto const &[name, type] = input;
        return std::pair<IdentifierDefinition, IdentifierDefinition>(input, IdentifierDefinition(std::to_string(traceIdentifiers1.at(name)), type));
    });

    // return the results
    Expressions result(declarations.cbegin(), declarations.cend());
    result.push_back(problemStatement);
    return {inputIdentifiers, result};
}

std::optional<ConcreteTrace> getDistinguishingTrace(Sfsm const &sfsm, Expression const &transitionRelationDefinition, Index state1, Index state2) {
    if(state1 == state2) {
        return std::nullopt;
    }
    auto [identifiers, problemFormula] = findDistinguishingTraceProblem(sfsm, transitionRelationDefinition, state1, state2);
    size_t upperBound = sfsm.size();

    auto distinguishingStep = makeIdentifier(STEP_IDENTIFIER);
    Expressions inputIdentifierExpressions;
    for(auto const &[_, newInput] : identifiers) {
        auto const &[name, type] = newInput;
        inputIdentifierExpressions.push_back(makeIdentifier(name));
    }
    std::map<Expression, Expressions> identifierQueriesByIdentifier;
    std::transform(inputIdentifierExpressions.cbegin(), inputIdentifierExpressions.cend(), std::inserter(identifierQueriesByIdentifier, identifierQueriesByIdentifier.end()),
                    [upperBound](auto const &identifier) {
        Expressions queriesForSteps;
        for(size_t step = 0; step < upperBound; ++step) {
            queriesForSteps.push_back(makeFunctionCall({identifier, makeIdentifier(std::to_string(step))}));
        }
        return std::pair<Expression, Expressions>{identifier, queriesForSteps};
    });
    Expressions expressionsToEvaluate { distinguishingStep };
    for(auto const &[identifier, expressions] : identifierQueriesByIdentifier) {
        expressionsToEvaluate.insert(expressionsToEvaluate.end(), expressions.cbegin(), expressions.cend());
    }
    Expressions queriesToSolver;
    std::transform(expressionsToEvaluate.cbegin(), expressionsToEvaluate.cend(), std::back_inserter(queriesToSolver), [](auto const &expression) {
        return makeFunctionCall({makeIdentifier("get-value"), makeFunctionCall({expression})});
    });
#ifndef NDEBUG
    auto const verboseSolver_envvar = getenv("LIBSFSMTEST_VERBOSE_SOLVER");
    bool const verboseSolver = verboseSolver_envvar != nullptr and (std::string("1") == verboseSolver_envvar);
    if(verboseSolver) {
        queriesToSolver.push_back(makeFunctionCall({makeIdentifier("get-model")}));
    }
#endif

    auto valuations = isSatisfiable(problemFormula,
                                    queriesToSolver);

    if(not valuations.has_value()) {
        return std::nullopt;
    }
    assert(valuations->size() == queriesToSolver.size());
    auto extractValue = [](auto const &expr) {
        auto const innerPart = splitFunctionCall(expr).value().front();
        auto const innerSplit = splitFunctionCall(innerPart).value();
        assert(innerSplit.size() == 2);
        return innerSplit.back();
    };
    auto distinguishingStepValue = valuations->at(queriesToSolver.front());
    auto splitDistinguishingStepValue = extractValue(distinguishingStepValue);
    auto splitDistinguishingStepAsString = std::to_string(splitDistinguishingStepValue);
    size_t distinguishingStepInteger = 0;
    auto conversionErrors = std::from_chars(splitDistinguishingStepAsString.c_str(), splitDistinguishingStepAsString.c_str() + splitDistinguishingStepAsString.size(), distinguishingStepInteger);
    assert(conversionErrors.ptr == splitDistinguishingStepAsString.c_str() + splitDistinguishingStepAsString.size());

    ConcreteTrace result;

    for(size_t step = 0; step <= distinguishingStepInteger; ++step) {
        result.emplace_back();
        Valuation &stepValues = result.back();
        for(auto const &[originalInput, newInput] : identifiers) {
            auto const inputNameExpression = makeIdentifier(std::get<std::string>(newInput));
            auto const valueToLookUp = makeFunctionCall({inputNameExpression, makeIdentifier(std::to_string(step))});
            auto const getValueExpression = makeFunctionCall({makeIdentifier("get-value"), makeFunctionCall({valueToLookUp})});
            auto value = extractValue(valuations->at(getValueExpression));
            stepValues = stepValues.modify(originalInput, std::to_string(value));
        }
        assert(not stepValues.getValuations().empty());
    }
    assert(not result.empty());
    return result;
}

std::optional<ConcreteTrace> getDistinguishingTrace(Sfsm const &sfsm, Expression const &transitionRelationDefinition, std::vector<Index> state1, std::vector<Index> state2) {
    std::sort(state1.begin(), state1.end());
    std::sort(state2.begin(), state2.end());
    if(state1 == state2) {
        return std::nullopt;
    }
    auto [identifiers, problemFormula] = findDistinguishingTraceProblem(sfsm, transitionRelationDefinition, state1, state2);
    size_t upperBound = sfsm.size();

    auto distinguishingStep = makeIdentifier(STEP_IDENTIFIER);
    Expressions inputIdentifierExpressions;
    for(auto const &[_, newInput] : identifiers) {
        auto const &[name, type] = newInput;
        inputIdentifierExpressions.push_back(makeIdentifier(name));
    }
    std::map<Expression, Expressions> identifierQueriesByIdentifier;
    std::transform(inputIdentifierExpressions.cbegin(), inputIdentifierExpressions.cend(), std::inserter(identifierQueriesByIdentifier, identifierQueriesByIdentifier.end()),
                    [upperBound](auto const &identifier) {
        Expressions queriesForSteps;
        for(size_t step = 0; step < upperBound; ++step) {
            queriesForSteps.push_back(makeFunctionCall({identifier, makeIdentifier(std::to_string(step))}));
        }
        return std::pair<Expression, Expressions>{identifier, queriesForSteps};
    });
    Expressions expressionsToEvaluate { distinguishingStep };
    for(auto const &[identifier, expressions] : identifierQueriesByIdentifier) {
        expressionsToEvaluate.insert(expressionsToEvaluate.end(), expressions.cbegin(), expressions.cend());
    }
    Expressions queriesToSolver;
    std::transform(expressionsToEvaluate.cbegin(), expressionsToEvaluate.cend(), std::back_inserter(queriesToSolver), [](auto const &expression) {
        return makeFunctionCall({makeIdentifier("get-value"), makeFunctionCall({expression})});
    });
#ifndef NDEBUG
    auto const verboseSolver_envvar = getenv("LIBSFSMTEST_VERBOSE_SOLVER");
    bool const verboseSolver = verboseSolver_envvar != nullptr and (std::string("1") == verboseSolver_envvar);
    if(verboseSolver) {
        queriesToSolver.push_back(makeFunctionCall({makeIdentifier("get-model")}));
    }
#endif

    auto valuations = isSatisfiable(problemFormula,
                                    queriesToSolver);

    if(not valuations.has_value()) {
        return std::nullopt;
    }
    assert(valuations->size() == queriesToSolver.size());
    auto extractValue = [](auto const &expr) {
        auto const innerPart = splitFunctionCall(expr).value().front();
        auto const innerSplit = splitFunctionCall(innerPart).value();
        assert(innerSplit.size() == 2);
        return innerSplit.back();
    };
    auto distinguishingStepValue = valuations->at(queriesToSolver.front());
    auto splitDistinguishingStepValue = extractValue(distinguishingStepValue);
    auto splitDistinguishingStepAsString = std::to_string(splitDistinguishingStepValue);
    size_t distinguishingStepInteger = 0;
    auto conversionErrors = std::from_chars(splitDistinguishingStepAsString.c_str(), splitDistinguishingStepAsString.c_str() + splitDistinguishingStepAsString.size(), distinguishingStepInteger);
    assert(conversionErrors.ptr == splitDistinguishingStepAsString.c_str() + splitDistinguishingStepAsString.size());

    ConcreteTrace result;

    for(size_t step = 0; step <= distinguishingStepInteger; ++step) {
        result.emplace_back();
        Valuation &stepValues = result.back();
        for(auto const &[originalInput, newInput] : identifiers) {
            auto const inputNameExpression = makeIdentifier(std::get<std::string>(newInput));
            auto const valueToLookUp = makeFunctionCall({inputNameExpression, makeIdentifier(std::to_string(step))});
            auto const getValueExpression = makeFunctionCall({makeIdentifier("get-value"), makeFunctionCall({valueToLookUp})});
            auto value = extractValue(valuations->at(getValueExpression));
            stepValues = stepValues.modify(originalInput, std::to_string(value));
        }
        assert(not stepValues.getValuations().empty());
    }
    assert(not result.empty());
    return result;
}

std::optional<ConcreteTrace> getDistinguishingTraceForSfsm(Sfsm const &sfsm, Index state1, Index state2) {
    auto const transitionRelationDefinition = generateTransitionRelationDefinitionForSfsm(sfsm);
    return getDistinguishingTrace(sfsm, transitionRelationDefinition, state1, state2);
}

std::optional<ConcreteTrace> getDistinguishingTraceForSfsm(Sfsm const &sfsm, std::vector<Index> const &state1, std::vector<Index> const &state2) {
    auto const transitionRelationDefinition = generateTransitionRelationDefinitionForSfsm(sfsm);
    return getDistinguishingTrace(sfsm, transitionRelationDefinition, state1, state2);
}

std::optional<ConcreteTrace> getDistinguishingTraceForAbstractedSfsm(Sfsm const &sfsm, Expressions const &propositionEquivalenceClasses, Index state1, Index state2) {
    auto const transitionRelationDefinition = generateAbstractTransitionRelationDefinitionForSfsm(sfsm, propositionEquivalenceClasses);
    return getDistinguishingTrace(sfsm, transitionRelationDefinition, state1, state2);
}

std::optional<ConcreteTrace> getDistinguishingTraceForAbstractedSfsm(Sfsm const &sfsm, Expressions const &propositionEquivalenceClasses, std::vector<Index> const &state1, std::vector<Index> const &state2) {
    auto const transitionRelationDefinition = generateAbstractTransitionRelationDefinitionForSfsm(sfsm, propositionEquivalenceClasses);
    return getDistinguishingTrace(sfsm, transitionRelationDefinition, state1, state2);
}

Expressions traceReachesStateFromStateProblem(Sfsm const &sfsm, Expression const &transitionRelationDefinition, std::vector<Index> const &initialStates, Index targetState, ConcreteTrace const &trace) {
    std::set<Expression> declarations;
    Expressions problemStatements;
    auto const &definitions = sfsm.getIdentifierDefinitions();

    // Get inputs and outputs
    IdentifierDefinitions inputs = getIdentifiersByLocation(definitions, IdentifierType::LocationInModel::INPUT);
    IdentifierDefinitions outputs = getIdentifiersByLocation(definitions, IdentifierType::LocationInModel::OUTPUT);

    // Allocate distinct identifiers for all identifiers in the model
    auto traceIdentifiers = allocateTransitionRelationArgumentIdentifierInstantiations(sfsm, {});

    auto stateFunctionGeneral = makeNewIdentifier("state");
    std::map<Expression, IdentifierType> argumentIdentifiers = { 
        {stateFunctionGeneral, {makeTraceType(IdentifierType::SimpleType(IdentifierType::SimpleBaseType::INTEGER)), IdentifierType::LocationInModel::OTHER}} };
    std::transform(definitions.cbegin(), definitions.cend(), std::inserter(argumentIdentifiers, argumentIdentifiers.end()), [&traceIdentifiers](auto const &definition) {
        auto const &[identifier, type] = definition;
        return std::make_pair(traceIdentifiers.at(identifier), IdentifierType(makeTraceType(type.getType()), type.getLocation()));
    });

    // Collect all identifiers allocated for this problem and create a declaration expression
    std::map<Expression, IdentifierType> toDeclare = argumentIdentifiers;

    std::transform(toDeclare.cbegin(), toDeclare.cend(), std::inserter(declarations, declarations.end()), [](auto const &declare) {
        auto const &[expr, type] = declare;
        return makeFunctionDeclaration({std::to_string(expr), type});
    });
    declarations.insert(transitionRelationDefinition);
    // Create step identifier declaration
    declarations.insert(makeFunctionDeclaration({STEP_IDENTIFIER, 
        {IdentifierType::SimpleType(IdentifierType::SimpleBaseType::INTEGER), IdentifierType::LocationInModel::OTHER}}));

    // Initial constraints
    auto const initialStateExpr = makeFunctionCall(stateFunctionGeneral, {makeIdentifier("0")});
    Expressions initialStateConstraints;
    std::transform(initialStates.cbegin(), initialStates.cend(), std::back_inserter(initialStateConstraints), [&initialStateExpr](auto const &state) {
        return makeComparisonEquality(initialStateExpr, makeIdentifier(std::to_string(state)));
    });
    problemStatements.push_back(makeLogicalDisjunction(initialStateConstraints));

    // Pre- and post states for state1 and state2
    auto const preStateGeneral = makeFunctionCall({stateFunctionGeneral, makeIdentifier(STEP_IDENTIFIER)});
    auto const postStateGeneral = makeFunctionCall({stateFunctionGeneral, makeArithmeticAddition(makeIdentifier(STEP_IDENTIFIER), makeIdentifier("1"))});

    // Create a transition relation call with all identifiers allocated for the first trace
    Expressions transRelCallArguments = {preStateGeneral, postStateGeneral};

    for(auto const &definition : definitions) {
        auto const &[identifier, type] = definition;
        auto const mappedIdentifier = traceIdentifiers.at(identifier);
        transRelCallArguments.push_back(makeFunctionCall({mappedIdentifier, makeIdentifier(STEP_IDENTIFIER)}));
        if(type.getLocation() == IdentifierType::LocationInModel::OUTPUT) {
            transRelCallArguments.push_back(makeFunctionCall({mappedIdentifier, makeArithmeticSubtraction(makeIdentifier(STEP_IDENTIFIER), makeIdentifier("1"))}));
        }
    }

    auto transRelCall = generateTransitionRelationCall(sfsm, transRelCallArguments);

    // Call transition relation k times, where k is the number of states
    for(size_t callIndex = 0; callIndex < trace.size(); ++callIndex) {
        // (transRel (state n) (state (+ n 1)) ...
        auto relation = replaceSubexpressions(transRelCall, {{makeIdentifier(STEP_IDENTIFIER), makeIdentifier(std::to_string(callIndex))}});
        problemStatements.push_back(relation);
    }

    auto const lastState = replaceSubexpressions(preStateGeneral, {{makeIdentifier(STEP_IDENTIFIER), makeIdentifier(std::to_string(trace.size()))}});
    problemStatements.push_back(makeComparisonEquality(lastState, makeIdentifier(std::to_string(targetState))));

    for(size_t index = 0; index < trace.size(); ++index) {
        auto const &stepValuation = trace.at(index);
        auto stepValuationExpression = stepValuation.toExpression();
        for(auto const &[identifier, mapped] : traceIdentifiers) {
            stepValuationExpression = replaceSubexpressions(stepValuationExpression, {{makeIdentifier(identifier), makeFunctionCall({mapped, makeIdentifier(std::to_string(index))})}});
        }
        problemStatements.push_back(stepValuationExpression);
    }

    auto problemStatement = makeFunctionCall(makeIdentifier("assert"), {makeLogicalConjunction(problemStatements)});
    std::unordered_map<IdentifierDefinition, IdentifierDefinition> inputIdentifiers;
    std::transform(inputs.cbegin(), inputs.cend(), std::inserter(inputIdentifiers, inputIdentifiers.end()), [&traceIdentifiers](auto const &input) {
        auto const &[name, type] = input;
        return std::pair<IdentifierDefinition, IdentifierDefinition>(input, IdentifierDefinition(std::to_string(traceIdentifiers.at(name)), type));
    });

    // return the results
    Expressions result(declarations.cbegin(), declarations.cend());
    result.push_back(problemStatement);
    return result;
}

Expressions traceReachesStateFromStateProblem(Sfsm const &sfsm, Expression const &transitionRelationDefinition, std::vector<Index> const &initialStates, Index targetState, SymbolicTrace const &trace) {
    std::set<Expression> declarations;
    Expressions problemStatements;
    auto const &definitions = sfsm.getIdentifierDefinitions();

    // Get inputs and outputs
    IdentifierDefinitions inputs = getIdentifiersByLocation(definitions, IdentifierType::LocationInModel::INPUT);
    IdentifierDefinitions outputs = getIdentifiersByLocation(definitions, IdentifierType::LocationInModel::OUTPUT);

    // Allocate distinct identifiers for all identifiers in the model
    auto traceIdentifiers = allocateTransitionRelationArgumentIdentifierInstantiations(sfsm, {});

    auto stateFunctionGeneral = makeNewIdentifier("state");
    std::map<Expression, IdentifierType> argumentIdentifiers = { 
        {stateFunctionGeneral, {makeTraceType(IdentifierType::SimpleType(IdentifierType::SimpleBaseType::INTEGER)), IdentifierType::LocationInModel::OTHER}} };
    std::transform(definitions.cbegin(), definitions.cend(), std::inserter(argumentIdentifiers, argumentIdentifiers.end()), [&traceIdentifiers](auto const &definition) {
        auto const &[identifier, type] = definition;
        return std::make_pair(traceIdentifiers.at(identifier), IdentifierType(makeTraceType(type.getType()), type.getLocation()));
    });

    // Collect all identifiers allocated for this problem and create a declaration expression
    std::map<Expression, IdentifierType> toDeclare = argumentIdentifiers;

    std::transform(toDeclare.cbegin(), toDeclare.cend(), std::inserter(declarations, declarations.end()), [](auto const &declare) {
        auto const &[expr, type] = declare;
        return makeFunctionDeclaration({std::to_string(expr), type});
    });
    declarations.insert(transitionRelationDefinition);
    // Create step identifier declaration
    declarations.insert(makeFunctionDeclaration({STEP_IDENTIFIER, 
        {IdentifierType::SimpleType(IdentifierType::SimpleBaseType::INTEGER), IdentifierType::LocationInModel::OTHER}}));

    // Initial constraints
    auto const initialStateExpr = makeFunctionCall(stateFunctionGeneral, {makeIdentifier("0")});
    Expressions initialStateConstraints;
    std::transform(initialStates.cbegin(), initialStates.cend(), std::back_inserter(initialStateConstraints), [&initialStateExpr](auto const &state) {
        return makeComparisonEquality(initialStateExpr, makeIdentifier(std::to_string(state)));
    });
    problemStatements.push_back(makeLogicalDisjunction(initialStateConstraints));

    // Pre- and post states for state1 and state2
    auto const preStateGeneral = makeFunctionCall({stateFunctionGeneral, makeIdentifier(STEP_IDENTIFIER)});
    auto const postStateGeneral = makeFunctionCall({stateFunctionGeneral, makeArithmeticAddition(makeIdentifier(STEP_IDENTIFIER), makeIdentifier("1"))});

    // Create a transition relation call with all identifiers allocated for the first trace
    Expressions transRelCallArguments = {preStateGeneral, postStateGeneral};

    for(auto const &definition : definitions) {
        auto const &[identifier, type] = definition;
        auto const mappedIdentifier = traceIdentifiers.at(identifier);
        transRelCallArguments.push_back(makeFunctionCall({mappedIdentifier, makeIdentifier(STEP_IDENTIFIER)}));
        if(type.getLocation() == IdentifierType::LocationInModel::OUTPUT) {
            transRelCallArguments.push_back(makeFunctionCall({mappedIdentifier, makeArithmeticSubtraction(makeIdentifier(STEP_IDENTIFIER), makeIdentifier("1"))}));
        }
    }

    auto transRelCall = generateTransitionRelationCall(sfsm, transRelCallArguments);

    // Call transition relation k times, where k is the number of states
    for(size_t callIndex = 0; callIndex < trace.size(); ++callIndex) {
        // (transRel (state n) (state (+ n 1)) ...
        auto relation = replaceSubexpressions(transRelCall, {{makeIdentifier(STEP_IDENTIFIER), makeIdentifier(std::to_string(callIndex))}});
        problemStatements.push_back(relation);
    }

    auto const lastState = replaceSubexpressions(preStateGeneral, {{makeIdentifier(STEP_IDENTIFIER), makeIdentifier(std::to_string(trace.size()))}});
    problemStatements.push_back(makeComparisonEquality(lastState, makeIdentifier(std::to_string(targetState))));

    for(size_t index = 0; index < trace.size(); ++index) {
        auto const &[inputExpression, outputExpression] = trace.at(index);
        auto stepValuationExpression = makeLogicalConjunction(inputExpression, outputExpression);
        for(auto const &[identifier, mapped] : traceIdentifiers) {
            stepValuationExpression = replaceSubexpressions(stepValuationExpression, {{makeIdentifier(identifier), makeFunctionCall({mapped, makeIdentifier(std::to_string(index))})}});
        }
        problemStatements.push_back(stepValuationExpression);
    }

    auto problemStatement = makeFunctionCall(makeIdentifier("assert"), {makeLogicalConjunction(problemStatements)});
    std::unordered_map<IdentifierDefinition, IdentifierDefinition> inputIdentifiers;
    std::transform(inputs.cbegin(), inputs.cend(), std::inserter(inputIdentifiers, inputIdentifiers.end()), [&traceIdentifiers](auto const &input) {
        auto const &[name, type] = input;
        return std::pair<IdentifierDefinition, IdentifierDefinition>(input, IdentifierDefinition(std::to_string(traceIdentifiers.at(name)), type));
    });

    // return the results
    Expressions result(declarations.cbegin(), declarations.cend());
    result.push_back(problemStatement);
    return result;
}

std::vector<Index> targetStates(Sfsm const &sfsm, Expression const &transitionRelationDefinition, std::vector<Index> const &initialStates, ConcreteTrace const &trace) {
    if(trace.empty()) {
        return initialStates;
    }
    std::vector<Index> result;
    for(Index index = 0; index < sfsm.size(); ++index) {
        auto const problem = traceReachesStateFromStateProblem(sfsm, transitionRelationDefinition, initialStates, index, trace);
        if(isSatisfiable(problem, {}).has_value()) {
            result.push_back(index);
        }
    }
    return result;
}

std::vector<Index> getTargetStatesForSfsm(Sfsm const &sfsm, std::vector<Index> const &initialStates, ConcreteTrace const &trace) {
    auto const transitionRelationDefinition = generateTransitionRelationDefinitionForSfsm(sfsm);
    return targetStates(sfsm, transitionRelationDefinition, initialStates, trace);
}

std::vector<Index> targetStates(Sfsm const &sfsm, Expression const &transitionRelationDefinition, std::vector<Index> const &initialStates, SymbolicTrace const &trace) {
    if(trace.empty()) {
        return initialStates;
    }
    std::vector<Index> result;
    for(Index index = 0; index < sfsm.size(); ++index) {
        auto const problem = traceReachesStateFromStateProblem(sfsm, transitionRelationDefinition, initialStates, index, trace);
        if(isSatisfiable(problem, {}).has_value()) {
            result.push_back(index);
        }
    }
    return result;
}

// TODO: Make target state getting template code with trace type as template type
std::vector<Index> getTargetStatesForSfsm(Sfsm const &sfsm, std::vector<Index> const &initialStates, SymbolicTrace const &trace) {
    auto const transitionRelationDefinition = generateTransitionRelationDefinitionForSfsm(sfsm);
    return targetStates(sfsm, transitionRelationDefinition, initialStates, trace);
}

std::vector<Index> getTargetStatesForAbstractedSfsm(Sfsm const &sfsm, Expressions const &propositionEquivalenceClasses, std::vector<Index> const &initialStates, ConcreteTrace const &trace) {
    auto const transitionRelationDefinition = generateAbstractTransitionRelationDefinitionForSfsm(sfsm, propositionEquivalenceClasses);
    return targetStates(sfsm, transitionRelationDefinition, initialStates, trace);
}

std::optional<ConcreteTrace> getDistinguishingTraceForSfsm(Sfsm const &sfsm, ConcreteTrace const &trace1, ConcreteTrace const &trace2) {
    if(trace1 == trace2) {
        return std::nullopt;
    }
    auto states1 = getTargetStatesForSfsm(sfsm, {sfsm.getInitialStateId()}, trace1);
    auto states2 = getTargetStatesForSfsm(sfsm, {sfsm.getInitialStateId()}, trace2);
    return getDistinguishingTraceForSfsm(sfsm, states1, states2);
}

std::optional<ConcreteTrace> getDistinguishingTraceForAbstractedSfsm(Sfsm const &sfsm, Expressions const &propositionEquivalenceClasses, ConcreteTrace const &trace1, ConcreteTrace const &trace2) {
    if(trace1 == trace2) {
        return std::nullopt;
    }
    auto states1 = getTargetStatesForAbstractedSfsm(sfsm, propositionEquivalenceClasses, {sfsm.getInitialStateId()}, trace1);
    auto states2 = getTargetStatesForAbstractedSfsm(sfsm, propositionEquivalenceClasses, {sfsm.getInitialStateId()}, trace2);
    return getDistinguishingTraceForAbstractedSfsm(sfsm, propositionEquivalenceClasses, states1, states2);
}

}

