#include "sfsm/Variable.hpp"

#include <algorithm>
#include <iterator>
#include <string>

#include "sfsm/ExpressionDefinitions.hpp"
#include "sfsm/Expressions.hpp"

namespace libfsmtest::SFSMs {

Valuation::Valuation(IdentifierValuations const &valuations) {
    for(auto const &[definition, value] : valuations) {
        this->valuationFunction[definition] = value;
    }
}

ValuationFunction Valuation::getValuations(IdentifierDefinitions const &variableSet) const {
    ValuationFunction result;
    for(auto const &var : variableSet) {
        result[var] = this->valuationFunction.at(var);
    }
    return result;
}

Value const& Valuation::getValuation(IdentifierDefinition const &variable) const {
    return this->valuationFunction.at(variable);
}

Valuation Valuation::modify(IdentifierDefinition const &variable, Value const &value) const {
    auto copy = this->valuationFunction;
    copy[variable] = value;
    return Valuation{copy};
}

bool Valuation::operator<(Valuation const &other) const {
    if(this->valuationFunction.size() != other.valuationFunction.size()) {
        return this->valuationFunction.size() < other.valuationFunction.size();
    }

    auto iter1 = this->valuationFunction.begin();
    auto iter2 = other.valuationFunction.begin();

    while(iter1 != this->valuationFunction.end()) {
        auto const &[id1, val1] = *iter1;
        auto const &[id2, val2] = *iter2;
        if(id1 != id2) {
            return id1 < id2;
        }
        if(val1 != val2) {
            return val1 < val2;
        }
        ++iter1;
        ++iter2;
    }
    return false;
}

bool Valuation::operator==(Valuation const &other) const {
    return this->valuationFunction == other.valuationFunction;
}

Expression Valuation::toExpression() const {
    Expressions equivalences;
    std::transform(valuationFunction.begin(), valuationFunction.end(), std::back_inserter(equivalences), [](auto const &idVal) {
        auto const &[id, val] = idVal;
        auto const &[name, type] = id;
        return makeComparisonEquality(makeIdentifier(name), makeIdentifier(val));
    });
    if(not equivalences.empty()) {
        if(equivalences.size() > 1) {
            return makeLogicalConjunction(equivalences);
        } else {
            return equivalences.front();
        }
    } else {
        return makeFunctionCall({});
    }
}

} // namespace libfsmtest

namespace std {
std::string to_string(libfsmtest::SFSMs::Valuation const &valuation) {
    auto stringRep = std::to_string(valuation.toExpression());
    if(stringRep.front() != '(') {
        return "(" + stringRep + ")";
    } else {
        return stringRep;
    }
}
std::string to_string(libfsmtest::SFSMs::ConcreteTrace const &concreteTrace) {
    if(concreteTrace.empty()) {
        return "ε";
    }
    std::string result;
    bool isFirst = true;
    for(auto const &valuation : concreteTrace) {
        if(not isFirst) {
            result += '.';
        }
        result += std::to_string(valuation);
        isFirst = false;
    }
    return result;
}
std::size_t hash<libfsmtest::SFSMs::ValuationFunction>::operator()(libfsmtest::SFSMs::ValuationFunction const &valuationFunction) const {
    using libfsmtest::SFSMs::combine_hashes;
    std::size_t intermediate_hash = 0;
    struct std::hash<libfsmtest::SFSMs::IdentifierDefinition> keyHasher;
    struct std::hash<libfsmtest::SFSMs::Value> valueHasher;
    for(auto const &[key, value] : valuationFunction) {
        intermediate_hash = combine_hashes(intermediate_hash, combine_hashes(keyHasher(key), valueHasher(value)));
    }
    return intermediate_hash;
}
std::size_t hash<libfsmtest::SFSMs::Valuation>::operator()(libfsmtest::SFSMs::Valuation const &valuation) const {
    return std::hash<libfsmtest::SFSMs::ValuationFunction>{}(valuation.getValuations());
}
}

