/* MIT License
 * 
 * Copyright (c) 2022 Niklas Krafczyk
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

/* s-expression parser */

%{
#include <vector>
#include <stdexcept>

#include "sfsm/Expressions.hpp"

typedef void* yyscan_t;
#include "sfsm/typeparser/TypeParser.tab.hpp"
#include "sfsm/typeparser/TypeParser.lex.hpp"

using namespace libfsmtest::SFSMs;

void sfsmIdentifierTypeerror(libfsmtest::SFSMs::IdentifierType *&, std::string &, yyscan_t, char const*);

%}

/* Declarations */

%union {
	libfsmtest::SFSMs::IdentifierType::LocationInModel location;
	libfsmtest::SFSMs::IdentifierType::SimpleType *simpleType;
	libfsmtest::SFSMs::IdentifierType::ComplexType *complexType;
	libfsmtest::SFSMs::IdentifierType::Type *type;
	std::vector<libfsmtest::SFSMs::IdentifierType::Type> *typeList; }

%token IN
%token OUT
%token ARRAY
%token INTEGER
%token REAL
%token BOOL
%token COMMA
%token LPAREN
%token RPAREN
%token WHITESPACE
%nterm <location> LOCATION
%nterm <simpleType> SIMPLE_TYPE
%nterm <complexType> FUNCTIONTYPE
%nterm <complexType> ARRAYTYPE
%nterm <complexType> COMPLEX_TYPE
%nterm <type> TYPE
%nterm <typeList> TYPE_LIST

%start IDENTIFIER_TYPE

%define parse.lac full
%define parse.error detailed
%define api.prefix {sfsmIdentifierType}
%define api.pure full
%parse-param {libfsmtest::SFSMs::IdentifierType *&result} {std::string &errorString}
%param {yyscan_t yyscanner}

%%

/* Grammar */
SIMPLE_TYPE: BOOL { $SIMPLE_TYPE = new IdentifierType::SimpleType(IdentifierType::SimpleBaseType::BOOLEAN); }
					 | INTEGER { $SIMPLE_TYPE = new IdentifierType::SimpleType(IdentifierType::SimpleBaseType::INTEGER); }
					 | REAL { $SIMPLE_TYPE = new IdentifierType::SimpleType(IdentifierType::SimpleBaseType::REAL); }
FUNCTIONTYPE: TYPE LPAREN TYPE_LIST RPAREN { std::vector<IdentifierType::Type> type; type.push_back(*$TYPE); type.insert(type.end(), $TYPE_LIST->begin(), $TYPE_LIST->end()); $FUNCTIONTYPE = new IdentifierType::ComplexType(IdentifierType::ComplexBaseType::FUNCTION, type); delete $TYPE; delete $TYPE_LIST; }
ARRAYTYPE: ARRAY LPAREN TYPE[key] COMMA TYPE[value] RPAREN { std::vector<IdentifierType::Type> type; type.push_back(*$key); type.push_back(*$value); $ARRAYTYPE = new IdentifierType::ComplexType(IdentifierType::ComplexBaseType::ARRAY, type); delete $key; delete $value; }
COMPLEX_TYPE: FUNCTIONTYPE
					  | ARRAYTYPE
LOCATION: IN { $LOCATION = IdentifierType::LocationInModel::INPUT; }
				| OUT { $LOCATION = IdentifierType::LocationInModel::OUTPUT; }
				| %empty { $LOCATION = IdentifierType::LocationInModel::OTHER; }
TYPE: SIMPLE_TYPE { $TYPE = new IdentifierType::Type(*$SIMPLE_TYPE); delete $SIMPLE_TYPE; }
	  | COMPLEX_TYPE { $TYPE = new IdentifierType::Type(*$COMPLEX_TYPE); delete $COMPLEX_TYPE; }
TYPE_LIST[result]: TYPE_LIST[list] COMMA TYPE { $result = $list; $result->push_back(std::move(*$TYPE)); delete $TYPE; }
				 | TYPE { $result = new std::vector<IdentifierType::Type>{}; $result->push_back(std::move(*$TYPE)); delete $TYPE; }
IDENTIFIER_TYPE: LOCATION TYPE { result = new IdentifierType(*$TYPE, $LOCATION); delete $TYPE; }
%%

/* Epilogue */
