/* 
 * Copyright 2022 Niklas Krafczyk, Jan Peleska
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#include "testgeneration/SfsmMethod.hpp"

#include <utility>
#include <vector>
#include <algorithm>
#include <optional>
#include <functional>
#include <map>
#include <cassert>
#include <iterator>
#include <unordered_map>
#include <type_traits>
#include <tuple>
#include <set>
#include <cstdlib>
#include <string>
#include <fstream>
#include <memory>
#include <stdexcept>
#include <iostream>

#include "sfsm/Expressions.hpp"
#include "sfsm/Variable.hpp"
#include "sfsm/definitions.hpp"
#include "sfsm/Sfsm.hpp"
#include "libexpressions/utils/trie_node.hpp"
#include "sfsm/TransitionRelationGenerator.hpp"
#include "creators/transformers/AbstractionTransformer.hpp"
#include "sfsm/SMTTools.hpp"
#include "utils/algorithms/FixpointIteration.hpp"
#include "sfsm/Transition.hpp"
#include "utils/algorithms/BreadthFirstSearch.hpp"
#include "utils/algorithms/FindAllSatisfiableConjunctions.hpp"
#include "utils/algorithms/iota.hpp"
#include "processing/checkers/UnsatisfiableTransitionsChecker.hpp"
#include "processing/checkers/RestrictednessChecker.hpp"
#include "processing/RestrictedSfsmDistinguishingInputCalculation.hpp"
#include "libfsmtest/src/libfsmtest/fsm/Fsm.hpp"
#include "libfsmtest/src/libfsmtest/visitors/TestGenerationFrame.hpp"
#include "libfsmtest/src/libfsmtest/visitors/testgeneration/WMethod.hpp"
#include "processing/transformations/ToFSMAbstraction.hpp"

namespace libfsmtest::SFSMs {

typedef std::pair<std::vector<ssize_t>, Expression> ConjunctionResult;

class TraceTargetDeterminator {
private:
    libexpressions::TrieNode<ConcreteTrace::value_type, std::vector<Index>> tracesToTargets;
    Sfsm const sfsm;
public:
    explicit TraceTargetDeterminator(Sfsm const &paramSfsm) : sfsm(paramSfsm) {
        tracesToTargets[ConcreteTrace{}] = {sfsm.getInitialStateId()};
    }
    
    std::vector<Index> getTargets(ConcreteTrace const &trace) {
        auto const longestPrefix = tracesToTargets.getLongestPrefixContainingData(trace);
        auto const &targets = tracesToTargets.at(longestPrefix).value();
        if(longestPrefix == trace) {
            return targets;
        }
        ConcreteTrace suffix(trace.cbegin() + static_cast<ssize_t>(longestPrefix.size()), trace.cend());
        std::vector<Index> targetStates = getTargetStatesForSfsm(sfsm, targets, suffix);
        std::sort(targetStates.begin(), targetStates.end());
        targetStates.erase(std::unique(targetStates.begin(), targetStates.end()), targetStates.end());
        tracesToTargets[trace] = targetStates;
        return targetStates;
    }

    Sfsm const &getSfsm() const {
        return this->sfsm;
    }
};
class AbstractTraceTargetDeterminator {
private:
    libexpressions::TrieNode<ConcreteTrace::value_type, std::vector<Index>> tracesToTargets;
    Expressions const &propositionEquivalenceClasses;
    Sfsm const sfsm;
public:
    AbstractTraceTargetDeterminator(Sfsm const &paramSfsm, Expressions const &paramPropositionEquivalenceClasses)
     : propositionEquivalenceClasses(paramPropositionEquivalenceClasses), sfsm(paramSfsm) {
        tracesToTargets[ConcreteTrace{}] = {sfsm.getInitialStateId()};
    }
    
    std::vector<Index> getTargets(ConcreteTrace const &trace) {
        auto const longestPrefix = tracesToTargets.getLongestPrefixContainingData(trace);
        auto const &targets = tracesToTargets.at(longestPrefix).value();
        if(longestPrefix == trace) {
            return targets;
        }
        ConcreteTrace suffix(trace.cbegin() + static_cast<ssize_t>(longestPrefix.size()), trace.cend());
        std::vector<Index> targetStates = getTargetStatesForAbstractedSfsm(sfsm, propositionEquivalenceClasses, targets, suffix);
        std::sort(targetStates.begin(), targetStates.end());
        targetStates.erase(std::unique(targetStates.begin(), targetStates.end()), targetStates.end());
        tracesToTargets[trace] = targetStates;
        return targetStates;
    }

    Sfsm const &getSfsm() const {
        return this->sfsm;
    }
};

class DistinguishingTraceGenerator {
private:
    std::vector<std::pair<std::vector<Index>, std::vector<Index>>> indistinguishablePairs;
    TraceTargetDeterminator targets;
    Sfsm const sfsm;
public:
    explicit DistinguishingTraceGenerator(Sfsm const &paramSfsm) : targets(paramSfsm), sfsm(paramSfsm) { }

    std::optional<ConcreteTrace> getDistinguishingTrace(std::vector<Index> states1, std::vector<Index> states2) {
        std::sort(states1.begin(), states1.end());
        std::sort(states2.begin(), states2.end());
        auto const pair = (std::less<std::vector<Index>>{}(states1, states2) 
                            ?
                                std::pair<std::vector<Index>, std::vector<Index>>(states1, states2)
                              : std::pair<std::vector<Index>, std::vector<Index>>(states2, states1));
        if(states1 != states2) {
            auto const found = std::find(indistinguishablePairs.cbegin(), indistinguishablePairs.cend(), pair);
            if(found == indistinguishablePairs.cend()) {
                auto trace = getDistinguishingTraceForSfsm(sfsm, states1, states2);
                if(not trace.has_value()) {
                    indistinguishablePairs.push_back(pair);
                }
                return trace;
            }
        }
        return std::nullopt;
    }
    std::optional<ConcreteTrace> getDistinguishingTrace(ConcreteTrace const &trace1, ConcreteTrace const &trace2) {
        if(trace1 == trace2) {
            return std::nullopt;
        }
        auto states1 = targets.getTargets(trace1);
        auto states2 = targets.getTargets(trace2);
        auto distResult = getDistinguishingTrace(states1, states2);
        if(distResult.has_value()) {
            return distResult;
        }
        return std::nullopt;
    }
    TraceTargetDeterminator &getTargetDeterminator() {
        return targets;
    }

    Sfsm const &getSfsm() const {
        return this->sfsm;
    }
};

class MemoDistinguishingTraceGenerator {
private:
    std::vector<std::pair<std::vector<Index>, std::vector<Index>>> indistinguishablePairs;
    std::map<std::pair<std::vector<Index>, std::vector<Index>>, ConcreteTrace> distinguishingTraces;
    TraceTargetDeterminator targets;
    Sfsm const sfsm;
public:
    explicit MemoDistinguishingTraceGenerator(Sfsm const &paramSfsm) : targets(paramSfsm), sfsm(paramSfsm) { }

    std::optional<ConcreteTrace> getDistinguishingTrace(std::vector<Index> states1, std::vector<Index> states2) {
        std::sort(states1.begin(), states1.end());
        std::sort(states2.begin(), states2.end());
        auto const pair = (std::less<std::vector<Index>>{}(states1, states2) 
                            ?
                                std::pair<std::vector<Index>, std::vector<Index>>(states1, states2)
                              : std::pair<std::vector<Index>, std::vector<Index>>(states2, states1));
        if(states1 != states2) {
            auto const found = std::find(indistinguishablePairs.cbegin(), indistinguishablePairs.cend(), pair);
            if(found == indistinguishablePairs.cend()) {
                if(distinguishingTraces.count(pair) == 0) {
                    auto trace = getDistinguishingTraceForSfsm(sfsm, states1, states2);
                    if(not trace.has_value()) {
                        indistinguishablePairs.push_back(pair);
                    } else {
                        distinguishingTraces[pair] = trace.value();
                    }
                    return trace;
                } else {
                    return distinguishingTraces.at(pair);
                }
            }
        }
        return std::nullopt;
    }
    std::optional<ConcreteTrace> getDistinguishingTrace(ConcreteTrace const &trace1, ConcreteTrace const &trace2) {
        if(trace1 == trace2) {
            return std::nullopt;
        }
        auto states1 = targets.getTargets(trace1);
        auto states2 = targets.getTargets(trace2);
        return getDistinguishingTrace(states1, states2);
    }
    TraceTargetDeterminator &getTargetDeterminator() {
        return targets;
    }

    Sfsm const &getSfsm() const {
        return this->sfsm;
    }
};

class AbstractionDistinguishingTraceGenerator {
private:
    std::vector<std::pair<std::vector<Index>, std::vector<Index>>> indistinguishablePairs;
    Sfsm const sfsm;
    TraceTargetDeterminator targets;
public:
    AbstractionDistinguishingTraceGenerator(Sfsm const &paramSfsm,
                                            Expressions const &paramPropositions)
        : sfsm(*transformSfsm<AbstractionTransformer>(&paramSfsm, paramPropositions).release()),
          targets(sfsm) { }

    std::optional<ConcreteTrace> getDistinguishingTrace(std::vector<Index> states1, std::vector<Index> states2) {
        std::sort(states1.begin(), states1.end());
        std::sort(states2.begin(), states2.end());
        auto const pair = (std::less<std::vector<Index>>{}(states1, states2) 
                            ?
                                std::pair<std::vector<Index>, std::vector<Index>>(states1, states2)
                              : std::pair<std::vector<Index>, std::vector<Index>>(states2, states1));
        if(states1 != states2) {
            auto const found = std::find(indistinguishablePairs.cbegin(), indistinguishablePairs.cend(), pair);
            if(found == indistinguishablePairs.cend()) {
                auto trace = getDistinguishingTraceForSfsm(sfsm, states1, states2);
                if(not trace.has_value()) {
                    indistinguishablePairs.push_back(pair);
                }
                return trace;
            }
        }
        return std::nullopt;
    }
    std::optional<ConcreteTrace> getDistinguishingTrace(ConcreteTrace const &trace1, ConcreteTrace const &trace2) {
        if(trace1 == trace2) {
            return std::nullopt;
        }
        auto states1 = targets.getTargets(trace1);
        auto states2 = targets.getTargets(trace2);
        return getDistinguishingTrace(states1, states2);
    }
    TraceTargetDeterminator &getTargetDeterminator() {
        return targets;
    }

    Sfsm const &getSfsm() const {
        return this->sfsm;
    }
};

class MemoAbstractionDistinguishingTraceGenerator {
private:
    std::vector<std::pair<std::vector<Index>, std::vector<Index>>> indistinguishablePairs;
    std::map<std::pair<std::vector<Index>, std::vector<Index>>, ConcreteTrace> distinguishingTraces;
    Sfsm const sfsm;
    TraceTargetDeterminator targets;
public:
    MemoAbstractionDistinguishingTraceGenerator(Sfsm const &paramSfsm,
                                            Expressions const &paramPropositions)
        : sfsm(*transformSfsm<AbstractionTransformer>(&paramSfsm, paramPropositions).release()),
          targets(sfsm) { }

    std::optional<ConcreteTrace> getDistinguishingTrace(std::vector<Index> states1, std::vector<Index> states2) {
        std::sort(states1.begin(), states1.end());
        std::sort(states2.begin(), states2.end());
        auto const pair = (std::less<std::vector<Index>>{}(states1, states2) 
                            ?
                                std::pair<std::vector<Index>, std::vector<Index>>(states1, states2)
                              : std::pair<std::vector<Index>, std::vector<Index>>(states2, states1));
        if(states1 != states2) {
            auto const found = std::find(indistinguishablePairs.cbegin(), indistinguishablePairs.cend(), pair);
            if(found == indistinguishablePairs.cend()) {
                if(distinguishingTraces.count(pair) == 0) {
                    auto trace = getDistinguishingTraceForSfsm(sfsm, states1, states2);
                    if(not trace.has_value()) {
                        indistinguishablePairs.push_back(pair);
                    } else {
                        distinguishingTraces[pair] = trace.value();
                    }
                    return trace;
                } else {
                    return distinguishingTraces.at(pair);
                }
            }
        }
        return std::nullopt;
    }
    std::optional<ConcreteTrace> getDistinguishingTrace(ConcreteTrace const &trace1, ConcreteTrace const &trace2) {
        if(trace1 == trace2) {
            return std::nullopt;
        }
        auto states1 = targets.getTargets(trace1);
        auto states2 = targets.getTargets(trace2);
        return getDistinguishingTrace(states1, states2);
    }
    //AbstractTraceTargetDeterminator &getTargetDeterminator() {
    TraceTargetDeterminator &getTargetDeterminator() {
        return targets;
    }

    Sfsm const &getSfsm() const {
        return this->sfsm;
    }
};

std::vector<Index> satisfiableGuardConditions(Sfsm const &sfsm) {
    std::vector<Index> result;
    auto const &guardConditions = sfsm.getGuardConditions();
    for(size_t index = 0; index < guardConditions.size(); ++index) {
        if(isSatisfiable(makeLogicalConjunction(sfsm.getInvariant(), guardConditions.at(index)), sfsm.getIdentifierDefinitions()).has_value()) {
            result.push_back(index);
        }
    }
    return result;
}

std::vector<Index> reachableStates(Sfsm const &sfsm) {
    std::vector<Index> reachedStates(1, sfsm.getInitialStateId());
    std::vector<Index> guardConditions = satisfiableGuardConditions(sfsm);
    return fixpointIteration(reachedStates, [&guardConditions,&sfsm](auto const &states) {
        std::vector<Index> result = states;
        for(auto const state : states) {
            auto const &transitions = sfsm.getStates().at(state).getTransitions();
            for(auto const &transition : transitions) {
                if(std::find(guardConditions.cbegin(), guardConditions.cend(), transition.getInputId()) != guardConditions.cend()) {
                    result.push_back(transition.getTargetId());
                }
            }
        }
        std::sort(result.begin(), result.end());
        result.erase(std::unique(result.begin(), result.end()), result.end());
        return result;
    });
}

SymbolicInputTrace getSymbolicInputsForStateTrace(Sfsm const &sfsm, std::vector<Index> const &stateTrace) {
    SymbolicInputTrace result;
    auto const &guardConditions = sfsm.getGuardConditions();
    for(size_t index = 0; index + 1 < stateTrace.size(); ++index) {
        auto const preStateIndex = stateTrace.at(index);
        auto const postStateIndex = stateTrace.at(index+1);
        auto const &preState = sfsm.getStates().at(preStateIndex);
        auto const &transitions = preState.getTransitions();
        auto const &transitionIter = std::find_if(transitions.cbegin(), transitions.cend(), [&postStateIndex](auto const &transition) {
            return transition.getTargetId() == postStateIndex;
        });
        assert(transitionIter != transitions.cend());
        result.push_back(guardConditions.at(transitionIter->getInputId()));
    }
    return result;
}
SymbolicInputTrace getSymbolicStateCoverTraceToState(Sfsm const &sfsm, Index const &state) {
    auto const childGetterFunction = [&states=sfsm.getStates()](auto const &preStateIndex) {
        std::vector<Index> postStates;
        auto const &preState = states.at(preStateIndex);
        auto const &transitions = preState.getTransitions();
        std::transform(transitions.cbegin(), transitions.cend(), std::back_inserter(postStates), [](auto const &transition) {
            return transition.getTargetId();
        });
        return postStates;
    };
    auto const stateTrace = breadthFirstSearch(childGetterFunction, sfsm.getInitialStateId(), state).value();

    return getSymbolicInputsForStateTrace(sfsm, stateTrace);
}

std::vector<SymbolicInputTrace> getSymbolicStateCoverForStates(Sfsm const &sfsm, std::vector<Index> const &states) {
    std::vector<SymbolicInputTrace> result;
    std::transform(states.cbegin(), states.cend(), std::back_inserter(result), [&sfsm](auto const &stateIndex) {
        return getSymbolicStateCoverTraceToState(sfsm, stateIndex);
    });
    return result;
}

std::vector<std::pair<std::vector<ssize_t>, Expression>> pickHittingSetForInputEquivalenceClasses(std::vector<std::vector<std::pair<std::vector<ssize_t>, Expression>>> const &conjunctions) {
    /*auto const equality = [](ConjunctionResult const &r1, ConjunctionResult const &r2) {
        return std::get<Expression>(r1) == std::get<Expression>(r2);
    };*/
    std::vector<ConjunctionResult> result;
    //TODO
    /*auto const &conjunctionResultIterators = estimateMinimalHittingSet(conjunctions, equality);
    std::transform(conjunctionResultIterators.begin(), conjunctionResultIterators.end(), std::back_inserter(result), [](auto const &iter) {
        return *iter;
    });*/
    std::transform(conjunctions.cbegin(), conjunctions.cend(), std::back_inserter(result), [](auto const &conjunction) {
        return conjunction.front();
    });
    return result;
}

std::vector<ConcreteTrace> translateSymbolicTracesToConcreteTracesByEquivalenceClasses(Sfsm const &sfsm, std::vector<SymbolicInputTrace> const &symbolicTraces) {
    auto const &identifierDefinitions = sfsm.getIdentifierDefinitions();
    auto const &invariant = sfsm.getInvariant();
    auto const guardConditions = sfsm.getGuardConditions();
    std::vector<ConjunctionResult> const inputEquivalenceClasses 
        = findAllSatisfiableConjunctions(identifierDefinitions, invariant, guardConditions);
    auto const findSatisfyingClasses = [&inputEquivalenceClasses](size_t guardIndex) {
        std::vector<ConjunctionResult> satisfyingSubset;
        std::copy_if(inputEquivalenceClasses.cbegin(), inputEquivalenceClasses.cend(), std::back_inserter(satisfyingSubset), [&guardIndex](auto const &eqClass) {
            auto const &[indices, _] = eqClass;
            return std::find(indices.cbegin(), indices.cend(), guardIndex+1) != indices.cend();
        });
        return satisfyingSubset;
    };
    std::vector<std::vector<ConjunctionResult>> indexToSatisfyingClasses;
    for(size_t guardIndex = 0; guardIndex < guardConditions.size(); ++guardIndex) {
        indexToSatisfyingClasses.push_back(findSatisfyingClasses(guardIndex));
    }

    std::vector<ConjunctionResult> indexToPickedClass = pickHittingSetForInputEquivalenceClasses(indexToSatisfyingClasses);

    std::unordered_map<Expression, Valuation> valuationsForGuardConditions;
    for(size_t index = 0; index < guardConditions.size(); ++index) {
        auto const &pickedClass = std::get<Expression>(indexToPickedClass.at(index));
        auto const model = isSatisfiable(makeLogicalConjunction(invariant, pickedClass), sfsm.getIdentifierDefinitions()).value();
        Valuation valuation;
        for(auto const &[definition, value] : model) {
            valuation = valuation.modify(definition, value);
        }
        valuationsForGuardConditions[guardConditions.at(index)] = valuation;
    }

    auto const applyMap = [&valuationsForGuardConditions](Expression const &expr) {
        return valuationsForGuardConditions.at(expr);
    };

    auto transformSymbolicToConcrete = [&applyMap](SymbolicInputTrace const &trace) {
        ConcreteTrace cTrace;
        std::transform(trace.cbegin(), trace.cend(), std::back_inserter(cTrace), applyMap);
        return cTrace;
    };

    std::vector<ConcreteTrace> result;
    std::transform(symbolicTraces.cbegin(), symbolicTraces.cend(), std::back_inserter(result), transformSymbolicToConcrete);
    return result;
}

template<typename Container>
static void eraseEverySecondElement(Container &c) {
    typedef std::decay_t<decltype(c.begin())> IterType;
    bool reachedEnd = false;
    for(size_t counter = 0; not reachedEnd; ++counter) {
        IterType iter = c.begin();
        for(size_t incrementsLeft = counter; incrementsLeft != 0 and not reachedEnd; --incrementsLeft) {
            if(iter == c.end()) {
                reachedEnd = true;
            } else {
                ++iter;
            }
        }
        if(not reachedEnd) {
            c.erase(iter);
        }
    }
}

Valuation translateSymbolToValuation(IdentifierDefinitions const &definitions, Expression const &symbol, std::vector<Expression> forbiddenValuations) {
    SatisfyingValuationResult result;
    while(not result.has_value()) {
        result = isSatisfiable(makeLogicalConjunction(symbol, makeLogicalConjunction(makeLogicalNegation(forbiddenValuations))), definitions);
        eraseEverySecondElement(forbiddenValuations);
    }
    return Valuation{result.value()};
}

std::vector<ConcreteTrace> toInputTraces(Sfsm const &sfsm, std::vector<ConcreteTrace> const &traces) {
    std::vector<ConcreteTrace> result;
    auto const &identifierDefinitions = sfsm.getIdentifierDefinitions();
    IdentifierDefinitions inputIdentifiers;
    std::copy_if(identifierDefinitions.cbegin(), identifierDefinitions.cend(), std::back_inserter(inputIdentifiers), [](auto const &definition) {
        return std::get<IdentifierType>(definition).getLocation() == IdentifierType::LocationInModel::INPUT;
    });
    for(ConcreteTrace const &trace : traces) {
        ConcreteTrace inputTrace;
        for(Valuation const &valuation : trace) {
            inputTrace.emplace_back(valuation.getValuations(inputIdentifiers));
        }
        result.push_back(inputTrace);
    }
    return result;
}

ConcreteTrace translateSymbolicTraceToConcreteTraceWithPreferrablyUniqueValuations(Sfsm const &sfsm, std::set<Valuation> const &forbiddenValuations, SymbolicInputTrace const &symbolicTrace) {
    auto const &invariant = sfsm.getInvariant();
    auto const &identifierDefinitions = sfsm.getIdentifierDefinitions();
    std::vector<Expression> forbiddenValuationExpressions;
    std::transform(forbiddenValuations.cbegin(), forbiddenValuations.cend(), std::back_inserter(forbiddenValuationExpressions), [](auto const &valuation) {
        return valuation.toExpression();
    });

    ConcreteTrace result;
    std::transform(symbolicTrace.cbegin(), symbolicTrace.cend(), std::back_inserter(result), [&identifierDefinitions, &invariant, &forbiddenValuationExpressions](auto const &symbol) {
        return translateSymbolToValuation(identifierDefinitions, makeLogicalConjunction(symbol, invariant), forbiddenValuationExpressions);
    });
    return result;
}

std::vector<SymbolicTrace> translateConcreteTraceToSymbolicTraces(Sfsm const &sfsm, Index const &startState, ConcreteTrace const &trace) {
    std::vector<std::pair<Index, SymbolicTrace>> worklist;
    std::vector<std::pair<Index, SymbolicTrace>> nextWorklist;
    worklist.emplace_back(startState, SymbolicTrace{});
    for(auto const &valuation : trace) {
        for(auto const &[stateIndex, symbolicTrace] : worklist) {
            auto const &state = sfsm.getStates().at(stateIndex);
            for(auto const &transition : state.getTransitions()) {
                auto const &guardCondition = sfsm.getUsedGuardConditions().at(transition.getInputId());
                auto const &outputExpression = sfsm.getUsedOutputExpressions().at(transition.getOutputId());
                auto const expression = makeLogicalConjunction(valuation.toExpression(), sfsm.getInvariant(), guardCondition, outputExpression);
                if(isSatisfiable(expression, sfsm.getIdentifierDefinitions()).has_value()) {
                    auto newSymbolicTrace = symbolicTrace;
                    newSymbolicTrace.emplace_back(guardCondition, outputExpression);
                    nextWorklist.emplace_back(transition.getTargetId(), newSymbolicTrace);
                }
            }
        }
        worklist = std::move(nextWorklist);
        nextWorklist.clear();
    }
    std::vector<SymbolicTrace> result;
    for(auto const &[_, symbolicTrace] : worklist) {
        assert(symbolicTrace.size() == trace.size());
        result.push_back(symbolicTrace);
    }
    return result;
}

std::vector<std::vector<Index>> getStateTracesForSymbolicTrace(Sfsm const &sfsm, Index const &startState, SymbolicTrace const &trace) {
    std::vector<std::vector<Index>> worklist = {{startState}};
    std::vector<std::vector<Index>> nextWorklist;
    for(auto const &[input, output] : trace) {
        for(auto const &stateTrace : worklist) {
            auto const &stateIndex = stateTrace.back();
            auto const &state = sfsm.getStates().at(stateIndex);
            for(auto const &transition : state.getTransitions()) {
                auto const &guardCondition = sfsm.getUsedGuardConditions().at(transition.getInputId());
                auto const &outputExpression = sfsm.getUsedOutputExpressions().at(transition.getOutputId());
                if(guardCondition == input
                    and outputExpression == output) {
                    auto newStateTrace = stateTrace;
                    newStateTrace.push_back(transition.getTargetId());
                    nextWorklist.push_back(newStateTrace);
                }
            }
        }
        worklist = std::move(nextWorklist);
    }

    return {};
}

ConcreteTrace concatenateConcreteTraces(ConcreteTrace tr1, ConcreteTrace const &tr2) {
    tr1.insert(tr1.end(), tr2.cbegin(), tr2.cend());
    return tr1;
}

std::vector<ConcreteTrace> crossConcatenateConcreteTraceSets(std::vector<ConcreteTrace> const &prefixes, std::vector<ConcreteTrace> const &suffixes) {
    std::vector<ConcreteTrace> result;
    for(auto const &prefix : prefixes) {
        for(auto const &suffix : suffixes) {
            result.push_back(concatenateConcreteTraces(prefix, suffix));
        }
    }
    return result;
}

std::vector<ConcreteTrace> generateStateCover(Sfsm const &sfsm, std::vector<Index> const &statesToCover) {
    auto const symbolicStateCover = getSymbolicStateCoverForStates(sfsm, statesToCover);
    return toInputTraces(sfsm, translateSymbolicTracesToConcreteTracesByEquivalenceClasses(sfsm, symbolicStateCover));
}

std::vector<ConcreteTrace> generateDiscoveryTraces(Sfsm const &sfsm, std::vector<Valuation> const &equivalenceClasses, size_t stateDifference) {
    std::vector<ConcreteTrace> result;
    std::vector<std::vector<Index>> accummulatedTraces;
    std::vector<std::vector<Index>> tracesOfPreviousLength = {{}};

    for(size_t i = 0; i < stateDifference+1; ++i) {
        std::vector<std::vector<Index>> tracesOfCurrentLength;
        for(auto const &traceOfPreviousLength : tracesOfPreviousLength) {
            for(size_t eqClassIndex = 0; eqClassIndex < equivalenceClasses.size(); ++eqClassIndex) {
                std::vector<Index> traceOfCurrentLength = traceOfPreviousLength;
                traceOfCurrentLength.push_back(eqClassIndex);
                tracesOfCurrentLength.push_back(traceOfCurrentLength);
            }
        }
        accummulatedTraces.insert(accummulatedTraces.end(), tracesOfCurrentLength.cbegin(), tracesOfCurrentLength.cend());
        tracesOfPreviousLength = tracesOfCurrentLength;
    }

    auto const mapToValuation = [&equivalenceClasses](auto const index) {
        return equivalenceClasses.at(index);
    };
    auto const mapToConcreteTrace = [&mapToValuation](auto const &indexTrace) {
        ConcreteTrace mappedConcreteTrace;
        std::transform(indexTrace.cbegin(), indexTrace.cend(), std::back_inserter(mappedConcreteTrace), mapToValuation);
        return mappedConcreteTrace;
    };
    std::transform(accummulatedTraces.cbegin(), accummulatedTraces.cend(), std::back_inserter(result), mapToConcreteTrace);
    return toInputTraces(sfsm, result);
}

std::vector<ConcreteTrace> generatePartOne(Sfsm const &sfsm, std::vector<Index> const &statesToCover) {
    auto stateCover = generateStateCover(sfsm, statesToCover /* = reachableStates(sfsm) */);
    if(auto const envVarPtr = getenv("LIBSFSMTEST_TEST_SUITE_TEX");
        envVarPtr != nullptr) {
        std::string const texOutputFile = envVarPtr;
        std::ofstream output(texOutputFile, std::ios_base::app);
        output << "\\subsection{State cover traces}" << std::endl;
        output << "\\begin{align*}" << std::endl;
        output << "C = \\{ ";
        for(auto const &testTrace : stateCover) {
            output << "&" << std::to_string(testTrace) << std::endl;
        }
        output << "\\}" << std::endl;
        output << "\\end{align*}" << std::endl;
    }
    return stateCover;
}

std::vector<ConcreteTrace> generatePartTwo(Sfsm const &sfsm, std::vector<Index> const &statesToCover) {
    std::vector<ConcreteTrace> result;
    auto const stateCover = generateStateCover(sfsm, statesToCover);
    for(size_t traceIndex1 = 0; traceIndex1 < stateCover.size() - 1; ++traceIndex1) {
        auto const &stateTrace1 = stateCover.at(traceIndex1);
        for(size_t traceIndex2 = traceIndex1 + 1; traceIndex2 < stateCover.size(); ++traceIndex2) {
            auto const &stateTrace2 = stateCover.at(traceIndex2);
            //TODO: Only correct for completely specified right now
            auto const distTraceOpt = getDistinguishingTraceForSfsm(sfsm, stateTrace1, stateTrace2);
            if(distTraceOpt.has_value()) {
                auto const &distTrace = distTraceOpt.value();
                auto newStateTrace1 = stateTrace1;
                auto newStateTrace2 = stateTrace2;
                newStateTrace1.insert(newStateTrace1.end(), distTrace.cbegin(), distTrace.cend());
                newStateTrace2.insert(newStateTrace2.end(), distTrace.cbegin(), distTrace.cend());
                result.push_back(newStateTrace1);
                result.push_back(newStateTrace2);
            }
        }
    }
    auto inputTraces = toInputTraces(sfsm, result);
    if(auto const envVarPtr = getenv("LIBSFSMTEST_TEST_SUITE_TEX");
        envVarPtr != nullptr) {
        std::string const texOutputFile = envVarPtr;
        std::ofstream output(texOutputFile, std::ios_base::app);
        output << "\\subsection{State cover + distinguishing traces}" << std::endl;
        output << "\\begin{align*}" << std::endl;
        output << R"(\{&\alpha.\delta \mid (\alpha,\beta) \in C \times C \land \delta \in \Delta(\alpha, \beta)\}\)" << std::endl;
        output << "= \\{";
        for(auto const &testTrace : inputTraces) {
            output << "&" << std::to_string(testTrace) << std::endl;
        }
        output << "\\}" << std::endl;
        output << "\\end{align*}" << std::endl;
    }
    return inputTraces;
}

static void equivalenceClassesToDotTree(std::string const &outputFilePath, Expressions const &expressionsForEq, std::vector<ConjunctionResult> const &equivalenceClasses, std::vector<Valuation> const &valuationsForClasses) {
    std::ofstream output(outputFilePath, std::ios_base::trunc);
    if(not output.is_open()) {
        throw std::runtime_error("Could not open path '" + outputFilePath + "' for output of the equivalence class tree graphviz data.");
    }
    output << "digraph {" << std::endl;
    typedef std::pair<std::string, std::vector<ConjunctionResult>> EquivalenceClassPartitionMember;
    std::vector<EquivalenceClassPartitionMember> partitioningOfEquivalenceClasses = { {"n0", equivalenceClasses} };
    std::vector<EquivalenceClassPartitionMember> nextPartitioningOfEquivalenceClasses;
    std::vector<std::string> intermediateNodeNames;
    for(size_t idx = 0; idx < expressionsForEq.size(); ++idx) {
        size_t const expressionIndex = idx + 1;
        for(auto const &[name, partition] : partitioningOfEquivalenceClasses) {
            intermediateNodeNames.emplace_back(name);
            std::vector<ConjunctionResult> positivePartition;
            std::vector<ConjunctionResult> negativePartition;
            for(auto const &partitionMember : partition) {
                auto const &[indices, _] = partitionMember;
                auto const iterPositive = std::find(indices.cbegin(), indices.cend(),  static_cast<ssize_t>(expressionIndex));
                auto const iterNegative = std::find(indices.cbegin(), indices.cend(), -static_cast<ssize_t>(expressionIndex));
                if(iterPositive != indices.end()) {
                    positivePartition.push_back(partitionMember);
                } else {
                    assert(iterNegative != indices.end());
                    negativePartition.push_back(partitionMember);
                }
            }
            if(not positivePartition.empty()) {
                std::string const newName = name + "1";
                output << name << " -> " << newName << R"( [label="+", tooltip=")" << std::to_string(expressionsForEq.at(idx)) << "\"];";
                nextPartitioningOfEquivalenceClasses.emplace_back(newName, positivePartition);
            }
            if(not negativePartition.empty()) {
                std::string const newName = name + "0";
                output << name << " -> " << newName << R"( [label="-", tooltip="(not )" << std::to_string(expressionsForEq.at(idx)) << ")\"];";
                nextPartitioningOfEquivalenceClasses.emplace_back(newName, negativePartition);
            }
        }
        // Note: `nextPartitioningOfEquivalenceClasses` is guaranteed to be empty after the next line
        partitioningOfEquivalenceClasses = std::move(nextPartitioningOfEquivalenceClasses);
    }
    std::vector<std::vector<std::string>> subgraphNodes(expressionsForEq.size());
    for(auto const &[className, partitioningOfEquivalenceClass] : partitioningOfEquivalenceClasses) {
        auto prefixOfClassName = className;
        while(prefixOfClassName.size() > 2) {
            auto &subgraphNode = subgraphNodes.at(prefixOfClassName.size()-3);
            auto iter = std::find(subgraphNode.cbegin(), subgraphNode.cend(), prefixOfClassName);
            if(iter == subgraphNode.cend()) {
                subgraphNode.push_back(prefixOfClassName);
            }
            prefixOfClassName = prefixOfClassName.substr(0, prefixOfClassName.size()-1);
        }
    }
    output << "n0[label=\"\", shape=circle, width=0.1, height=0.1, style=filled];" << std::endl;
    for(size_t subgraphIndex = 0; subgraphIndex < subgraphNodes.size(); ++subgraphIndex) {
        auto const &subgraph = subgraphNodes.at(subgraphIndex);
        output << "subgraph cluster_for_expression" << subgraphIndex << " { " << std::endl;
        for(auto const &intermediateNodeName : subgraph) {
            output << intermediateNodeName << R"([label="", shape=circle, width=0.1, height=0.1, style=filled, tooltip="(and)";
            for(size_t nodeNameIndex = 2; nodeNameIndex < intermediateNodeName.size(); ++nodeNameIndex) {
                auto const &expression = expressionsForEq.at(nodeNameIndex-2);
                if(intermediateNodeName.at(nodeNameIndex) == '1') {
                    output << "\\n     " << std::to_string(expression);
                } else {
                    output << "\\n(not " << std::to_string(expression) << ")";
                }
            }
            output << "\\n)\"];" << std::endl;
        }
        output << "label = \"" << std::to_string(expressionsForEq.at(subgraphIndex)) << "\";" << std::endl;
    }
    for(auto const &[name, partitionSet] : partitioningOfEquivalenceClasses) {
        assert(partitionSet.size() == 1);
        auto const &partitionElement = std::get<std::vector<ssize_t>>(partitionSet.front());
        auto const iterToEquivalenceClass = std::find_if(equivalenceClasses.cbegin(), equivalenceClasses.cend(), [&partitionElement](auto const &equivalenceClass) {
            auto const &[indices, _] = equivalenceClass;
            return partitionElement == indices;
        });
        auto const indexToEquivalenceClass = static_cast<size_t>(std::distance(equivalenceClasses.cbegin(), iterToEquivalenceClass));
        auto const &valuation = valuationsForClasses.at(indexToEquivalenceClass);
        output << name << " [shape=box, label=\"";
        for(auto const &[id, value] : valuation.getValuations()) {
            auto const &[varName, type] = id;
            if(auto const location = type.getLocation();
                location == IdentifierType::LocationInModel::INPUT
              or location == IdentifierType::LocationInModel::OUTPUT) {
                output << varName << " = " << value << "\\n";
            }
        }
        output << "\"];" << std::endl;
    }
    for(size_t subgraphIndex = 0; subgraphIndex < subgraphNodes.size(); ++subgraphIndex) {
        output << "}";
    }
    output << std::endl;
    output << "}";
}

std::vector<ConcreteTrace> generatePartThree(Sfsm const &sfsm, std::vector<Index> const &statesToCover, size_t stateDifference, Expressions const &propositions) {
    auto const &identifierDefinitions = sfsm.getIdentifierDefinitions();
    auto const &invariant = sfsm.getInvariant();
    auto const &guardConditions = sfsm.getUsedGuardConditions();
    auto const &outputExpressions = sfsm.getUsedOutputExpressions();
    auto const &additionalGuardConditions = sfsm.getAdditionalGuardConditions();
    auto const &additionalOutputExpressions = sfsm.getAdditionalOutputExpressions();
    Expressions expressionsForEq = guardConditions;
    expressionsForEq.insert(expressionsForEq.end(), outputExpressions.begin(), outputExpressions.end());
    expressionsForEq.insert(expressionsForEq.end(), additionalGuardConditions.begin(), additionalGuardConditions.end());
    expressionsForEq.insert(expressionsForEq.end(), additionalOutputExpressions.begin(), additionalOutputExpressions.end());
    expressionsForEq.insert(expressionsForEq.end(), propositions.begin(), propositions.end());

    std::vector<ConjunctionResult> const equivalenceClasses 
        = findAllSatisfiableConjunctions(identifierDefinitions, invariant, expressionsForEq);
    std::vector<Valuation> valuationsForClasses;
    std::transform(equivalenceClasses.cbegin(), equivalenceClasses.cend(), std::back_inserter(valuationsForClasses),
        [&identifierDefinitions, &invariant](auto const &equivalenceClass) {
        auto const &[_, expression] = equivalenceClass;
        return Valuation{isSatisfiable(makeLogicalConjunction(invariant, expression), identifierDefinitions).value()};
    });
    if(auto const envVarPtr = getenv("LIBSFSMTEST_EQUIVALENCE_CLASS_TREE_DOT_OUTPUT");
        envVarPtr != nullptr) {
        std::string const dotOutputFile = envVarPtr;
        equivalenceClassesToDotTree(dotOutputFile, expressionsForEq, equivalenceClasses, valuationsForClasses);
    }

    auto const stateCover = generateStateCover(sfsm, statesToCover);
    auto const discoveryTraces = generateDiscoveryTraces(sfsm, valuationsForClasses, stateDifference);
    return toInputTraces(sfsm, crossConcatenateConcreteTraceSets(stateCover, discoveryTraces));
}

std::vector<ConcreteTrace> generatePartFour(Sfsm const &sfsm, std::vector<Index> const &statesToCover, size_t stateDifference, Expressions const &/*propositionEquivalenceClasses*/, Expressions const &propositions) {
    std::vector<ConcreteTrace> result;
    bool const abstract = not propositions.empty();
    auto const stateCover = generateStateCover(sfsm, statesToCover);
    auto const partThree = generatePartThree(sfsm, statesToCover, stateDifference, propositions);
    std::vector<ConcreteTrace> tracesToDistinguish = stateCover;
    tracesToDistinguish.insert(tracesToDistinguish.end(), partThree.begin(), partThree.end());
    std::sort(tracesToDistinguish.begin(), tracesToDistinguish.end());
    auto const newEnd = std::unique(tracesToDistinguish.begin(), tracesToDistinguish.end());
    tracesToDistinguish.erase(newEnd, tracesToDistinguish.end());

    auto const progress = [&tracesToDistinguish](size_t index) {
        auto const &size = tracesToDistinguish.size();
        auto const N = size-index;
        auto const M = size;
        auto const total = static_cast<double>(M*(M-1)) / 2.0;
        auto const toDo = static_cast<double>(N*(N-1)) / 2.0;
        return 100.0*(total-toDo)/total;
    };
    MemoDistinguishingTraceGenerator distTraceGenerator(sfsm);
    std::unique_ptr<MemoAbstractionDistinguishingTraceGenerator> abstDistTraceGenerator;
    if(abstract) {
        abstDistTraceGenerator = std::make_unique<MemoAbstractionDistinguishingTraceGenerator>(sfsm, propositions);
    }
    std::cerr << "Calculating the first part of the final test suite..." << std::endl;
    for(size_t traceIndex1 = 0; traceIndex1 < tracesToDistinguish.size() - 1; ++traceIndex1) {
        std::cerr << progress(traceIndex1) << "%" << std::endl;
        auto const &trace1 = tracesToDistinguish.at(traceIndex1);
        for(size_t traceIndex2 = traceIndex1 + 1; traceIndex2 < tracesToDistinguish.size(); ++traceIndex2) {
            auto const &trace2 = tracesToDistinguish.at(traceIndex2);
            if(abstract) {
                if(auto abstractDistTrace = abstDistTraceGenerator->getDistinguishingTrace(trace1, trace2); abstractDistTrace.has_value()) {
                    auto const distTrace = distTraceGenerator.getDistinguishingTrace(trace1, trace2);
                    if(not distTrace.has_value()) {
                        auto const &abstractSfsm = abstDistTraceGenerator->getSfsm();
                        std::string exceptionMessage = "Abstraction is not state preserving.\nThe following traces are distinguishable under abstraction but not in the SFSM:\n";
                        exceptionMessage += std::to_string(trace1);
                        exceptionMessage += "\n";
                        exceptionMessage += std::to_string(trace2);
                        exceptionMessage += "\n";
                        exceptionMessage += "Trace " + std::to_string(trace1) + " reaches the following states:\n";
                        for(auto const &target : distTraceGenerator.getTargetDeterminator().getTargets(trace1)) {
                            exceptionMessage += " " + std::to_string(target);
                        }
                        exceptionMessage += "\n";
                        exceptionMessage += "Trace " + std::to_string(trace2) + " reaches the following states:\n";
                        for(auto const &target : distTraceGenerator.getTargetDeterminator().getTargets(trace2)) {
                            exceptionMessage += " " + std::to_string(target);
                        }
                        exceptionMessage += "\n";
                        exceptionMessage += "Trace " + std::to_string(trace1) + " reaches the following states under abstraction:\n";
                        for(auto const &target : abstDistTraceGenerator->getTargetDeterminator().getTargets(trace1)) {
                            exceptionMessage += " " + std::to_string(target);
                        }
                        exceptionMessage += "\n";
                        exceptionMessage += "Trace " + std::to_string(trace2) + " reaches the following states under abstraction:\n";
                        for(auto const &target : abstDistTraceGenerator->getTargetDeterminator().getTargets(trace2)) {
                            exceptionMessage += " " + std::to_string(target);
                        }
                        exceptionMessage += "\n";
                        exceptionMessage += "States are reached that are distinguishable under abstraction by the following trace:\n";
                        exceptionMessage += std::to_string(abstractDistTrace.value());

                        //Calculating symbolic traces
                        auto targets1 = abstDistTraceGenerator->getTargetDeterminator().getTargets(trace1);
                        std::sort(targets1.begin(), targets1.end());
                        auto targets2 = abstDistTraceGenerator->getTargetDeterminator().getTargets(trace2);
                        std::sort(targets2.begin(), targets2.end());
                        std::vector<SymbolicTrace> tracesIn1;
                        std::vector<SymbolicTrace> tracesIn2;
                        for(auto const &state : targets1) {
                            auto const traces = translateConcreteTraceToSymbolicTraces(abstractSfsm, state, abstractDistTrace.value());
                            tracesIn1.insert(tracesIn1.end(), traces.begin(), traces.end());
                        }
                        for(auto const &state : targets2) {
                            auto const traces = translateConcreteTraceToSymbolicTraces(abstractSfsm, state, abstractDistTrace.value());
                            tracesIn2.insert(tracesIn2.end(), traces.begin(), traces.end());
                        }
                        std::sort(tracesIn1.begin(), tracesIn1.end());
                        std::sort(tracesIn2.begin(), tracesIn2.end());
                        tracesIn1.erase(std::unique(tracesIn1.begin(), tracesIn1.end()), tracesIn1.end());
                        tracesIn2.erase(std::unique(tracesIn2.begin(), tracesIn2.end()), tracesIn2.end());
                        std::vector<SymbolicTrace> tracesOnlyIn1;
                        std::vector<SymbolicTrace> tracesOnlyIn2;
                        std::set_difference(tracesIn1.begin(), tracesIn1.end(), tracesIn2.begin(), tracesIn2.end(), std::back_inserter(tracesOnlyIn1));
                        std::set_difference(tracesIn2.begin(), tracesIn2.end(), tracesIn1.begin(), tracesIn1.end(), std::back_inserter(tracesOnlyIn2));
                        assert((not tracesOnlyIn1.empty()) or (not tracesOnlyIn2.empty()));
                        exceptionMessage += "\n";
                        if(not tracesOnlyIn1.empty()) {
                            exceptionMessage += "The following abstract symbolic traces are only possible after " + std::to_string(trace1);
                            for(auto const &trace : tracesOnlyIn1) {
                                exceptionMessage += "\n";
                                exceptionMessage += std::to_string(trace);
                                exceptionMessage += "\n\ttraversing states:";
                                for(auto const &target : targets1) {
                                    auto const stateTraces = getStateTracesForSymbolicTrace(abstractSfsm, target, trace);
                                    for(auto const &stateTrace : stateTraces) {
                                        exceptionMessage += "\n\t";
                                        exceptionMessage += std::to_string(stateTrace);
                                    }
                                }
                            }
                        }
                        exceptionMessage += "\n";
                        if(not tracesOnlyIn2.empty()) {
                            exceptionMessage += "The following abstract symbolic traces are only possible after " + std::to_string(trace2);
                            for(auto const &trace : tracesOnlyIn2) {
                                exceptionMessage += "\n";
                                exceptionMessage += std::to_string(trace);
                                exceptionMessage += "\n\ttraversing states:";
                                for(auto const &target : targets2) {
                                    auto const stateTraces = getStateTracesForSymbolicTrace(abstractSfsm, target, trace);
                                    for(auto const &stateTrace : stateTraces) {
                                        exceptionMessage += "\n\t";
                                        exceptionMessage += std::to_string(stateTrace);
                                    }
                                }
                            }
                        }
                        throw std::runtime_error(exceptionMessage);
                    }
                    auto newTrace1 = trace1;
                    newTrace1.insert(newTrace1.end(), distTrace->begin(), distTrace->end());
                    auto newTrace2 = trace2;
                    newTrace2.insert(newTrace2.end(), distTrace->begin(), distTrace->end());
                    result.push_back(newTrace1);
                    result.push_back(newTrace2);
                }
            } else {
                auto const &distTrace = distTraceGenerator.getDistinguishingTrace(trace1, trace2);
                if(distTrace.has_value()) {
                    auto newTrace1 = trace1;
                    newTrace1.insert(newTrace1.end(), distTrace->begin(), distTrace->end());
                    auto newTrace2 = trace2;
                    newTrace2.insert(newTrace2.end(), distTrace->begin(), distTrace->end());
                    result.push_back(newTrace1);
                    result.push_back(newTrace2);
                }
            }
        }
    }
    return toInputTraces(sfsm, result);
}

static std::vector<ConcreteTrace> prefixClosure(ConcreteTrace const &trace) {
    std::vector<ConcreteTrace> result;
    for(auto iter = trace.cbegin(); iter != trace.cend(); ++iter) {
        result.emplace_back(trace.cbegin(), iter);
    }
    return result;
}

std::vector<ConcreteTrace> generatePartFive(Sfsm const &sfsm, std::vector<Index> const &statesToCover, size_t stateDifference, Expressions const &/*propositionEquivalenceClasses*/, Expressions const &propositions) {
    std::vector<ConcreteTrace> result;
    bool const abstract = not propositions.empty();
    std::vector<ConcreteTrace> tracesToDistinguish = generatePartThree(sfsm, statesToCover, stateDifference, propositions);

    auto const numberOfTracesUpToNotIncluding = [&tracesToDistinguish](auto const &upperBoundIndex) {
        size_t numTracesResult = 0;
        for(size_t idx = 0; idx < upperBoundIndex; ++idx) {
            numTracesResult += tracesToDistinguish.at(idx).size();
        }
        return numTracesResult;
    };
    size_t const totalNumberOfTraces = numberOfTracesUpToNotIncluding(tracesToDistinguish.size());
    auto const progress = [&totalNumberOfTraces,&numberOfTracesUpToNotIncluding](size_t index) {
        return 100.0*static_cast<double>(numberOfTracesUpToNotIncluding(index))/static_cast<double>(totalNumberOfTraces);
    };
    MemoDistinguishingTraceGenerator distTraceGenerator(sfsm);
    std::unique_ptr<MemoAbstractionDistinguishingTraceGenerator> abstDistTraceGenerator;
    if(abstract) {
        abstDistTraceGenerator= std::make_unique<MemoAbstractionDistinguishingTraceGenerator>(sfsm, propositions);
    }
    std::cerr << "Calculating the second part of the final test suite..." << std::endl;
    for(size_t traceIndex1 = 0; traceIndex1 < tracesToDistinguish.size(); ++traceIndex1) {
        std::cerr << progress(traceIndex1) << "%" << std::endl;
        auto const &trace1 = tracesToDistinguish.at(traceIndex1);
        auto const prefixClosureTraces = prefixClosure(trace1);
        for(auto const &trace2 : prefixClosureTraces) {
            if(abstract) {
                if(auto abstractDistTrace = abstDistTraceGenerator->getDistinguishingTrace(trace1, trace2); abstractDistTrace.has_value()) {
                    auto const distTrace = distTraceGenerator.getDistinguishingTrace(trace1, trace2);
                    if(not distTrace.has_value()) {
                        auto const &abstractSfsm = abstDistTraceGenerator->getSfsm();
                        std::string exceptionMessage = "Abstraction is not state preserving.\nThe following traces are distinguishable under abstraction but not in the SFSM:\n";
                        exceptionMessage += std::to_string(trace1);
                        exceptionMessage += "\n";
                        exceptionMessage += std::to_string(trace2);
                        exceptionMessage += "\n";
                        exceptionMessage += "Trace " + std::to_string(trace1) + " reaches the following states:\n";
                        for(auto const &target : distTraceGenerator.getTargetDeterminator().getTargets(trace1)) {
                            exceptionMessage += " " + std::to_string(target);
                        }
                        exceptionMessage += "\n";
                        exceptionMessage += "Trace " + std::to_string(trace2) + " reaches the following states:\n";
                        for(auto const &target : distTraceGenerator.getTargetDeterminator().getTargets(trace2)) {
                            exceptionMessage += " " + std::to_string(target);
                        }
                        exceptionMessage += "\n";
                        exceptionMessage += "Trace " + std::to_string(trace1) + " reaches the following states under abstraction:\n";
                        for(auto const &target : abstDistTraceGenerator->getTargetDeterminator().getTargets(trace1)) {
                            exceptionMessage += " " + std::to_string(target);
                        }
                        exceptionMessage += "\n";
                        exceptionMessage += "Trace " + std::to_string(trace2) + " reaches the following states under abstraction:\n";
                        for(auto const &target : abstDistTraceGenerator->getTargetDeterminator().getTargets(trace2)) {
                            exceptionMessage += " " + std::to_string(target);
                        }
                        exceptionMessage += "\n";
                        exceptionMessage += "States are reached that are distinguishable under abstraction by the following trace:\n";
                        exceptionMessage += std::to_string(abstractDistTrace.value());

                        //Calculating symbolic traces
                        auto targets1 = abstDistTraceGenerator->getTargetDeterminator().getTargets(trace1);
                        std::sort(targets1.begin(), targets1.end());
                        auto targets2 = abstDistTraceGenerator->getTargetDeterminator().getTargets(trace2);
                        std::sort(targets2.begin(), targets2.end());
                        std::vector<SymbolicTrace> tracesIn1;
                        std::vector<SymbolicTrace> tracesIn2;
                        for(auto const &state : targets1) {
                            auto const traces = translateConcreteTraceToSymbolicTraces(abstractSfsm, state, abstractDistTrace.value());
                            tracesIn1.insert(tracesIn1.end(), traces.begin(), traces.end());
                        }
                        for(auto const &state : targets2) {
                            auto const traces = translateConcreteTraceToSymbolicTraces(abstractSfsm, state, abstractDistTrace.value());
                            tracesIn2.insert(tracesIn2.end(), traces.begin(), traces.end());
                        }
                        std::sort(tracesIn1.begin(), tracesIn1.end());
                        std::sort(tracesIn2.begin(), tracesIn2.end());
                        tracesIn1.erase(std::unique(tracesIn1.begin(), tracesIn1.end()), tracesIn1.end());
                        tracesIn2.erase(std::unique(tracesIn2.begin(), tracesIn2.end()), tracesIn2.end());
                        std::vector<SymbolicTrace> tracesOnlyIn1;
                        std::vector<SymbolicTrace> tracesOnlyIn2;
                        std::set_difference(tracesIn1.begin(), tracesIn1.end(), tracesIn2.begin(), tracesIn2.end(), std::back_inserter(tracesOnlyIn1));
                        std::set_difference(tracesIn2.begin(), tracesIn2.end(), tracesIn1.begin(), tracesIn1.end(), std::back_inserter(tracesOnlyIn2));
                        assert((not tracesOnlyIn1.empty()) or (not tracesOnlyIn2.empty()));
                        exceptionMessage += "\n";
                        if(not tracesOnlyIn1.empty()) {
                            exceptionMessage += "The following abstract symbolic traces are only possible after " + std::to_string(trace1);
                            for(auto const &trace : tracesOnlyIn1) {
                                exceptionMessage += "\n";
                                exceptionMessage += std::to_string(trace);
                                exceptionMessage += "\n\ttraversing states:";
                                for(auto const &target : targets1) {
                                    auto const stateTraces = getStateTracesForSymbolicTrace(abstractSfsm, target, trace);
                                    for(auto const &stateTrace : stateTraces) {
                                        exceptionMessage += "\n\t";
                                        exceptionMessage += std::to_string(stateTrace);
                                    }
                                }
                            }
                        }
                        exceptionMessage += "\n";
                        if(not tracesOnlyIn2.empty()) {
                            exceptionMessage += "The following abstract symbolic traces are only possible after " + std::to_string(trace2);
                            for(auto const &trace : tracesOnlyIn2) {
                                exceptionMessage += "\n";
                                exceptionMessage += std::to_string(trace);
                                exceptionMessage += "\n\ttraversing states:";
                                for(auto const &target : targets2) {
                                    auto const stateTraces = getStateTracesForSymbolicTrace(abstractSfsm, target, trace);
                                    for(auto const &stateTrace : stateTraces) {
                                        exceptionMessage += "\n\t";
                                        exceptionMessage += std::to_string(stateTrace);
                                    }
                                }
                            }
                        }
                        throw std::runtime_error(exceptionMessage);
                    }
                    auto newTrace1 = trace1;
                    newTrace1.insert(newTrace1.end(), distTrace->begin(), distTrace->end());
                    auto newTrace2 = trace2;
                    newTrace2.insert(newTrace2.end(), distTrace->begin(), distTrace->end());
                    result.push_back(newTrace1);
                    result.push_back(newTrace2);
                }
            } else {
                auto const &distTrace = distTraceGenerator.getDistinguishingTrace(trace1, trace2);
                if(distTrace.has_value()) {
                    auto newTrace1 = trace1;
                    newTrace1.insert(newTrace1.end(), distTrace->begin(), distTrace->end());
                    auto newTrace2 = trace2;
                    newTrace2.insert(newTrace2.end(), distTrace->begin(), distTrace->end());
                    result.push_back(newTrace1);
                    result.push_back(newTrace2);
                }
            }
        }
    }
    std::cerr << "100%" << std::endl;
    return toInputTraces(sfsm, result);
}

void checkForStatePreservation(Sfsm const &sfsm, std::vector<size_t> const &statesToCover, Expressions const &equivalenceClasses) {
    for(size_t stateIndex1 = 0; stateIndex1 < statesToCover.size() - 1; ++stateIndex1) {
        auto const state1 = statesToCover.at(stateIndex1);
        for(size_t stateIndex2 = stateIndex1; stateIndex2 < statesToCover.size(); ++stateIndex2) {
            auto const state2 = statesToCover.at(stateIndex2);
            if(auto const abstrDistTrace = getDistinguishingTraceForAbstractedSfsm(sfsm, equivalenceClasses, state1, state1); abstrDistTrace.has_value()) {
                auto const distTrace = getDistinguishingTraceForSfsm(sfsm, state1, state1);
                if(not distTrace.has_value()) {
                    std::string errorString = "Abstraction is not state preserving: States " + std::to_string(state1) + " and " + std::to_string(state2);
                    errorString += " are distinguishable under abstraction but not in the actual SFSM.";
                    errorString += "\nDistinguishing abstract trace: " + std::to_string(abstrDistTrace.value());
                    throw std::runtime_error(errorString);
                }
            }
        }
    }
}

bool isPrefix(ConcreteTrace const &potentialPrefix, ConcreteTrace const &trace) {
    if(potentialPrefix.size() > trace.size()) {
        return false;
    }
    for(size_t idx = 0; idx < potentialPrefix.size(); ++idx) {
        if(not (potentialPrefix.at(idx) == trace.at(idx))) {
            return false;
        }
    }
    return true;
}

template<typename IIter1, typename IIter2, typename OIter>
void filterPrefixes(IIter1 begin, IIter2 &&end, OIter &&out) {
    if(begin != end) {
        for(auto iterNext = begin; ++iterNext != end; ++begin) {
            if(not isPrefix(*begin, *iterNext)) {
                *out = *begin;
            }
        }
        *out = *begin;
    }
}

Sfsm removeInfeasibleTransitions(Sfsm const &sfsm) {
    if(auto const unsatisfiableTransitions = applyProcessor<UnsatisfiableTransitionsChecker>(sfsm);
        not unsatisfiableTransitions.empty()) {
        std::cerr << "Warning: Found infeasible transitions:" << std::endl;
        for(auto const transition : unsatisfiableTransitions) {
            std::cerr << "Infeasible transition from state '";
            std::cerr << sfsm.getStates().at(transition->getSourceId()).getName();
            std::cerr << "' to state '";
            std::cerr << sfsm.getStates().at(transition->getTargetId()).getName();
            std::cerr << "' with guard condition '";
            std::cerr << std::to_string(sfsm.getGuardConditions().at(transition->getInputId()));
            std::cerr << "' and output expression '";
            std::cerr << std::to_string(sfsm.getOutputExpressions().at(transition->getOutputId()));
            std::cerr << "'" << std::endl;
        }
        std::cerr << "The test suite generation may show unexpected behaviour." << std::endl;

        // Prepare a set of the indices of the guards and output expressions
        // used in the infeasible transitions
        std::vector<Index> guardConditionIdsOfInfeasibleTransitions;
        std::vector<Index> outputExpressionIdsOfInfeasibleTransitions;
        for(auto const unsatisfiableTransition : unsatisfiableTransitions) {
            guardConditionIdsOfInfeasibleTransitions.push_back(unsatisfiableTransition->getInputId());
            outputExpressionIdsOfInfeasibleTransitions.push_back(unsatisfiableTransition->getOutputId());
        }
        std::sort(guardConditionIdsOfInfeasibleTransitions.begin(), guardConditionIdsOfInfeasibleTransitions.end());
        std::sort(outputExpressionIdsOfInfeasibleTransitions.begin(), outputExpressionIdsOfInfeasibleTransitions.end());
        guardConditionIdsOfInfeasibleTransitions.erase(
            std::unique(guardConditionIdsOfInfeasibleTransitions.begin(),
                        guardConditionIdsOfInfeasibleTransitions.end()),
            guardConditionIdsOfInfeasibleTransitions.end());
        outputExpressionIdsOfInfeasibleTransitions.erase(
            std::unique(outputExpressionIdsOfInfeasibleTransitions.begin(),
                        outputExpressionIdsOfInfeasibleTransitions.end()),
            outputExpressionIdsOfInfeasibleTransitions.end());
        
        // Prepare a set of the indices of the guards and output expressions
        // used in all other transitions
        std::vector<Index> otherGuardIndices;
        std::vector<Index> otherOutputIndices;
        for(auto const &state : sfsm.getStates()) {
            for(auto const &transition : state.getTransitions()) {
                if(std::find(unsatisfiableTransitions.begin(),
                             unsatisfiableTransitions.end(),
                             &transition)
                    == unsatisfiableTransitions.end()) {
                    otherGuardIndices.push_back(transition.getInputId());
                    otherOutputIndices.push_back(transition.getOutputId());
                }
            }
        }
        std::sort(otherGuardIndices.begin(), otherGuardIndices.end());
        std::sort(otherOutputIndices.begin(), otherOutputIndices.end());
        otherGuardIndices.erase(
            std::unique(otherGuardIndices.begin(), otherGuardIndices.end()),
            otherGuardIndices.end());
        otherOutputIndices.erase(
            std::unique(otherOutputIndices.begin(), otherOutputIndices.end()),
            otherOutputIndices.end());

        // Calculate the guard conditions und output expressions to remove
        std::vector<Index> guardsToRemove;
        std::vector<Index> outputsToRemove;
        std::set_difference(guardConditionIdsOfInfeasibleTransitions.cbegin(),
                            guardConditionIdsOfInfeasibleTransitions.cend(),
                            otherGuardIndices.cbegin(),
                            otherGuardIndices.cend(),
                            std::back_inserter(guardsToRemove));
        std::set_difference(outputExpressionIdsOfInfeasibleTransitions.cbegin(),
                            outputExpressionIdsOfInfeasibleTransitions.cend(),
                            otherOutputIndices.cbegin(),
                            otherOutputIndices.cend(),
                            std::back_inserter(outputsToRemove));
        for(auto const guardIdx : guardsToRemove) {
            std::cerr << "Removing guard condition '" << std::to_string(sfsm.getUsedGuardConditions().at(guardIdx)) << "' from input alphabet as it is not used by any feasible transition." << std::endl;
        }
        for(auto const outputIdx : outputsToRemove) {
            std::cerr << "Removing output expression '" << std::to_string(sfsm.getUsedOutputExpressions().at(outputIdx)) << "' from output alphabet as it is not used by any feasible transition." << std::endl;
        }

        // Remove the guard conditions and output expressions from the sets of
        // all guard conditions and output expressions used
        auto allCurrentGuards = iota<std::vector<size_t>>(0, sfsm.getInputAlphabetSize()-1);
        auto allCurrentOutputs = iota<std::vector<size_t>>(0, sfsm.getOutputAlphabetSize()-1);
        std::vector<Index> newGuards;
        std::vector<Index> newOutputs;
        std::set_difference(allCurrentGuards.cbegin(), allCurrentGuards.cend(),
                            guardsToRemove.cbegin(), guardsToRemove.cend(),
                            std::back_inserter(newGuards));
        std::set_difference(allCurrentOutputs.cbegin(), allCurrentOutputs.cend(),
                            outputsToRemove.cbegin(), outputsToRemove.cend(),
                            std::back_inserter(newOutputs));
        std::decay_t<decltype(sfsm.getUsedGuardConditions())> newGuardConditions;
        std::decay_t<decltype(sfsm.getUsedOutputExpressions())> newOutputExpressions;
        // Generate a map from old to new IDs
        std::map<Index, Index> guardMap;
        std::map<Index, Index> outputMap;
        for(size_t idx = 0; idx < newGuards.size(); ++idx) {
            guardMap[newGuards.at(idx)] = idx;
            if(newGuards.at(idx) < sfsm.getUsedGuardConditions().size()) {
                newGuardConditions.push_back(sfsm.getUsedGuardConditions().at(idx));
            }
        }
        for(size_t idx = 0; idx < newOutputs.size(); ++idx) {
            outputMap[newOutputs.at(idx)] = idx;
            if(newOutputs.at(idx) < sfsm.getUsedOutputExpressions().size()) {
                newOutputExpressions.push_back(sfsm.getUsedOutputExpressions().at(idx));
            }
        }

        // Removing transitions that are not feasible
        auto const &states = sfsm.getStates();
        std::vector<State> transformedStates;
        std::transform(states.cbegin(), states.cend(), std::back_inserter(transformedStates),
            [&unsatisfiableTransitions,&guardMap,&outputMap](auto const &state) {
            std::vector<Transition> transitionsToKeep;
            auto const &transitions = state.getTransitions();
            std::remove_copy_if(transitions.cbegin(), transitions.cend(), std::back_inserter(transitionsToKeep),
                [&unsatisfiableTransitions](auto const &transition) {
                return std::find(unsatisfiableTransitions.cbegin(),
                                 unsatisfiableTransitions.cend(),
                                 &transition)
                        != unsatisfiableTransitions.cend();
            });
            std::vector<Transition> translatedTransitions;
            std::transform(transitionsToKeep.cbegin(), transitionsToKeep.cend(), std::back_inserter(translatedTransitions),
                [&guardMap,&outputMap](auto const &transition) {
                    auto const &[pre, input, output, post] = transition;
                    return Transition(pre, guardMap.at(input), outputMap.at(output), post);
            });

            return state.modify(std::nullopt, std::nullopt, translatedTransitions);
        });

        return *sfsm.modify(transformedStates,
                            std::nullopt,
                            newGuardConditions,
                            newOutputExpressions,
                            std::nullopt,
                            std::nullopt,
                            std::nullopt,
                            std::nullopt,
                            std::nullopt);
    }
    return sfsm;
}

Sfsm preprocessSfsm(Sfsm const &sfsm) {
    return removeInfeasibleTransitions(sfsm);
}

std::vector<ConcreteTrace> sfsmMethod(Sfsm const &sfsm, size_t stateDifference, Expressions const &propositions) {
    Sfsm preprocessedSfsm = preprocessSfsm(sfsm);
    auto const statesToCover = reachableStates(preprocessedSfsm);
    Expressions equivalenceClasses;

    if(not propositions.empty()) {
        std::vector<ConjunctionResult> const propositionEquivalenceClasses 
            = findAllSatisfiableConjunctions(preprocessedSfsm.getIdentifierDefinitions(), preprocessedSfsm.getInvariant(), propositions);
        std::transform(propositionEquivalenceClasses.cbegin(), propositionEquivalenceClasses.cend(), std::back_inserter(equivalenceClasses), [](auto const &elem) {
            return std::get<1>(elem);
        });
    } else if(applyProcessor<RestrictednessChecker>(sfsm, propositions)) {
        std::cerr << "Model fulfills conditions for FSM based generation. Using that approach." << std::endl;
        auto translationFunction = applyProcessor<RestrictedSfsmDistinguishingInputCalculator>(sfsm);
        auto fsm = applyProcessor<ToFSMAbstraction>(sfsm);
        auto concreteTestSuiteGenerator = libfsmtest::TestGenerationFrame<libfsmtest::WMethod>("Test suite", fsm->clone(), stateDifference);
        concreteTestSuiteGenerator.generateTestSuite();
        auto concreteTestSuite = concreteTestSuiteGenerator.getTestSuite()->getInputTracesRaw();
        std::vector<libfsmtest::IOTrace> ioTraces;
        for(auto const &inputTrace : concreteTestSuite) {
            auto translatedTraces = fsm->getIOTraces(inputTrace, false);
            ioTraces.insert(ioTraces.end(), translatedTraces.begin(), translatedTraces.end());
        }
        std::sort(ioTraces.begin(), ioTraces.end(), [](auto const &trace1, auto const &trace2) {
            auto inputLess = std::lexicographical_compare(trace1.input.cbegin(), trace1.input.cend(), trace2.input.cbegin(), trace2.input.cend());
            auto inputGreater = std::lexicographical_compare(trace2.input.cbegin(), trace2.input.cend(), trace1.input.cbegin(), trace1.input.cend());
            auto outputLess = std::lexicographical_compare(trace1.output.cbegin(), trace1.output.cend(), trace2.output.cbegin(), trace2.output.cend());
            return inputLess || (not inputGreater && outputLess);
        });
        ioTraces.erase(std::unique(ioTraces.begin(), ioTraces.end()), ioTraces.end());
        std::vector<ConcreteTrace> testSuite;
        for(auto const &ioTrace : ioTraces) {
            ConcreteTrace translatedTestCase;
            auto inputIter = ioTrace.input.cbegin();
            auto outputIter = ioTrace.output.cbegin();
            for(size_t idx = 0; idx < ioTrace.input.size(); ++idx) {
                translatedTestCase.push_back(translationFunction.at({*inputIter, *outputIter}));
                ++inputIter; ++outputIter;
            }
            testSuite.push_back(translatedTestCase);
        }
        std::sort(testSuite.begin(), testSuite.end());
        testSuite.erase(std::unique(testSuite.begin(), testSuite.end()), testSuite.end());
        return testSuite;
    }
    if(auto const envVarPtr = getenv("LIBSFSMTEST_TEST_SUITE_TEX");
        envVarPtr != nullptr) {
        std::string const texOutputFile = envVarPtr;
        std::ofstream output(texOutputFile, std::ios_base::app);
        output << "\\section{Test suite calculation}" << std::endl;
    }
    checkForStatePreservation(preprocessedSfsm, statesToCover, equivalenceClasses);
    auto const partOne = generatePartOne(preprocessedSfsm, statesToCover);
    std::cerr << "State cover traces:" << std::endl;
    for(auto const &trace : partOne) {
        std::cerr << std::to_string(trace) << std::endl;
    }
    auto const partTwo = generatePartTwo(preprocessedSfsm, statesToCover);
    std::cerr << "Distinguishing traces:" << std::endl;
    for(auto const &trace : partTwo) {
        std::cerr << std::to_string(trace) << std::endl;
    }
    auto const partThree = generatePartThree(preprocessedSfsm, statesToCover, stateDifference, propositions);
    std::cerr << "Traversal set:" << std::endl;
    for(auto const &trace : partThree) {
        std::cerr << std::to_string(trace) << std::endl;
    }
    auto const partFour = generatePartFour(preprocessedSfsm, statesToCover, stateDifference, equivalenceClasses, propositions);
    auto const partFive = generatePartFive(preprocessedSfsm, statesToCover, stateDifference, equivalenceClasses, propositions);

    std::vector<ConcreteTrace> testSuite = partOne;
    testSuite.insert(testSuite.end(), partTwo.begin(), partTwo.end());
    testSuite.insert(testSuite.end(), partThree.begin(), partThree.end());
    testSuite.insert(testSuite.end(), partFour.begin(), partFour.end());
    testSuite.insert(testSuite.end(), partFive.begin(), partFive.end());

    std::sort(testSuite.begin(), testSuite.end());
    auto const newEnd = std::unique(testSuite.begin(), testSuite.end());
    testSuite.erase(newEnd, testSuite.end());
    std::vector<ConcreteTrace> prefixFreeTestSuite;
    filterPrefixes(testSuite.begin(), testSuite.end(), std::back_inserter(prefixFreeTestSuite));
    if(auto const envVarPtr = getenv("LIBSFSMTEST_TEST_SUITE_TEX");
        envVarPtr != nullptr) {
        std::string const texOutputFile = envVarPtr;
        std::ofstream output(texOutputFile, std::ios_base::app);
        output << "\\subsection{Test suite}" << std::endl;
        output << "\\begin{align*}" << std::endl;
        for(auto const &testTrace : prefixFreeTestSuite) {
            output << "&" << std::to_string(testTrace) << std::endl;
        }
        output << "\\end{align*}" << std::endl;
    }
    std::cerr << "Test suite generation complete!" << std::endl;
    return prefixFreeTestSuite;
}

}
