/*
 * Copyright 2022 Niklas Krafczyk, Jan Peleska
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#pragma once

#include <vector>
#include <utility>
#include <algorithm>
#include <iterator>

#include "sfsm/Expressions.hpp"
#include "sfsm/ExpressionDefinitions.hpp"
#include "sfsm/SMTTools.hpp"

namespace libfsmtest::SFSMs {

    inline std::vector<std::pair<std::vector<ssize_t>, Expression>> findAllSatisfiableConjunctions(IdentifierDefinitions const &definitions, Expression const &invariant, Expressions const &expressions) {
        Expressions negatedExpressions;
        std::transform(expressions.cbegin(), expressions.cend(), std::back_inserter(negatedExpressions), [](auto const &expression) {
            return makeLogicalNegation(expression);
        });
        std::vector<Expressions> currentRound, nextRound;
        currentRound.emplace_back();

        for(size_t expressionIndex = 0; expressionIndex < expressions.size(); ++expressionIndex) {
            for(auto const &currExpressions : currentRound) {
                Expression newExpressionPositive;
                Expression newExpressionNegative;
                if (expressionIndex == 0) {
                    newExpressionPositive = expressions.at(expressionIndex);
                    newExpressionNegative = negatedExpressions.at(expressionIndex);
                } else {
                    newExpressionPositive = makeLogicalConjunction(makeLogicalConjunction(currExpressions), expressions.at(expressionIndex));
                    newExpressionNegative = makeLogicalConjunction(makeLogicalConjunction(currExpressions), negatedExpressions.at(expressionIndex));
                }
                auto positiveWithInvariant = makeLogicalConjunction(newExpressionPositive, invariant);
                auto negativeWithInvariant = makeLogicalConjunction(newExpressionNegative, invariant);

                // If there is no solution to newExpressionPositive
                // we know that there is a solution to newExpressionNegative as currExpressions was satisfiable previously.
                if(not isSatisfiable(positiveWithInvariant, definitions).has_value()) {
                    nextRound.emplace_back(currExpressions);
                    nextRound.back().push_back(negatedExpressions.at(expressionIndex));
                } else {
                    nextRound.emplace_back(currExpressions);
                    nextRound.back().push_back(expressions.at(expressionIndex));
                    // However, if there is a solution to newExpressionPositive
                    // there still might be a solution to newExpressionNegative
                    if(isSatisfiable(negativeWithInvariant, definitions).has_value()) {
                        nextRound.emplace_back(currExpressions);
                        nextRound.back().push_back(negatedExpressions.at(expressionIndex));
                    }
                }
            }
            currentRound = std::move(nextRound);
        }

        std::vector<std::pair<std::vector<ssize_t>, Expression>> result;
        std::transform(currentRound.cbegin(), currentRound.cend(), std::back_inserter(result), 
          [&expressions](auto const &exp) -> std::pair<std::vector<ssize_t>, Expression> {
            std::vector<ssize_t> indices;
            for(auto const &expression : exp) {
                auto const index = indices.size();
                if(expression == expressions.at(index)) {
                    indices.push_back(static_cast<ssize_t>(index+1));
                } else {
                    assert(expression == makeLogicalNegation(expressions.at(index)));
                    indices.push_back(static_cast<ssize_t>(-(index + 1)));
                }
            }
            return {indices, makeLogicalConjunction(exp)};
        });

        return result;
    }

}

