/*
 * Copyright 2022 Jan Peleska, Moritz Bergenthal, Niklas Krafczyk
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#include "visitors/output/ToDotFileVisitor.hpp"

#include <fstream>
#include <stdexcept>
#include <string>
#include <algorithm>

#include "sfsm/State.hpp"
#include "sfsm/Transition.hpp"
#include "sfsm/Sfsm.hpp"
#include "sfsm/Expressions.hpp"

using namespace std;

namespace libfsmtest::SFSMs {

void ToDotFileVisitor::writeToFile() {
    std::ofstream outStream(fileName);
    if (!outStream.is_open()) throw std::runtime_error("Failed to open output file");
    
    outStream << "digraph g { " << endl;
    
    outStream << dotAsString.str() << endl << "}" << endl;
}

void ToDotFileVisitor::visit(State& state) {
    
    if ( state.getId() != initialState ) {
        // If state name equals its number , do not
        // add ( number )
        if ( state.getName() == std::to_string(state.getId()) ) {
            dotAsString << state.getId()
            << R"([fontname="helvetica",label=")"
            << state.getName()
            << "\"];" << endl;
        }
        else {
            dotAsString << state.getId()
            << R"([fontname="helvetica",label=")"
            << state.getName()
            << "(" << state.getId() << ")"
            << "\"];" << endl;
        }
    }
    else {
        // Special format for initial state
        if ( state.getName() == std::to_string(state.getId()) ) {
            dotAsString << state.getId()
            << R"([shape=doublecircle,fontname="helvetica",label=")"
            << state.getName()
            << "\"];" << endl;
        }
        else {
            dotAsString << state.getId()
            << R"([shape=doublecircle,fontname="helvetica",label=")"
            << state.getName()
            << "(" << state.getId() << ")"
            << "\"];" << endl;
        }
    }
    
    for(auto &tr : state.getTransitions()) {
        tr.accept(*this);
    }
}

void ToDotFileVisitor::visit(Transition& tr) {
    
    std::string inStr = std::to_string(sfsmInputAlphabet.at(tr.getInputId()));
    std::string outStr = std::to_string(sfsmOutputAlphabet.at(tr.getOutputId()));
    
    if (tr.getSourceId() != tr.getTargetId() ) {
        dotAsString << tr.getSourceId()
        << " -> "
        << tr.getTargetId()
        << R"([fontname="helvetica",label=")"
        << inStr
        << "/"
        << outStr
        << "\"];" << endl;
    }
    else {
        // Leave more space for labels attached to self-loops
        dotAsString << tr.getSourceId()
        << " -> "
        << tr.getTargetId()
        << R"([fontname="helvetica",label="  )"
        << inStr
        << "/"
        << outStr
        << "  \"];" << endl;
    }
    
}

void ToDotFileVisitor::createDotAsString(Sfsm& sfsm) {
    
    this->sfsmInputAlphabet = sfsm.getGuardConditions();
    this->sfsmOutputAlphabet = sfsm.getOutputExpressions();
    
    for( auto &state : sfsm.getStates() ) {
        state.accept(*this);
    }
}

void ToDotFileVisitor::visit(Sfsm& sfsm)  {
    initialState = sfsm.getInitialStateId();
    createDotAsString(sfsm);
}


} //namespace libfsmtest
