/*
 * Copyright 2021 Jan Peleska, Moritz Bergenthal, Niklas Krafczyk
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#pragma once

#include <string>
#include <sstream>      // std::ostringstream
#include <vector>

#include "sfsm/definitions.hpp"
#include "sfsm/Sfsm.hpp"
#include "visitors/SfsmVisitor.hpp"

namespace libfsmtest::SFSMs {

/**
 * Class for writing a SFSMs to GraphViz format (also called dot-format)
 */
class ToDotFileVisitor : public SfsmVisitor {
private:
    std::string fileName;
    Index initialState;

    std::ostringstream dotAsString;
    
    void createDotAsString(Sfsm& sfsm);
    
    /** SFSMs */
    std::vector<Sfsm::GuardCondition_t> sfsmInputAlphabet;
    std::vector<Sfsm::OutputExpression_t> sfsmOutputAlphabet;

public:
    /**
     * Constructor.
     * @see visit(Sfsm&)
     * @see writeToFile()
     * @param fileNameParam The name of the file to write to.
     */
    explicit ToDotFileVisitor(const std::string& fileNameParam) {
        this->initialState = 0;
        this->fileName = fileNameParam;
    }

    /** SFSMs */
    /**
     * Starts gathering data of an SFSM.
     * Only called indirectly. This uses the
     * same method createDotAsString() as the Fsm-visit method.
     * @param sfsm The SFSM to gather data from.
     */
    void visit(Sfsm& sfsm) override;
    /**
     * Starts gathering data of an SFSM state.
     * Only called indirectly.
     * @see State::accept()
     * @param state The SFSM state to gather data from.
     */
    void visit(State& state) override;
    /**
     * Starts gathering data of an SFSM transition.
     * Only called indirectly.
     * @see Transition::accept()
     * @param transition The SFSM transition to gather data from.
     */
    void visit(Transition& transition) override;

    /**
     * Writes out gathered data to the given file.
     * Has to be called after this visitor has been accepted on the DFSM.
     */
    void writeToFile();

};

} // namespace libfsmtest
