/*
 * Copyright 2022 Niklas Krafczyk, Jan Peleska
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <cassert>
#include <string>
#include <iostream>
#include <cassert>

#include "version.hpp"
#include "sfsm/Expressions.hpp"
#include "creators/input/SfsmFromFileCreator.hpp"
#include "utils/algorithms/FindAllSatisfiableConjunctions.hpp"
#include "sfsm/TransitionRelationGenerator.hpp"
#include "visitors/output/ToDotFileVisitor.hpp"
#include "testgeneration/SfsmMethod.hpp"
#include "creators/transformers/AbstractionTransformer.hpp"

#ifndef RESOURCEPATH
#define RESOURCEPATH std::string("../../resources/")
#endif

using namespace libfsmtest;

[[maybe_unused]] static void demo_ExpressionOperations() {
    auto x = SFSMs::makeIdentifier("x");
    auto y = SFSMs::makeIdentifier("y");
    auto twentythree = SFSMs::makeIdentifier("23");
    auto two = SFSMs::makeIdentifier("2");
    auto mult = SFSMs::makeArithmeticMultiplication(x, y);
    auto equals = SFSMs::makeComparisonEquality(mult, twentythree);
    auto xLessThan24 = SFSMs::makeComparisonLess(x, twentythree);
    auto yLessThan24 = SFSMs::makeComparisonLess(y, twentythree);
    auto xGreaterThan2 = SFSMs::makeComparisonGreater(x, two);
    auto yGreaterThan2 = SFSMs::makeComparisonGreater(y, two);
    auto everything = SFSMs::makeLogicalConjunction(xLessThan24, yLessThan24, xGreaterThan2, yGreaterThan2, equals);

    SFSMs::IdentifierDefinitions definitions;
    definitions.emplace_back("x", SFSMs::IdentifierType{SFSMs::IdentifierType::SimpleType(SFSMs::IdentifierType::SimpleBaseType::REAL), SFSMs::IdentifierType::LocationInModel::INPUT});
    definitions.emplace_back("y", SFSMs::IdentifierType{SFSMs::IdentifierType::SimpleType(SFSMs::IdentifierType::SimpleBaseType::REAL), SFSMs::IdentifierType::LocationInModel::OUTPUT});
    auto sat = SFSMs::isSatisfiable(everything, definitions);
    if(sat.has_value()) {
        std::cout << "Formula \"" << std::to_string(everything) << "\" has solution: " << std::endl;
        for(auto const &[id, value] : sat.value()) {
            auto const &[name, type] = id;
            std::cout << "  " << name << ": " << value << std::endl;
        }
    }
}

[[maybe_unused]] static void demo_SfsmFromFileCreator() {
    std::unique_ptr<SFSMs::Sfsm> sfsm = SFSMs::createSfsm<SFSMs::SfsmFromFileCreator>(RESOURCEPATH "sfsm.sfsm");
    SFSMs::ToDotFileVisitor dotFsm("f-sfsm.dot");
    sfsm->accept(dotFsm);
    dotFsm.writeToFile();
}

[[maybe_unused]] static void demo_SfsmFromCsvFileCreator() {
    std::unique_ptr<SFSMs::Sfsm> sfsm = SFSMs::createSfsm<SFSMs::SfsmFromFileCreator>(RESOURCEPATH "garage-door-controller-symbolic.csv");
    SFSMs::ToDotFileVisitor dotFsm("csv-sfsm.dot");
    sfsm->accept(dotFsm);
    dotFsm.writeToFile();
}

[[maybe_unused]] static void demo_EquivalenceClassCalculation() {
    std::unique_ptr<SFSMs::Sfsm> sfsm = SFSMs::createSfsm<SFSMs::SfsmFromFileCreator>(RESOURCEPATH "brake/brake.csv");
    SFSMs::Expressions APs;
    APs.push_back(SFSMs::makeComparisonLessOrEquals(SFSMs::makeIdentifier("y"), SFSMs::makeIdentifier("B1")));
    APs.push_back(SFSMs::makeComparisonGreater(SFSMs::makeIdentifier("x"), SFSMs::makeIdentifier("max")));
    APs.push_back(SFSMs::makeComparisonGreaterOrEquals(SFSMs::makeIdentifier("y"), makeArithmeticSubtraction(SFSMs::makeIdentifier("B2"), SFSMs::makeIdentifier("0.1"))));
    APs.push_back(SFSMs::makeComparisonLess(SFSMs::makeIdentifier("x"), makeArithmeticSubtraction(SFSMs::makeIdentifier("max"), SFSMs::makeIdentifier("delta"))));
    auto guardConditions = sfsm->getGuardConditions();
    auto outputExpressions = sfsm->getOutputExpressions();

    SFSMs::Expressions allExpressions;
    allExpressions.insert(allExpressions.end(), guardConditions.cbegin(), guardConditions.cend());
    allExpressions.insert(allExpressions.end(), outputExpressions.cbegin(), outputExpressions.cend());
    allExpressions.insert(allExpressions.end(), APs.cbegin(), APs.cend());
    
    auto definitions = sfsm->getIdentifierDefinitions();
    auto equivalenceClasses = SFSMs::findAllSatisfiableConjunctions(definitions, sfsm->getInvariant(), allExpressions);
    std::cout << std::endl << "Equivalence classes for the brake example:" << std::endl;
    for(auto const &[_, eqClass] : equivalenceClasses) {
        std::cout << std::to_string(eqClass) << std::endl;
        auto const model = SFSMs::isSatisfiable(makeLogicalConjunction(eqClass, sfsm->getInvariant()), definitions).value();
        SFSMs::IdentifierValuations inputOutputValuations;
        std::copy_if(model.begin(), model.end(), std::back_inserter(inputOutputValuations), [&definitions](auto const &valuation) {
            auto const &[id, value] = valuation;
            auto const &[name, _1] = id;
            auto identifierIter = std::find_if(definitions.cbegin(), definitions.cend(), [&name=name](auto const &definition) {
                auto const &[idName, _trash] = definition;
                return name == idName;
            });
            assert(identifierIter != definitions.cend());
            auto const &[_2, type] = *identifierIter;
            return type.getLocation() != SFSMs::IdentifierType::LocationInModel::OTHER;
        });
        for(auto const &[id, value] : inputOutputValuations) {
            auto const &[name, type] = id;
            std::cout << name << " = " << value << std::endl;
        }
        std::cout << std::endl;
    }
    std::cout << std::endl;
}

[[maybe_unused]] static void demo_DistinguishingTraceCalculation() {
    std::unique_ptr<SFSMs::Sfsm> sfsm = SFSMs::createSfsm<SFSMs::SfsmFromFileCreator>(RESOURCEPATH "alarm/alarm.csv");
    auto const [_, problemStatements] = findDistinguishingTraceProblem(*sfsm, generateTransitionRelationDefinitionForSfsm(*sfsm), 0, 1);
    for(auto const &statement : problemStatements) {
        std::cout << std::to_string(statement) << std::endl;
    }

    auto result = getDistinguishingTraceForSfsm(*sfsm, 0, 1).value();
    for(size_t index = 0; index < result.size(); ++index) {
        auto const &valuationFunction = result.at(index);
        for(auto const &[id, valuation] : valuationFunction) {
            auto const &[name, type] = id;
            std::cout << std::endl << index << ": " << std::endl;
            std::cout << "  " << name << ": " << valuation << std::endl;
        }
    }
}

[[maybe_unused]] static void demo_SfsmMethodPOT() {
    std::unique_ptr<SFSMs::Sfsm> sfsm = SFSMs::createSfsm<SFSMs::SfsmFromFileCreator>(RESOURCEPATH "brake/brake.csv");
    /*
     * props = { (y <= B1), (y >= B2 - 0.1), (x > max), (x < max - delta) }
     */
    SFSMs::Expressions const props = {SFSMs::makeComparisonLessOrEquals(SFSMs::makeIdentifier("y"), SFSMs::makeIdentifier("B1")),
                               SFSMs::makeComparisonGreaterOrEquals(SFSMs::makeIdentifier("y"), SFSMs::makeArithmeticSubtraction(SFSMs::makeIdentifier("B2"), SFSMs::makeIdentifier("0.1"))),
                               SFSMs::makeComparisonGreater(SFSMs::makeIdentifier("x"), SFSMs::makeIdentifier("max")),
                               SFSMs::makeComparisonLess(SFSMs::makeIdentifier("x"), SFSMs::makeArithmeticSubtraction(SFSMs::makeIdentifier("max"), SFSMs::makeIdentifier("delta")))};
    std::unique_ptr<SFSMs::Sfsm> abstraction = SFSMs::transformSfsm<SFSMs::AbstractionTransformer>(sfsm.get(), props);
    SFSMs::ToDotFileVisitor brake("brake.csv");
    SFSMs::ToDotFileVisitor abstractBrake("abrake.csv");
    sfsm->accept(brake);
    brake.writeToFile();
    abstraction->accept(abstractBrake);
    abstractBrake.writeToFile();
    auto result = SFSMs::sfsmMethod(*sfsm, 0, props);
    std::cout << "Test suite POT (size: " << result.size() << "):" << std::endl;
    for(auto const &trace : result) {
        std::cout << std::to_string(trace) << std::endl;
    }
}

[[maybe_unused]] static void demo_SfsmMethod() {
    std::unique_ptr<SFSMs::Sfsm> sfsm = SFSMs::createSfsm<SFSMs::SfsmFromFileCreator>(RESOURCEPATH "brake/brake.csv");
    auto result = SFSMs::sfsmMethod(*sfsm, 0, {});
    std::cout << "Test suite (size: " << result.size() << "):" << std::endl;
    for(auto const &trace : result) {
        std::cout << std::to_string(trace) << std::endl;
    }
}

int main(int /* argc */, char** /* argv[] */) {
    std::cout << "Usage demo for the libsfsmtest " << version.major << "." << version.minor << std::endl;
    
    //demo_ExpressionOperations();

    //demo_SfsmFromFileCreator();

    //demo_SfsmFromCsvFileCreator();

    //demo_EquivalenceClassCalculation();

    //demo_DistinguishingTraceCalculation();

    demo_SfsmMethodPOT();
    demo_SfsmMethod();
    
    return 0;
}
